#include <v_lib.h>
#include <math.h>
#include <malloc.h>

#define	power2(op)	(1<<(op))

void fft_calc(mvec_t *ReIn,mvec_t *ImIn,mvec_t *ReOut,mvec_t *ImOut,
	mvec_t dir_sign,unsigned size_power)
{
	mvec_t *xr, *xi, *yr, *yi, w, p, si, co, xa, xb, sc;
	unsigned size,m,k,l,i,j,ni,jm,k1,l1,p1,p2,pm,is,js;
	size=(1+power2(size_power))*sizeof(mvec_t);
	xr=(mvec_t*)malloc(size);
	xi=(mvec_t*)malloc(size);
	yr=(mvec_t*)malloc(size);
	yi=(mvec_t*)malloc(size);
	memCopy(&xr[1],ReIn,size-sizeof(mvec_t));
	memCopy(&xi[1],ImIn,size-sizeof(mvec_t));
	p1=power2(size_power-1);
	for(m=1;m<=size_power;m++)
	{
		k1=power2(size_power-m)-1;
		pm=power2(m);
		p2=power2(m-1);
		p=M_PI/(mvec_t)p2;
		l1=p2-1;
		is=1;
		js=1;
		for(k=0;k<=k1;k++,is+=p2,js+=pm)
		{
			w=0;
			i=is;
			j=js;
			for(l=0;l<=l1;l++,i++,j++,w+=p)
			{
				ni=p1+i;
				jm=j+p2;
				si=dir_sign*(mvec_t)sin(w);
				co=(mvec_t)cos(w);
				xa=xr[ni]*co+xi[ni]*si;
				xb=xi[ni]*co-xr[ni]*si;
				yr[j]=xr[i]+xa;
				yi[j]=xi[i]+xb;
				yr[jm]=xr[i]-xa;
				yi[jm]=xi[i]-xb;
			}
		}
		memCopy(xr,yr,size);
		memCopy(xi,yi,size);
	}
	memCopy(ReOut,&xr[1],size-sizeof(mvec_t));
	memCopy(ImOut,&xi[1],size-sizeof(mvec_t));
	size=power2(size_power);
	if(dir_sign<0)
		sc=(mvec_t)size;
	else
		sc=(mvec_t)1/(mvec_t)size;
	for(i=0;i<size;i++)
	{
		ReOut[i]*=sc;
		ImOut[i]*=sc;
	}
	free(xr);
	free(xi);
	free(yr);
	free(yi);
}
	
/*
forw:
	for(i=0;i<size;i++)
	{
		Re[i]*=total_time;
		Im[i]*=total_time;
	}
inv:
	inv_len=1/total_time;
	for(i=0;i<size;i++)
	{
		Re[i]*=inv_len;
		Im[i]*=inv_len;
	}
*/

#define	SAMPLE_RATE (11025)

V_LOCAL	void
generate_sines(mvec_t *buffer,u32_t samples,u32_t channels,
	mvec_t freqs[],mvec_t volumes[],mvec_t *phase_info)
{
	u32_t s,c;
	mvec_t out;
	mvec_t *phase=(vec_t*)malloc(channels * sizeof(mvec_t));
	mvec_t *dph=(vec_t*)malloc(channels * sizeof(mvec_t));
	mvec_t *amp=(vec_t*)malloc(channels * sizeof(mvec_t));
	if(phase_info)
		memCopy(phase,phase_info,channels * sizeof(mvec_t));
	else
		memZero(phase,channels * sizeof(mvec_t));
	for(c=0;c<channels;c++)
	{
		amp[c] = volumes[c] / (mvec_t)channels;
		dph[c] = M_2PI * freqs[c] / (mvec_t)SAMPLE_RATE;
	}
	for(s=0;s<samples;s++)
	{
		out = 0;
		for(c=0;c<channels;c++)
		{
			out += amp[c] * (mvec_t)sin(phase[c]);
			phase[c]+=dph[c];
		}
		if(out < -1)
			out = -1;
		else
			if(out > 1)
				out = 1;
		buffer[s]=out;
	}
	if(phase_info)
		memCopy(phase_info,phase,channels * sizeof(mvec_t));
	free(phase);
	free(dph);
	free(amp);
}

int main()
{
	unsigned i,p=8;
	unsigned len=power2(p);
	mvec_t *Re,*Im,freqs[10],volumes[10];
	Re=(mvec_t*)memAlloc(len*sizeof(mvec_t));
	Im=(mvec_t*)memAlloc(len*sizeof(mvec_t));

	freqs[0]=100;
	volumes[0]=0.5;
	
	generate_sines(Re,len,1,freqs,volumes,NULL);
	
	fft_calc(Re,Im,Re,Im,1,p);

	for(i=0;i<len;i++)
		/* printf("[%8u]\tPower=%g\n", */
		printf("%8u\t%g\n",
			i,sqrt(Re[i]*Re[i]+Im[i]*Im[i]));

	memFree(Re);
	memFree(Im);
	return 0;
}

	
	