#pragma warning(disable:4115)
#pragma warning(disable:4214)
#pragma warning(disable:4201)
#include <windows.h>
#pragma warning(default:4115)
#include <GL/gl.h>

#ifdef __BORLANDC__
#include <float.h>	/* For masking floating point exceptions. */
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* Refresh driver parameters: visual */
typedef struct ref_visual_s
{
	struct
	{
		unsigned char hardware;		/* Hardware renderer */
        unsigned char offscreen;	/* Offscreen frame buffer */
		unsigned char pad[2];
	} flags;
	int	color_bits;		/* Bit depth of visual */
	int depth_bits;		/* Bit depth of Z - buffer */
	int stencil_bits;	/* Bit depth of stencil buffer */
}	ref_visual_t, *ref_visual_p;

/* Refresh driver window properties */
typedef enum
{
	ref_wm_title,		/* param is character string (of type char*) */
	ref_wm_left,		/* param is pointer to 32-bit unsigned */
	ref_wm_top,			/* param is pointer to 32-bit unsigned */
	ref_wm_width,		/* param is pointer to 32-bit unsigned */
	ref_wm_height,		/* param is pointer to 32-bit unsigned */
	ref_wm_icon			/* param is HICON */
}	ref_wm_property;

/* Refresh driver attributes */
typedef struct ref_visual_attributes_s
{
	struct
	{
		unsigned char force_software;
		unsigned char force_hardware;
        unsigned char force_offscreen;
	} flags;
	int min_color_bits;
	int min_depth_bits;
	int min_stencil_bits;
}	ref_visual_attributes_t, *ref_visual_attributes_p;

typedef long (__stdcall *ref_message_handler_t)(void*,unsigned,unsigned,long);

void	*refCreateView(ref_visual_attributes_t *attributes);
void	refDestroyView(void *context);
void	refSetWndProc(void *context,ref_message_handler_t proc);

/* IMPLEMENTATION */

/*
 *	Predefined sets of attributes to use as initializers for refChooseVisual().
 */
/* 'Choose what available' */
#define REF_VISUAL_ATTRIBUTES_DEFAULT 	{ {0,0,0}, 8, 16, 0 }
/* Hints for SW visual */
#define REF_VISUAL_ATTRIBUTES_SW 		{ {1,0,0}, 8, 16, 0 }
/* Hints for HW visual */
#define REF_VISUAL_ATTRIBUTES_HW 		{ {0,1,0}, 15, 16, 0 }
/* Hints for HW visual with stencil */
#define REF_VISUAL_ATTRIBUTES_FX 		{ {0,1,0}, 15, 16, 1 }

/*
 *	Converts OS-dependant visual info struct into ref_visual.
 *	Returns 1 on success, 0 otherwise.
 */
int
ref_visual_from_gdi_pfd(ref_visual_t *visual,PIXELFORMATDESCRIPTOR *pfd)
{
#ifndef PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED         0x00001000
#endif	/* !PFD_GENERIC_ACCELERATED */
	if(	(pfd->dwFlags & PFD_DRAW_TO_WINDOW && !(pfd->dwFlags & PFD_DOUBLEBUFFER)) || 
		!(pfd->dwFlags & PFD_SUPPORT_OPENGL) || 
		(pfd->iPixelType!=PFD_TYPE_RGBA))
		return 0;
	visual->flags.hardware=(unsigned char)(((pfd->dwFlags & PFD_GENERIC_FORMAT) != 0) ==
		((pfd->dwFlags & PFD_GENERIC_ACCELERATED) != 0) ? 1 : 0);
	visual->flags.offscreen=(unsigned char)((pfd->dwFlags & PFD_DRAW_TO_BITMAP) ? 1 : 0);
	visual->color_bits=pfd->cColorBits;
	visual->depth_bits=pfd->cDepthBits;
	visual->stencil_bits=pfd->cStencilBits;
	return 1;
}

int
ref_accept_visual(ref_visual_t *visual,ref_visual_attributes_t *attributes)
{
	/* Reject HW visual if attributes->flags.force_software */
	if(attributes->flags.force_software!=0 && visual->flags.hardware!=0)
		return 0;
	/* Reject SW visual if attributes->flags.force_hardware */
	if(attributes->flags.force_hardware!=0 && visual->flags.hardware==0)
		return 0;
	/*
	 *	Check for color, depth and stencil.
	 */
	if(visual->color_bits < attributes->min_color_bits ||
		visual->depth_bits < attributes->min_depth_bits ||
		visual->stencil_bits < attributes->min_stencil_bits)
		return 0;
	/*
	 *	The format is suitable.
	 */
	return 1;
}

/*
 *	TODO: choose logic of this procedure.
 */
int
ref_best_visual(ref_visual_t *visuals,int num_visuals)
{
	int i, charge, best_charge, best_was_hw, best;
	best=0;
	best_charge=0;
	best_was_hw=0;
	for(i=0;i<num_visuals;i++)
	{
		charge=visuals[i].color_bits + 
			visuals[i].depth_bits + 
			visuals[i].stencil_bits;
		if(best_was_hw)
		{
			if(visuals[i].flags.hardware && (charge>best_charge))
			{
				best_charge=charge;
				best=i;
			}
		}
		else
		{
			if(visuals[i].flags.hardware || (charge>best_charge))
			{
				best_was_hw=visuals[i].flags.hardware;
				best_charge=charge;
				best=i;
			}
		}
		/*
		printf("Visual %d:\tHW=%d\tCOLOR=%d\tDEPTH=%d\tSTENCIL=%d\n",i,
			visuals[i].flags.hardware,
			visuals[i].color_bits,
			visuals[i].depth_bits,
			visuals[i].stencil_bits);
		*/
	}
	printf("Best visual %d:\tHW=%d\tCOLOR=%d\tDEPTH=%d\tSTENCIL=%d\n",i,
		visuals[best].flags.hardware,
		visuals[best].color_bits,
		visuals[best].depth_bits,
		visuals[best].stencil_bits);
	return best;
}

typedef struct ref_context_s
{
	ref_visual_t visual;
	HWND		wnd;
	HDC			dc;
	HGLRC		rc;
	int			width, height;
	GLfloat		left, right, bottom, top; /* frustum */
	WNDCLASS	wc;
	ref_message_handler_t message_handler;
}	ref_context_t, *ref_context_p;

void
refSetWndProc(void *context,ref_message_handler_t proc)
{
	if(!context)
		return;
	if(proc)
		((ref_context_t*)context)->message_handler=proc;
	else
		((ref_context_t*)context)->message_handler=DefWindowProc;
}

void	conError(char *msg)
{
	char exe_name[256];
	GetModuleFileName(GetModuleHandle(NULL),exe_name,256);
	MessageBox(NULL,msg,exe_name,MB_OK|MB_ICONHAND);
	return;
}

ref_context_t *thread_context=NULL;

typedef unsigned u32_t;

void refSetWMProperty(ref_wm_property pname,void *param)
{
	RECT rc;
	ref_context_t *ctx=thread_context;
    if(!ctx)
    	return;
	switch(pname)
    {
    	case ref_wm_title:
        	SetWindowTextA(ctx->wnd,(char*)param);
        	break;
    	case ref_wm_left:
    		if(!param)
    			return;
        	GetWindowRect(ctx->wnd,&rc);
        	MoveWindow(ctx->wnd,*((u32_t*)param),rc.top,
        		rc.right-rc.left,rc.top-rc.bottom,TRUE);
        	break;
    	case ref_wm_top:
    		if(!param)
    			return;
        	GetWindowRect(ctx->wnd,&rc);
        	MoveWindow(ctx->wnd,rc.left,*((u32_t*)param),
        		rc.right-rc.left,rc.top-rc.bottom,TRUE);
        	break;
    	case ref_wm_width:
    		if(!param)
    			return;
        	GetWindowRect(ctx->wnd,&rc);
        	MoveWindow(ctx->wnd,rc.left,rc.top,*((u32_t*)param),
        		rc.top-rc.bottom,TRUE);
			ctx->width=*((u32_t*)param);
			glViewport(0,0,ctx->width,ctx->height);
        	break;
    	case ref_wm_height:
    		if(!param)
    			return;
        	GetWindowRect(ctx->wnd,&rc);
        	MoveWindow(ctx->wnd,rc.left,rc.top,
        		rc.right-rc.left,*((u32_t*)param),TRUE);
			ctx->height=*((u32_t*)param);
			glViewport(0,0,ctx->width,ctx->height);
        	break;
    	case ref_wm_icon:
			SendMessage(ctx->wnd,WM_SETICON,(WPARAM)TRUE,*((LPARAM*)param);
			SendMessage(ctx->wnd,WM_SETICON,(WPARAM)FALSE,*((LPARAM*)param);
        	break;
		default:
        	return;
	}
    return;
}

GLfloat a = 0, b = 0;

static void rect(void)
{
	glBegin(GL_TRIANGLE_FAN);
		glColor4f(1,1,0,.5f);
		glVertex2f(-1,-1);
		glColor4f(0,1,1,.5f);
		glVertex2f(-1,1);
		glColor4f(1,0,1,.5f);
		glVertex2f(1,1);
		glColor4f(1,1,1,.5f);
		glVertex2f(1,-1);
	glEnd();
}
		
static void redraw()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_DEPTH_TEST);
	rect();
	
	
	glScalef(.5f,.5f,.5f);
	glScalef(.5f,.5f,.5f);
	glRotatef(a,1,0,0);
	glRotatef(b,0,1,0);
	rect();
	SwapBuffers(wglGetCurrentDC());
	a+=1;
	b+=.2f;
}

/*
if(ctx)
{
	int i;
	GLfloat aspect;
	
	glClearColor(0,0.5f,0.5f,1);
	glClearDepth(1);
	glDepthFunc(GL_LESS);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	aspect=(GLfloat)ctx->width/(GLfloat)ctx->height;
	glOrtho(-aspect * 100,aspect * 100,-100,100,-1,1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glClear(GL_COLOR_BUFFER_BIT);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);

	for(i=10;i>0;i--)
	{
		glColor4b(rand() %  256,rand() %  256,rand() %  256,128);
		glRectf(-8 * i,-8 * i,8*i,8*i);
	}

	glFlush();
	glFinish();
	SwapBuffers(wglGetCurrentDC());
	GdiFlush();
}
*/

LRESULT CALLBACK
refWndProc(HWND wnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	PAINTSTRUCT ps;
	ref_context_t *ctx;
	ctx=(ref_context_t*)thread_context;
	switch(msg)
	{
		case WM_ERASEBKGND:
			return 1;
		case WM_PAINT:
			BeginPaint(wnd,&ps);
			EndPaint(wnd,&ps);
			return 0;
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
		default:
			break;
	}
	if(ctx)
		return ctx->message_handler(wnd,msg,wparam,lparam);
	else
		return DefWindowProc(wnd,msg,wparam,lparam);
}

static LONG class_counter_minus_one=-1;

void *refCreateView(ref_visual_attributes_t *attributes)
{
	void *context;
	ref_context_t ctx;
	int *pfds=NULL;
	ref_visual_t visual;
	ref_visual_t *visuals=NULL;
	PIXELFORMATDESCRIPTOR pfd;
	int i, pfd_max, vis_count, best_visual;
	static char *gl_class_name="GL_ViewWindowClass";
	ref_visual_attributes_t default_attributes=REF_VISUAL_ATTRIBUTES_DEFAULT;
	if(!attributes)
		attributes=&default_attributes;
	else
	{
		/* Check attributes */
		if(attributes->flags.force_software && attributes->flags.force_hardware)
		{
			conError("refCreateView(); bad attributes, both force_software and force_hardware specified.");
			return NULL;
		}
	}
	ZeroMemory(&ctx,sizeof(ctx));
	if(InterlockedIncrement(&class_counter_minus_one)==0) /* was -1 */
	{
		ctx.wc.hInstance=GetModuleHandle(NULL);
		ctx.wc.style=CS_HREDRAW|CS_VREDRAW|CS_OWNDC|CS_NOCLOSE;
		ctx.wc.lpfnWndProc=refWndProc;
		ctx.message_handler=DefWindowProc;
		ctx.wc.hIcon=LoadIcon(NULL,IDI_WINLOGO);
		ctx.wc.hCursor=LoadCursor(NULL,IDC_ARROW);
		ctx.wc.hbrBackground=GetStockObject(HOLLOW_BRUSH);
		ctx.wc.lpszClassName=gl_class_name;
		if(!RegisterClass(&ctx.wc))
		{
			conError("RegisterClass failed.");
			return NULL;
		}
	}
	else
		while(!GetClassInfo(GetModuleHandle(NULL),gl_class_name,&ctx.wc))
			Sleep(1);
	ctx.width=GetSystemMetrics(SM_CXSCREEN);
	ctx.height=GetSystemMetrics(SM_CYSCREEN);
	ctx.wnd=CreateWindow(
		ctx.wc.lpszClassName,
		"Initializing ...",
		WS_POPUP,
		0,0,ctx.width,ctx.height,
		NULL,NULL,ctx.wc.hInstance,NULL);
	if(!ctx.wnd)
	{
		conError("refCreateView(): CreateWindow failed.");
		refDestroyView((void*)&ctx);
		return NULL;
	}
	ShowWindow(ctx.wnd,SW_SHOW);
	ctx.dc=GetDC(ctx.wnd);
	/*
	 *	Enumerate all pixel formats.
	 */
	ZeroMemory(&pfd,sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
	pfd_max=DescribePixelFormat(ctx.dc,1,sizeof(PIXELFORMATDESCRIPTOR),&pfd);
	if(pfd_max)
	{
		pfds=(int*)malloc(pfd_max*sizeof(int));
		visuals=(ref_visual_t*)malloc(pfd_max*sizeof(ref_visual_t));
	}
	vis_count=0;
	for(i=1;i<=pfd_max;i++)
	{
		if(DescribePixelFormat(ctx.dc,i,sizeof(PIXELFORMATDESCRIPTOR),&pfd))
		{
			if(ref_visual_from_gdi_pfd(&visual,&pfd) &&
				ref_accept_visual(&visual,attributes))
			{
				CopyMemory(&visuals[vis_count],&visual,sizeof(visual));
				pfds[vis_count]=i;
				vis_count++;
			}
		}
	}
	if(!vis_count)
	{
		conError("refCreateView(): no appropriate pixel format.\n");
		refDestroyView((void*)&ctx);
		return NULL;
	}
	best_visual=ref_best_visual(visuals,vis_count);
	if(SetPixelFormat(ctx.dc,pfds[best_visual],NULL)==FALSE)
	{
		conError("refCreateView(): SetPixelFormat() failed.\n");
		refDestroyView((void*)&ctx);
		return NULL;
	}
#ifdef __BORLANDC__
	_control87(MCW_EM,MCW_EM);
#endif
	ctx.rc=wglCreateContext(ctx.dc);
	if(!ctx.rc)
	{
		conError("refCreateView(): wglCreateContext() failed.\n");
		refDestroyView((void*)&ctx);
		return NULL;
	}
	if(wglMakeCurrent(ctx.dc,ctx.rc)==FALSE)
	{
		conError("refCreateView(): wglMakeCurrent() failed.\n");
		refDestroyView((void*)&ctx);
		return NULL;
	}
	CopyMemory(&ctx.visual,&visuals[best_visual],sizeof(ref_visual_t));
	context=malloc(sizeof(ref_context_t));
	CopyMemory(context,&ctx,sizeof(ref_context_t));
	
	thread_context=context;
	
	return context;
}

void
refDestroyView(void *context)
{
	ref_context_t *ctx=(ref_context_t*)context;
	if(!ctx)
		return;
	ctx->message_handler=DefWindowProc;
	if(ctx->rc)
	{
		wglMakeCurrent(NULL,NULL);
		wglDeleteContext(ctx->rc);
	}
	if(ctx->wnd)
	{
		DestroyWindow(ctx->wnd);
		if(InterlockedDecrement(&class_counter_minus_one)<0)
			UnregisterClass(ctx->wc.lpszClassName,ctx->wc.hInstance);
	}
	free(ctx);
	return;
}

void refSetWindowPlacement(int left,int top,int width,int height)
{
	if(left>0)
		refSetWMProperty(ref_wm_left,&left);
	if(top>0)
		refSetWMProperty(ref_wm_top,&top);
	if(width>0)
		refSetWMProperty(ref_wm_width,&width);
	if(height>0)
		refSetWMProperty(ref_wm_height,&height);
}

int main()	
{
	int i;
	GLfloat aspect;
	void *context;
	ref_context_t *ctx;
	ref_visual_attributes_t attr_list=REF_VISUAL_ATTRIBUTES_DEFAULT;
	
	context=refCreateView(&attr_list);
	if(!context)
		return -1;
//	refSetWindowPlacement(20,20,256,256);
	
	ctx=(ref_context_t*)context;

	glDisable(GL_CULL_FACE);
	
	for(;;) {
		MSG msg;
		while(PeekMessage(&msg,ctx->wnd,0,0,PM_REMOVE))
		{
			if(msg.wParam==VK_ESCAPE)
				goto done;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		redraw();
		Sleep(10);
	}
done:;
	refDestroyView(context);
	return 1;
}
