#include <v_lib.h>
#include <../source/vfs/iobase.h>

#define	MP3_IN_BUF_SIZE		16384
#define	MP3_OUT_BUF_SIZE	8192

typedef struct mp3_decoder_s
{
	ioport_base_t	vtbl;
	file_t			in_stream;
	struct			mpstr mp;
	u8_fifo_t		decompressed;
	char			in_buf[MP3_IN_BUF_SIZE];
	char			out_buf[MP3_OUT_BUF_SIZE];
}	mp3_decoder_t, *mp3_decoder_p;

V_LOCAL	u32_t
mp3_read(file_t handle,void *buf,u32_t len)
{
	int size, ret;
	u32_t read_in;
	mp3_decoder_t *dec=(mp3_decoder_t *)handle;
	/* while not enought data in 'decompressed', decompress */
	while(u8fCount(&dec->decompressed)<len)
	{
		read_in=fileRead(dec->in_stream,in_buf,MP3_IN_BUF_SIZE);
		if(!read_in)	/* EOF */
			break;
		/* decompress input buffer */
		ret = mp3Decode(&dec->mp,dec->in_buf,read_in,dec->out_buf,
			MP3_OUT_BUF_SIZE,&size);
		while(ret == MP3_OK)
		{
			u8fPushBuffer(&dec->decompressed,dec->out_buf,(u32_t)size);
			ret = mp3Decode(&dec->mp,NULL,0,dec->out_buf,
				MP3_OUT_BUF_SIZE,&size);
		}
	}
	if(len>u8fCount(&dec->decompressed))
		len=u8fCount(&dec->decompressed);
	u8fPopBuffer(&dec->decompressed,buf,len);
	return len;
}

V_LOCAL void
mp3_close(file_t handle)
{
	fileClose(((mp3_decoder_t*)handle)->in_stream);
	memFree(handle);
}

V_LOCAL mp3_decoder_p
mp3AllocDecoder(file_t in_stream)
{
	mp3_decoder_t *dec;
	dec=ioportAlloc(sizeof(mp3_decoder_t));
	mp3Init(&dec->mp);
	u8fInit(&dec->decompressed);
	dec->in_stream=in_stream;
	return dec;
}	

file_t V_GLOBAL
mp3Open(char *name)
{
	mp3_decoder_t *dec;
	file_t in;
	in=fileOpenRead(name);
	if(!in)
		return NULL;
	dec=mp3AllocDecoder(in);
	dec->vtbl.read=mp3_read;
	dec->vtbl.cloae=mp3_close;
	return (file_t)dec;
}

#if 0

char buf[16384];
struct mpstr mp;

int	main(void)
{
	int size;
	char out[8192];
	int len,ret;
	mp3Init(&mp);
	for(;;)
	{
		len = read(0,buf,16384);
		if(len <= 0)
			break;
		ret = mp3Decode(&mp,buf,len,out,8192,&size);
		while(ret == MP3_OK)
		{
			write(1,out,size);
			ret = mp3Decode(&mp,NULL,0,out,8192,&size);
		}
	}
	return 0;
}

#endif

