#define         MAX_OFFSCREEN_SURFACE 64
#define         MAX_SCREEN_MODES      256

typedef struct
{
  DWORD cx, cy, bpp;
} DDMODEINFO;

class           C2DEngine
#ifdef          MEM_OVERRIDE
  : public MemOverride, public CFormattedMessage
#else
  : public CFormattedMessage
#endif
{
protected:
  HWND                  hWnd;
  IDirectDraw           *lpDD;
  IDirectDrawPalette    *lpDDP;
  IDirectDrawSurface    *lpDDSPrimary, *lpDDSBackBuffer;
  IDirectDrawClipper    *lpDDC;
  DDCAPS                HELCaps, HALCaps;
  struct
  {
    DWORD               Width, Height, Caps;
  } OSD[MAX_OFFSCREEN_SURFACE+1];
  IDirectDrawSurface    *OffScreenSurface[MAX_OFFSCREEN_SURFACE+1];
  char                  *AttachedBitmapId[MAX_OFFSCREEN_SURFACE+1];
  BYTE                  IsColorKeySet[MAX_OFFSCREEN_SURFACE+1];
  int                   iNumModes;
  DDMODEINFO            aModeInfo[MAX_SCREEN_MODES];
  // bPrimaryComplex==TRUE    
  BOOL                  bPrimaryComplex;
  // bNLVEnabled==TRUE -     (AGP) 
  BOOL                  bNLVEnabled;
  // bPageFlip==FALSE -     'Blt'
  BOOL                  bPageFlip;
  // bModeSet==FALSE -    ,   
  BOOL                  bModeSet;
  void                  ResetAll();
  void                  ReleaseAllSurfaces();
  void                  Create();
  void                  Destroy();
  DWORD                 GetAvailableMem(DWORD dwSurfaceCaps);
  //    ,  'DisplayString'   
  virtual void          DisplayString(char *lpszString)
  {
#ifdef DEBUG
    if(lpDD) lpDD->FlipToGDISurface();
    ::MessageBox(GetFocus(),lpszString,"C2DEngine",0);
#endif
  }
public:
  int                   LastErrorCode;
  BOOL                  LastError() {return LastErrorCode!=DD_OK;}
                        C2DEngine(HWND hWnd);
  virtual               ~C2DEngine();
  // =====================================================================
  //        
  HWND                  GetWnd() {return hWnd;}
  IDirectDraw           *GetDD() {return lpDD;}
  IDirectDrawSurface    *GetPrimarySurface() {return lpDDSPrimary;}
  IDirectDrawSurface    *GetBackBuffer() {return lpDDSBackBuffer;}
  IDirectDrawSurface    *GetOffScreenSurface(int iSurfNum)
  {
    return OffScreenSurface[iSurfNum];
  }
  IDirectDrawPalette    *GetDDPalette() {return lpDDP;}
  IDirectDrawClipper    *GetClipper() {return lpDDC;}
  //   
  int                   GetNumModes() {return iNumModes;}
  int                   GetModeWidth(int iM) {return aModeInfo[iM].cx;}
  int                   GetModeHeight(int iM) {return aModeInfo[iM].cy;}
  int                   GetModeBpp(int iM) {return aModeInfo[iM].bpp;}
  DDCAPS                *GetHALCaps() {return &HALCaps;}
  DDCAPS                *GetHELCaps() {return &HELCaps;}
  BOOL                  IsNLVAvailable(void);
  BOOL                  IsNLVEnabled(void);
  //      
  DWORD                 GetAvailableLocalVidMem(void);
  DWORD                 GetAvailableNonlocalVidMem(void);
  // =====================================================================
  //   -    ..
  void                  SetFullScreenMode(int cx, int cy, int bpp,
    BOOL bForceSystemMemory=FALSE);
  void                  SetPalette(PALETTEENTRY *PalEntries);
  void                  SetDIBPalette(char *lpszId);
  void                  EnableNLV(BOOL bEnable);
  void                  RestoreSurfaces();
  //   IDirectDrawSurface::GetDC
  HDC                   GetPrimaryDC();
  void                  ReleasePrimaryDC(HDC dc);
  void                  ClearPrimarySurface(DWORD dwColor);
  //   IDirectDrawSurface::ReleaseDC
  HDC                   GetBackBufferDC();
  void                  ReleaseBackBufferDC(HDC dc);
  void                  ClearBackBuffer(DWORD dwColor);
  //   
  void                  Flip();
  void                  SetFlipMode(BOOL bPageFlip);
  //     
  int                   CreateOffScreenSurface(int width,int height);
  BOOL                  IsValidSurface(int iSurfNum);
  HDC                   GetOffScreenDC(int iSurfNum);
  void                  ReleaseOffScreenDC(int iSurfNum, HDC dc);
  void                  ClearSurface(int iSurfNum, DWORD dwColor);
  //      ; 
  //  'lpszId' - .   'CDib'
  int                   AttachBitmap(char *lpszId);
  //    
  void                  BlitBitmap( int iDestSurf, RECT *lprcDest,
                                    RECT *lprcSrc, void *lpBits,
                                    BITMAPINFO *lpBitsInfo, UINT uUsage,
                                    DWORD dwRop);
  //    
  virtual void          RepaintOffScreenSurface(int iSurfNum);
  //     
  void                  SetColorKey(int iSurfNum, int lk, int hk);
  //      
  void                  Blit(IDirectDrawSurface *lpDDSDest,
                             RECT *lprcDest,
                             IDirectDrawSurface *lpDDSSrc,
                             RECT *lprcSrc);
  //     
  void                  Blit(RECT *lprcDest,int iDDSSrc, RECT *lprcSrc);
  // ,    
  DWORD                 GetOSWidth(int iSurfNum)
  {
    return OSD[iSurfNum].Width;
  }
  DWORD                 GetOSHeight(int iSurfNum)
  {
    return OSD[iSurfNum].Height;
  }
  DWORD                 GetOSCaps(int iSurfNum)
  {
    return OSD[iSurfNum].Caps;
  }
  void                  GetOSD(int iSurfNum, DWORD *lpdwWidth,
    DWORD *lpdwHeight, DWORD *lpdwCaps);
  //    
  void                  ReleaseOffScreenSurface(int iSurfNum);
};

// =======================================================================
// CWave     WAV -      
// =======================================================================
//  :
// =======================================================================
// :   
// =======================================================================
// BOOL CWave::LastError();
//  TRUE       FALSE .
// =======================================================================
// void CWave::LoadHeader(char *lpszId);
//    WAV-      .
// :
// lpszId
//         ( -
//    "RT_RCDATA")    .
//       LastError();
// =======================================================================
// WAVEFORMATEX CWave::GetWaveFormatEx();
//    WAVEFORMATEX   .
// =======================================================================
// BYTE *CWave::GetWaveBits();
//     (      
//  ).
// =======================================================================
// DWORD CWave::GetWaveSize();
//     .
// =======================================================================
// DWORD CWave::GetWaveChannels();
//     .
// =======================================================================
// DWORD CWave::GetWaveFrequency();
//     .
// =======================================================================
// DWORD CWave::GetWaveBitsPerSample();
//        .
// =======================================================================
// void CWave::LoadBits(void *lpExternalBuffer=NULL);
//   .
// :
//   lpExternalBuffer -  ,      -
//  .    ,    
//             
//      .
// =======================================================================

//      ,    
#define CWF_HEADER_LOADED               1
//      ,      
#define CWF_LOCATED_IN_RESOURCE         2
//      ,        
#define CWF_PRIVATE_BUFFER_ALLOCATED    4

class CWave
#ifdef          MEM_OVERRIDE
  : public MemOverride
#endif
{
protected:
  BOOL                  bLastError;
  DWORD                 dwFlags;
  DWORD                 dwSize;
  WAVEFORMATEX          WFE;
  DWORD                 dwDataBlockOffset;
  BYTE                  *lpWaveBits;
  char                  *lpszId;
  void                  ResetAll();
  void                  ReleaseAll();
public:
  BOOL                  LastError() {return bLastError;}
  WAVEFORMATEX          *GetWaveFormatEx() {return &WFE;}
  BYTE                  *GetWaveBits() {return lpWaveBits;}
  DWORD                 GetWaveSize() {return dwSize;}
  DWORD                 GetWaveChannels() {return (DWORD)WFE.nChannels;}
  DWORD                 GetWaveFrequency() {return WFE.nSamplesPerSec;}
  DWORD                 GetWaveBitsPerSample()
                        {
                          return (DWORD)WFE.wBitsPerSample;
                        }
                        CWave();
                        ~CWave();
  void                  LoadHeader(char *lpszId);
  void                  LoadBits(void *lpExternalBuffer=NULL);
  void                  Load(char *lpszId);
};

// =======================================================================
//  'CWavePlayer'     -
//  .
// =======================================================================
//  :
// ) CWavePlayer::Prepare(CWave *wvSoundData);
// ) CWavePlayer::Play();
// ) while(CWavePlayer::GetProgress1000()<1000);
// ) CWavePlayer::Close();
// =======================================================================
//    CWavePlayer::iPlayerState
// =======================================================================
//  
#define         WPL_CLOSED              0
//       
#define         WPL_OPENED              1
//       
#define         WPL_PREPARED            2
//  
#define         WPL_PLAYING             3
//   
#define         WPL_PAUSED              4
// =======================================================================
class           CWavePlayer
#ifdef          MEM_OVERRIDE
  : public MemOverride
#endif
{
  int           iNumDevs;
  HWAVEOUT      hSoundDevice;
  WAVEHDR       wh;
  HWND          hWnd;
  CWave         *lpWave;
  DWORD         dwSamplesTotal;
  BOOL          bLastError;
  int           iPlayerState;
  void          ResetAll();
  void          ReleaseAll();
public:
                CWavePlayer(HWND hWnd=NULL);
  virtual       ~CWavePlayer();
  BOOL          LastError() {return bLastError;}
  void          Open(CWave *lpWave);
  void          Prepare(CWave *lpWave, BOOL bLooped=FALSE);
  void          Play();
  void          Stop();
  void          Pause();
  void          Resume();
  void          SetVolume(DWORD dwLeft, DWORD dwRight);
  void          Close();
  BOOL          IsPlaying();
  DWORD         GetProgress1000();
};

#endif
