/* $Id: osdemo.c,v 1.1.1.1 1999/08/19 00:55:40 jtg Exp $ */

/*
 * Demo of off-screen Mesa rendering
 *
 * See Mesa/include/GL/osmesa.h for documentation of the OSMesa functions.
 *
 * If you want to render BIG images you'll probably have to increase
 * MAX_WIDTH and MAX_HEIGHT in src/config.h.
 *
 * This program is in the public domain.
 *
 * Brian Paul
 *
 * PPM output provided by Joerg Schmalzl.
 * ASCII PPM output added by Brian Paul.
 */


/*
 * $Log: osdemo.c,v $
 * Revision 1.1.1.1  1999/08/19 00:55:40  jtg
 * Imported sources
 *
 * Revision 3.0  1998/02/14 18:42:29  brianp
 * initial rev
 *
 */



#include <stdio.h>
#include <stdlib.h>
#include "GL/osmesa.h"
#include "GL/glut.h"


#define	SGI_GL_H
#include <v_lib.h>

/*
**	Loads image and binds it to new texture object with given parameters.
*/
GLuint	V_GLOBAL
vglLoadTexture(char *file_name,
	u8_p (V_CALL *LoadProc)(char *name,u32_t *width,u32_t *height),
	GLenum filter,GLenum clamp,GLenum env)
{
	u32_t width,height;
	GLuint id;
	u8_t *texels;
	GLenum magf=filter;
	int level;
	if(!file_name || !file_name[0] || !LoadProc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglLoadTexture('%s',%X): incorrect parameter(s).\n",
			dbs(file_name),LoadProc);
#endif	/* CHECKED_BUILD */
		return 0;
	}
	texels=LoadProc(file_name,&width,&height);
	if(!texels)
		return 0;
	if(!IS_POWER_TWO(width) || !IS_POWER_TWO(height))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglLoadTexture(): wrong texture dimensions %dX%d.\n",
			width,height);
#endif	/* CHECKED_BUILD */
		memFree(texels);
		return 0;
	}
	switch(filter)
	{
		case GL_LINEAR_MIPMAP_NEAREST:
		case GL_LINEAR_MIPMAP_LINEAR:
		case GL_LINEAR:
			magf=GL_LINEAR;
			break;
		case GL_NEAREST_MIPMAP_LINEAR:
		case GL_NEAREST_MIPMAP_NEAREST:
		case GL_NEAREST:
			magf=GL_NEAREST;
			break;
		default:
			magf=filter;
			break;
	}
	/* Generate texture and bind it to target */
	glGenTextures(1,&id);
	glBindTexture(GL_TEXTURE_2D,id);
	/* Download image data to renderer */
	level=0;
	if(magf!=filter)	/* generate all except last mipmap levels */
		for(;width>1 && height>1;level++)
		{
			glTexImage2D(GL_TEXTURE_2D,level,GL_RGBA,width,height,0,GL_RGBA,
				GL_UNSIGNED_BYTE,texels);
			texels=(u8_p)imgDownsample(texels,width,height,false);
			width>>=1;
			height>>=1;
		}
	glTexImage2D(GL_TEXTURE_2D,level,GL_RGBA,width,height,0,GL_RGBA,
		GL_UNSIGNED_BYTE,texels);
	memFree(texels);
	/* Set texture parameters */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,filter);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,magf);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,clamp);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,clamp);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,env);
	/* Clear current texture object */
	glBindTexture(GL_TEXTURE_2D,0);
	return id;
}


#define WIDTH 400
#define HEIGHT 300

int tex_obj;

static void render_image( void )
{
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(-60,60,-60,60,-60,60);
   glMatrixMode(GL_MODELVIEW);
   glClearColor(0,.5f,.5f,.5f);
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glBegin(GL_QUADS);
	   	glColor3f(1,1,1);
		glVertex2f(-50,-50);
		glColor3f(1,0,0);
		glVertex2f(-50,50);
	   	glColor3f(0,1,0);
		glVertex2f(50,-50);
		glColor3f(0,0,1);
		glVertex2f(50,50);
	glEnd();
	
	glScalef(1.1f,1.1f,1.1f);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glBindTexture(GL_TEXTURE_2D,tex_obj);
	glEnable(GL_TEXTURE_2D);
	glBegin(GL_QUADS);
	   	glColor4f(1,1,1,.5f);
		glTexCoord2f(0,0);
		glVertex2f(-50,-50);
		glColor4f(1,0,0,.5f);
		glTexCoord2f(0,1);
		glVertex2f(-50,50);
	   	glColor4f(0,1,0,.5f);
		glTexCoord2f(1,1);
		glVertex2f(50,-50);
		glColor4f(0,0,1,.5f);
		glTexCoord2f(1,0);
		glVertex2f(50,50);
	glEnd();
	glDisable(GL_BLEND);
	glBindTexture(GL_TEXTURE_2D,0);
	glDisable(GL_TEXTURE_2D);
	
	glFlush();
	glFinish();
}



int main( int argc, char *argv[] )
{
   OSMesaContext ctx;
   void *buffer;
   
   /* Create an RGBA-mode context */
   ctx = OSMesaCreateContext( GL_RGBA, NULL );

   /* Allocate the image buffer */
   buffer = malloc( WIDTH * HEIGHT * 4 );

   /* Bind the buffer to the context and make it current */
   OSMesaMakeCurrent( ctx, buffer, GL_UNSIGNED_BYTE, WIDTH, HEIGHT );

	do {
		timer_t timer;
		float dt;
		int i,nf=10;
		logStartConsoleLog();
		logUpperLevel(LOG_TRACE);
		tex_obj=vglLoadTexture("../../samples/media/mist.jpg",jpgLoad,
			GL_LINEAR,GL_REPEAT,GL_MODULATE);
		if(!tex_obj)
		{
			puts("Failed to load texture");
			return 0;
		}
		printf("Rendering %d frames ...\n",nf);
		fflush(stdout);
		timerStart(&timer);
		for(i=0;i<nf;i++)
			render_image();
		dt = 0.001 * timerMilliseconds(&timer);
		printf("Done: %g seconds, %g fps\n",dt,nf/dt);
		fflush(stdout);
	} while(0);
	
   if (argc>1) {
      /* write PPM file */
      FILE *f = fopen( argv[1], "w" );
      if (f) {
         int i, x, y;
         GLubyte *ptr = (GLubyte *) buffer;
#define BINARY 0
#if BINARY
         fprintf(f,"P6\n");
         fprintf(f,"# ppm-file created by %s\n",  argv[0]);
         fprintf(f,"%i %i\n", WIDTH,HEIGHT);
         fprintf(f,"255\n");
         fclose(f);
         f = fopen( argv[1], "ab" );  /* reopen in binary append mode */
         for (y=HEIGHT-1; y>=0; y--) {
            for (x=0; x<WIDTH; x++) {
               i = (y*WIDTH + x) * 4;
               fputc(ptr[i], f);   /* write red */
               fputc(ptr[i+1], f); /* write green */
               fputc(ptr[i+2], f); /* write blue */
            }
         }
#else /*ASCII*/
         int counter = 0;
         fprintf(f,"P3\n");
         fprintf(f,"# ascii ppm file created by %s\n", argv[0]);
         fprintf(f,"%i %i\n", WIDTH, HEIGHT);
         fprintf(f,"255\n");
         for (y=HEIGHT-1; y>=0; y--) {
            for (x=0; x<WIDTH; x++) {
               i = (y*WIDTH + x) * 4;
               fprintf(f, " %3d %3d %3d", ptr[i], ptr[i+1], ptr[i+2]);
               counter++;
               if (counter % 5 == 0)
                  fprintf(f, "\n");
            }
         }
#endif
         fclose(f);
      }
   }
   else {
      printf("Specify a filename if you want to make a ppm file\n");
   }

   printf("all done\n");

   /* free the image buffer */
   free( buffer );

   /* destroy the context */
   OSMesaDestroyContext( ctx );

   return 0;
}
