#include  "..\include\v_lib.h"
#include  "..\source\v_mt.h"
#pragma   hdrstop

void V_GLOBAL graphSetHandler(gc_t c,event_type_t evtype,void *handler)
{
  if(evtype<(int)evtype_idle || evtype>=(int)evtype_maxevent) {
    conlPrintf(LOG_WARNING,"graphSetHandler(): unknown event type '%d'.\n",
      (int)evtype);
    return;
  }
  switch(evtype) {
    case evtype_idle:
      c->on_idle=(notifyevent_t)handler;
      break;
    case evtype_activate:
      c->on_activate=(notifyevent_t)handler;
      break;
    case evtype_deactivate:
      c->on_deactivate=(notifyevent_t)handler;
      break;
    case evtype_destroy:
      c->on_destroy=(notifyevent_t)handler;
      break;
    case evtype_keydown:
      c->on_keydown=(keyevent_t)handler;
      break;
    case evtype_keyup:
      c->on_keyup=(keyevent_t)handler;
      break;
    case evtype_mousedown:
      c->on_mousedown=(mouseevent_t)handler;
      break;
    case evtype_mouseup:
      c->on_mouseup=(mouseevent_t)handler;
      break;
    case evtype_mousemove:
      c->on_mousemove=(mouseevent_t)handler;
      break;
    case evtype_sysevent:
      c->sysevent=(sysevent_t)handler;
      break;
    default:
      conlPrintf(LOG_WARNING,"graphSetHandler(event %d): not set.\n",
        (int)evtype);
  }
}

void V_GLOBAL graphResetHandler(gc_t c,event_type_t evtype)
{
  if(evtype<(int)evtype_idle || evtype>=(int)evtype_maxevent) {
    conlPrintf(LOG_WARNING,"graphSetHandler(): unknown event type '%d'.\n",
      (int)evtype);
    return;
  }
  switch(evtype) {
    case evtype_idle:
      c->on_idle=NULL;
      break;
    case evtype_activate:
      c->on_activate=NULL;
      break;
    case evtype_deactivate:
      c->on_deactivate=NULL;
      break;
    case evtype_destroy:
      c->on_destroy=NULL;
      break;
    case evtype_keydown:
      c->on_keydown=NULL;
      break;
    case evtype_keyup:
      c->on_keyup=NULL;
      break;
    case evtype_mousedown:
      c->on_mousedown=NULL;
      break;
    case evtype_mouseup:
      c->on_mouseup=NULL;
      break;
    case evtype_mousemove:
      c->on_mousemove=NULL;
      break;
    case evtype_sysevent:
      c->sysevent=NULL;
      break;
    default:
      conlPrintf(LOG_WARNING,"graphResetHandler(event %d): not removed.\n",
        (int)evtype);
  }
}

void    V_GLOBAL graphMakeCurrent(gc_t context)
{
  GetThreadData()->_gc_=context;
}

gc_t    V_GLOBAL graphGetCurrentContext(void)
{
  return GetThreadData()->_gc_;
}

int       V_GLOBAL getmaxx()
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  return _gc_->sizex;
}

int       V_GLOBAL getmaxy()
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  return _gc_->sizey;
}

u_32      V_GLOBAL getcaps()
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  return _gc_->caps;
}

void    V_GLOBAL setcolor(int r,int g,int b)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.setcolor(_gc_,r,g,b);
}

void    V_GLOBAL setindex(int index)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.setindex(_gc_,index);
}

void    V_GLOBAL setbkcolor(int r,int g,int b)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.setbkcolor(_gc_,r,g,b);
}

void    V_GLOBAL setbkindex(int index)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.setbkindex(_gc_,index);
}

void    V_GLOBAL settextcolor(int r,int g,int b)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.settextcolor(_gc_,r,g,b);
}

void    V_GLOBAL settextindex(int index)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.settextindex(_gc_,index);
}

void    V_GLOBAL moveto(int x,int y)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.moveto(_gc_,x,y);
}

void    V_GLOBAL lineto(int x,int y)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.lineto(_gc_,x,y);
}
  
void    V_GLOBAL line(int x1,int y1,int x2,int y2)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.line(_gc_,x1,y1,x2,y2);
}

void    V_GLOBAL rectangle(int x1,int y1,int x2,int y2)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.rectangle(_gc_,x1,y1,x2,y2);
}

void    V_GLOBAL bar(int x1,int y1,int x2,int y2)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.bar(_gc_,x1,y1,x2,y2);
}
  
void    V_GLOBAL clearviewport(void)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.clearviewport(_gc_);
}

void    V_GLOBAL circle(int x,int y,int radius)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.circle(_gc_,x,y,radius);
}

void    V_GLOBAL ellipse(int x1,int y1,int x2,int y2)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.ellipse(_gc_,x1,y1,x2,y2);
}

void    V_GLOBAL triangle(int x1,int y1,int x2,int y2,int x3,int y3)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.triangle(_gc_,x1,y1,x2,y2,x3,y3);
}

void V_GLOBAL usefont(int sizex,int sizey,vec_t direction,u_32 effects,
                     char *name)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.usefont(_gc_,sizex,sizey,direction,effects,name);
}

void    V_GLOBAL getextent(char *str,int *sx,int *sy)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.getextent(_gc_,str,sx,sy);
}

void    V_GLOBAL outtext(char *str)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.outtext(_gc_,str);
}

void    V_GLOBAL outtextxy(int x,int y,char *str)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r2d.outtextxy(_gc_,x,y,str);
}

vboolean V_GLOBAL loadpalette(char *unix_resname)
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  return _gc_->r2d.loadpalette(_gc_,unix_resname);
}

/* TODO: implement 3D rendering code                                    */
/*

void  V_GLOBAL glEnable(...
{
  gc_t _gc_=GetThreadData()->_gc_;
  V_ASSERT(_gc_);
  _gc_->r3d.gl_enable(_gc_,...
}

*/

