#include <v_lib.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define	PEAK		0x7fff
#define	OFFSET		0x0
#define	MAXOUTBURST	(16 * 1024)
u32_t	outburst	= MAXOUTBURST;
#define	SAMPLE_RATE (11025 * 4)

#define	FMT			i16_t

V_LOCAL	FMT			buf[MAXOUTBURST];
V_LOCAL	FMT			large_buf[256 * 1024];

V_LOCAL	void
generate_sines(FMT *buffer,int samples,u32_t channels,vec_t freqs[],vec_t volumes[],vec_t *phase_info)
{
	u32_t s,c;
	vec_t out;
	vec_t *phase=(vec_t*)alloca(channels * sizeof(vec_t));
	vec_t *dph=(vec_t*)alloca(channels * sizeof(vec_t));
	vec_t *amp=(vec_t*)alloca(channels * sizeof(vec_t));
	if(phase_info)
		memcpy(phase,phase_info,channels * sizeof(vec_t));
	else
		memset(phase,0,channels * sizeof(vec_t));
	for(c=0;c<channels;c++)
	{
		amp[c] = PEAK * volumes[c] / (vec_t)channels;
		dph[c] = M_2PI * freqs[c] / (vec_t)SAMPLE_RATE;
	}
	for(s=0;s<samples;s++)
	{
		out = 0;
		for(c=0;c<channels;c++)
		{
			out += amp[c] * sin(phase[c]);
			phase[c]+=dph[c];
		}
		if(out < -(vec_t)PEAK)
			out = -(vec_t)PEAK;
		else
			if(out > (vec_t)PEAK)
				out = (vec_t)PEAK;
		buffer[s]=OFFSET + (FMT)out;
	}
	if(phase_info)
		memcpy(phase_info,phase,channels * sizeof(vec_t));
}

cond_t cond;

V_LOCAL	void
quiet_buffer()
{
	int i;
	for(i=0;i<outburst;i++)
		buf[i]=OFFSET;
}

void V_CALL outproc(void *arg)
{
	int fd=*((int*)arg);
	vec_t freqs[]={10000,1000,200};
	vec_t volumes[]={.4,.1,.1};
	vec_t phases[]={0,0,0};
	quiet_buffer();
	generate_sines(large_buf,256*1024,1,freqs,volumes,phases);
	while(!condTimedWait(cond,0))
	{
		generate_sines(buf,outburst,3,freqs,volumes,phases);
		printf("Phase %g,%g,%g\n",phases[0],phases[1],phases[2]);
		write(fd,buf,outburst * sizeof(FMT));
	}
}

int main()
{
	int fd;
	u32_t arg;
	thread_t out;
	fd=open("/dev/dsp",O_WRONLY | O_SYNC,00200);
	if(!fd)
	{
		perror("open()");
		return -1;
	}
	ioctl(fd,SNDCTL_DSP_GETBLKSIZE,&outburst);
	if(outburst > MAXOUTBURST)
    	outburst = MAXOUTBURST;
	printf("DMA buffer size: %u\n",outburst);
	arg=AFMT_S16_LE;
	// arg=AFMT_U8;
	ioctl(fd,SNDCTL_DSP_SETFMT,&arg);
	arg=SAMPLE_RATE;
	ioctl(fd,SNDCTL_DSP_SPEED,&arg);
	cond=condCreate(false);
	out=threadCreate(outproc,&fd);
	fgetc(stdin);
	condSet(cond);
	threadJoin(out,NULL);
	ioctl(fd,SNDCTL_DSP_RESET);
	condDestroy(cond);
	close(fd);

	return 0;
}
