#include <windows.h>
#include <stdio.h>
#include "../../include/v_lib.h"
#include "../../source/v_mt.h"

V_LOCAL void_p	V_CALL
libvIsHandle(void *handle,char *caller,char *object,u32_t magic)
{
	if(!handle)
	{
		conlPrintf(LOG_ERROR,"%s(): NULL descriptor.\n",caller);
		return NULL;
	}
	if(*((u32_t*)handle) != magic)
	{
		conlPrintf(LOG_ERROR,"%s(): %s damaged: bad magic.\n",caller,
			object ? object : "object");
		return NULL;
	}
	return handle;
}

typedef void	*sound_engine_t;

#define	INVALID_INDEX			0xFFFFFFFF
#define	INVALID_BUFFER_VALUE	INVALID_INDEX

/*
 *	Generic implemenation
 */


typedef struct	wave_sound_info_s
{
	char		wave_name[_MAX_PATH];	/* variable sized */
}	_sound_info_t, *_sound_info_p;

typedef struct	_sound_engine_s
{
#ifdef	CHECKED_BUILD
	u32_t				magic;
#endif	/* CHECKED_BUILD */
}	ds_sound_engine_t, *ds_sound_engine_p;

#ifdef	THIS
#undef	THIS
#endif
#define	THIS(handle)	((_sound_engine_t*)(handle))

#ifdef	CHECKED_BUILD
V_LOCAL	char		*se_object="_Sound engine";
V_LOCAL	u32_t		se_magic=0x00534E44;	/* _SND */
#endif	/* CHECKED_BUILD */

void	V_GLOBAL
seDestroyEngine(sound_engine_t engine)
{
	int i,n;
	_sound_info_t **sounds;
	if(!libvIsHandle(engine,"seDestroyEngine",se_object,se_magic))
		return;
	/* Release all buffers */
	sounds=(_sound_info_t**)vpsGetStorage(&THIS(engine)->sounds);
	n=vpsCount(&THIS(engine)->sounds);
	for(i=0;i<n;i++)
	{
		if(sounds[i])
		{
			// release sound
			libvFree(sounds[i]);
		}
	}
	vpsClear(&THIS(engine)->sounds);
	libvFree(engine);
}

/*
 */
sound_engine_t	V_GLOBAL
seCreateEngine(void)
{
	_sound_engine_t *se;
	ALLOC_OBJECT(se,_sound_engine_t,0,se_magic);
	vpsInit(&se->sounds);
	return (sound_engine_t)se;
}

/*
**	Returns sound ID or INVALID_BUFFER_VALUE on error
*/
u32_t	V_GLOBAL
seLoadSound(sound_engine_t engine,char *wave_name)
{
	wave_handle_t wave;
	WAVEFORMATEX wf;
	int name_len;
	_sound_info_t *dsi;
	u32_t buffer_num;
	if(!libvIsHandle(engine,"seLoadSound",se_object,se_magic))
		return INVALID_BUFFER_VALUE;
	wave=wavLoad(wave_name);
	if(!wave)
		return INVALID_BUFFER_VALUE;
	memZero(&wf,sizeof(WAVEFORMATEX));
	wf.wFormatTag=WAVE_FORMAT_PCM;
	wf.nChannels=(SHORT)wavNumChannels(wave);
	wf.nSamplesPerSec=wavSamplingFrequency(wave);
	wf.wBitsPerSample=(SHORT)wavBitsPerSample(wave);
	wf.nBlockAlign=(SHORT)((wf.wBitsPerSample>>3) * wavNumChannels(wave));
	wf.nAvgBytesPerSec=wf.nSamplesPerSec * wf.nBlockAlign;
	/* lock buffer and copy data */
	wavClose(wave);
	return buffer_num;
}

/*
 *	Returns sound info or NULL.
 */
_sound_info_p	V_LOCAL
#ifdef	CHECKED_BUILD
get_sound_info(char *caller,sound_engine_t *engine,u32_t buffer_num)
#else	/* !CHECKED_BUILD */
#define	get_sound_info(caller,engine,buffer)	\
		get_sound_info_quiet(engine,buffer)
get_sound_info_quiet(sound_engine_t *engine,u32_t buffer_num)
#endif	/* !CHECKED_BUILD
 */
{
	u32_t len;
	_sound_info_t *info=NULL;
	if(!libvIsHandle(engine,caller,se_object,se_magic))
		return NULL;
	len=vpsCount(&THIS(engine)->sounds);
	if(buffer_num<len)
		info=(_sound_info_t*)vpsGet(&THIS(engine)->sounds,buffer_num);
	if(!info)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"%s(%d): %s buffer.\n",
			caller,buffer_num,(buffer_num<len) ? "empty" : "invalid");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	return info;
}

vboolean	V_GLOBAL
sePlaySound(sound_engine_t *engine,u32_t buffer_num)
{
	_sound_info_t *info;
	info=get_sound_info("sePlaySound",engine,buffer_num);
	if(!info)
		return false;
	return false;
}

vboolean	V_GLOBAL
seFreeSound(sound_engine_t *engine,u32_t buffer_num)
{
	_sound_info_t *info;
	info=get_sound_info("seFreeSound",engine,buffer_num);
	if(!info)
		return false;
	vpsSet(&THIS(engine)->sounds,buffer_num,NULL);
	return false;
}

/*
 *	Float attenuation in decibels; 0 mean full folume.
 */
vboolean	V_GLOBAL
seSetVolume(sound_engine_t *engine,u32_t buffer_num,u32_t attenuation)
{
	_sound_info_t *info;
	info=get_sound_info("seSetVolume",engine,buffer_num);
	if(!info)
		return false;
	return false;
}

/*
 *	Float pan in decibels; 0 mean full folume for both channels,
 *	negative mean right attenuation, positive - left attenuation.
 */
vboolean	V_GLOBAL
seSetPan(sound_engine_t *engine,u32_t buffer_num,vec_t pan)
{
	ds_sound_info_t *info;
	info=get_sound_info("seSetPan",engine,buffer_num);
	if(!info)
		return false;
	return false;
}

vboolean	V_GLOBAL
seSetRate(sound_engine_t *engine,u32_t buffer_num,u32_t rate)
{
	ds_sound_info_t *info;
	info=get_sound_info("seSetRate",engine,buffer_num);
	if(!info)
		return false;
	return false;
}

int main()
{
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	se=seCreateEngine();
	if(!se)
		return -1;
	sound=seLoadSound(se,"../../samples/media/guide.wav");
	if(sound!=INVALID_BUFFER_VALUE)
	{
		sePlaySound(se,sound);
		fgetc(stdin);
	}
	seDestroyEngine(se);
	return 0;
}


// =======================================================================
void CWavePlayer::ResetAll()
{
  iNumDevs=0;
  hSoundDevice=0;
  memset(&wh,0,sizeof(WAVEHDR));
  lpWave=NULL;
  dwSamplesTotal=0;
  bLastError=TRUE;
  iPlayerState=WPL_CLOSED;
}

// =======================================================================
void CWavePlayer::Open(CWave *lpWave)
{
  MMRESULT mmr;
  ReleaseAll();
  ResetAll();
  iNumDevs=waveOutGetNumDevs();
  if(iNumDevs==0) return;
  if(!lpWave) return;
  this->lpWave=lpWave;
  dwSamplesTotal=lpWave->GetWaveSize()/lpWave->GetWaveChannels()/
    (lpWave->GetWaveBitsPerSample()>>3);
  mmr=::waveOutOpen(&hSoundDevice,WAVE_MAPPER,lpWave->GetWaveFormatEx(),
    (DWORD)hWnd,0,(hWnd) ? CALLBACK_WINDOW : CALLBACK_NULL);
  if(mmr!=MMSYSERR_NOERROR) return;
  bLastError=FALSE;
  iPlayerState=WPL_OPENED;
  return;
}

void CWavePlayer::Prepare(CWave *lpWave, BOOL bLooped)
{
  MMRESULT mmr;
  bLastError=TRUE;
  if(iPlayerState>WPL_OPENED)
  {
    if(iPlayerState>WPL_PREPARED) ::waveOutReset(hSoundDevice);
    ::waveOutUnprepareHeader(hSoundDevice,&wh,sizeof(WAVEHDR));
  }
  else if(iPlayerState<WPL_OPENED) return;
  wh.lpData=lpWave->GetWaveBits();
  wh.dwBufferLength=lpWave->GetWaveSize();
  if(bLooped)
  {
    wh.dwFlags=WHDR_BEGINLOOP;
  }
  else
  {
    wh.dwFlags=0;
  }
  wh.dwLoops=1;
  mmr=::waveOutPrepareHeader(hSoundDevice,&wh,sizeof(WAVEHDR));
  if(mmr!=MMSYSERR_NOERROR) return;
  bLastError=FALSE;
  iPlayerState=WPL_PREPARED;
  return;
}

void CWavePlayer::Play()
{
  if(iPlayerState>=WPL_PREPARED)
  {
    ::waveOutWrite(hSoundDevice,&wh,sizeof(WAVEHDR));
    iPlayerState=WPL_PLAYING;
  }
}

void CWavePlayer::Stop()
{
  if(iPlayerState>=WPL_PLAYING)
  {
    ::waveOutReset(hSoundDevice);
    iPlayerState=WPL_PREPARED;
  }
}

void CWavePlayer::Pause()
{
  if(iPlayerState==WPL_PLAYING)
  {
    ::waveOutPause(hSoundDevice);
    iPlayerState=WPL_PAUSED;
  }
}

void CWavePlayer::Resume()
{
  if(iPlayerState==WPL_PAUSED)
  {
    ::waveOutRestart(hSoundDevice);
  }
  iPlayerState=WPL_PLAYING;
}

void CWavePlayer::Close()
{
  if(iPlayerState>=WPL_PLAYING)
  {
    ::waveOutReset(hSoundDevice);
    iPlayerState=WPL_PREPARED;
  }
  if(iPlayerState>=WPL_PREPARED)
  {
    ::waveOutUnprepareHeader(hSoundDevice,&wh,sizeof(WAVEHDR));
    iPlayerState=WPL_OPENED;
  }
  if(iPlayerState>=WPL_OPENED)
  {
    ::waveOutClose(hSoundDevice);
    iPlayerState=WPL_CLOSED;
    hSoundDevice=NULL;
  }
  return;
}

BOOL CWavePlayer::IsPlaying()
{
  return iPlayerState==WPL_PLAYING;
}

DWORD CWavePlayer::GetProgress1000()
{
  MMTIME mmt;
  if(iPlayerState>=WPL_PLAYING)
  {
    mmt.wType=TIME_SAMPLES;
    ::waveOutGetPosition(hSoundDevice,&mmt,sizeof(MMTIME));
    return (mmt.u.sample*1000)/dwSamplesTotal;
  }
  return 0;
}

void CWavePlayer::SetVolume(DWORD dwLeft, DWORD dwRight)
{
  if(iPlayerState>WPL_PREPARED)
  {
    dwLeft&=0xFFFF;
    dwRight&=0xFFFF;
    ::waveOutSetVolume(hSoundDevice,dwLeft|(dwRight<<16));
  }
}

