/*<html><body bgcolor=#B4C4B0><font color=#000000><strong><pre>*/
#ifndef V_GL_H
#define V_GL_H

/*		Pointer internal to context struct */
typedef	void		*glw_t;

/*	Six sky textures */
typedef	GLuint		quake_sky_t[6];

/*	Outline font adjustment */
typedef enum
{
	fa_left_top=0,		fa_center_top=1,	fa_right_top=2,
	fa_left_center=3,	fa_center_center=4,	fa_right_center=5,
	fa_left_bottom=6,	fa_center_bottom=7,	fa_right_bottom=8,
	fa_custom=9
}	gl_font_adjustment_t;

#define	GL_FONT_LOD_LEVELS			3

/*	Outline font drawing parameters */
typedef struct	gl_string_parameters_s
{
	vec_t 		scale;
	vec_t		fov;
	vboolean	use_fov;
	gl_font_adjustment_t	adjustment;
	vec3_t		custom_adjustment;
}	gl_string_parameters_t, gl_string_parameters_p;

/*	Outline font */
typedef struct	gl_outline_font_s
{
	u32_t		magic;
	u32_t		lod;			/* current LOD list index (NOT list base) */
	GLuint		bases[GL_FONT_LOD_LEVELS];	/* OpenGL lists for chars */
	GLfloat		line_width;		/* used for outlined fonts */
	vboolean	line_smooth;	/* used for outlined fonts */
	vboolean	polygonal;
	vec_t		extrusion;		/* used for polygonal fonts */
	glw_t		ctx_origin;		/* context font belongs to */
	u32_t		cache_index;	/* index in context's font cache */
	/*	String drawing parameters: adjustment, etc. */
	gl_string_parameters_t	sp;
	struct gl_outline_font_s *ref;
	/*			Platform-dependent part */
#ifdef	__WIN32__
	void		*gm;			/* WIN32: GLYPHMETRICSFLOAT */
	void		*hfont;			/* WIN32: font handle */
	void		*logfont;		/* WIN32: struct LOGFONT */
#endif	/* __WIN32__ */
}	gl_outline_font_t, *gl_outline_font_p;

/*
 *	TL rendering API to pass to vglDecodeFeedback()
 */
typedef	struct	gl_tl_callbacks_s
{
	/*		To draw point */
	void	(V_CALL *point)(void *dev,vec3_t,color4b_t,vec2_t);
	/*		Should draw Gouraud - shaded line */
	void	(V_CALL *line)(void *dev,vec3_t,color4b_t,vec2_t,
									 vec3_t,color4b_t,vec2_t);
	/*		To draw Gouraud - shaded polygon */
	void	(V_CALL *begin_polygon)(void *dev,int);
	void	(V_CALL *polygon_vertex)(void *dev,vec3_t,color4b_t,vec2_t);
	/*		To respond to pass-through tokens */
	void	(V_CALL	*pass_though)(void *dev,vec_t);
}	gl_tl_callbacks_t, *gl_tl_callbacks_p;

typedef	float	*GLfloat_p;

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
	/*	INI / FINI ==================================================== */
	/*	Creates OpenGL-capable overlapped (but non-user-resizeable) window. */
	glw_t	V_GLOBAL		vglCreateWindow(char *title,
    	u32_t x,u32_t y,u32_t width,u32_t height);
	/*	Creates OpenGL-capable fullscreen window. */
	glw_t	V_GLOBAL		vglCreateFullscreenWindow(char *title,
    	u32_t width,u32_t height,u32_t bpp,u32_t refresh);
	/*	Creates OpenGL-capable child window (without borders etc.) */
	glw_t	V_GLOBAL		vglCreateChildWindow(char *title,
    	u32_t x,u32_t y,u32_t width,u32_t height,void *parent);
	/*	Creates bitmap-based OpenGL-capable window. */
	glw_t	V_GLOBAL		vglCreateMappedWindow(u32_t width,u32_t height);
	/*	Destroys window. */
	void	V_GLOBAL		vglDestroyWindow(glw_t win);
	/*	Makes given window current (but NOT moves input focust to it). */
	vboolean	V_GLOBAL	vglMakeCurrent(glw_t win);
	/*	Hides current window */
	vboolean	V_GLOBAL	vglHideWindow(void);
	/*	Shows current window */
	vboolean	V_GLOBAL	vglShowWindow(void);
	/*	Returns current window. */
	glw_t	V_GLOBAL		vglGetCurrentWindow(void);
    /*	Returns true (and copies entry point) if given extension
     *	available for current renderer; otherwise returns false */
    vboolean	V_GLOBAL 	vglQueryExtension(char *extension,
    	void **entry);
	/*	Window management ============================================= */
	/*	Returns true if threre are pending events in queue of current window */
	vboolean	V_GLOBAL	vglPending(void);
	/*	Waits for input event in current window and returns it in 'event_return'. */
	vboolean	V_GLOBAL	vglNextEvent(input_record_t *event_return);
	/*	Moves input focus to current window. */
	vboolean	V_GLOBAL	vglSetFocus(void);
	/*	Resizes window. */
	vboolean	V_GLOBAL	vglResizeWindow(u32_t width,u32_t height);
	/*	Moves window. */
	vboolean	V_GLOBAL	vglMoveWindow(u32_t left,u32_t top);
	/*	Swaps buffers of current window. */
	void		V_GLOBAL	vglSwapBuffers(void);
	/*	Copies contents of framebuffer / offscreen bitmap into malloc()'ed
	 *	buffer and returns pointer to it. */
	u8_p		V_GLOBAL	vglReadPixels(u32_t *width,u32_t *height);
	/*	Font management =============================================== */
	/*	Creates outlined or polygonal font.
	 *	Input:
	 *		face_name	- font face (may be NULL);
	 *		polygonal	- if zero, creates outlined font;
	 *		extrusion	- amount of extrusion along z-axis 
	 *			(font itself lies in xOy plane);
	 *		os_font		- os-dependent handle of font (may be NULL).
	 *	Notes.
	 *		1. Win32 only: os_font must be HFONT object that was previously
	 *	created via CreateFont() / CreateFontIndirect().
	 *		2. Win32 only: if os_font is non-NULL, font_face ignored.
	 *		3. If both face_name and os_font are NULL, function creates
	 *			font with face 'Arial' (aka 'Helvetica'; TODO: implement
	 *			automatic replacement ;).
	 *	Warning: fonts created in current context automatically deleted
	 *		with context; attemping to create font before selecting
	 *		valid context is an error.
	 */
	gl_outline_font_p V_GLOBAL	vglCreateFont(char *face_name,
		vboolean polygonal,vec_t extrusion,void *os_font);
	/*	Destroys font. */
	void	V_GLOBAL	vglDeleteFont(gl_outline_font_t *fnt);
	/*	These adjusts text setings.
	 *	LOD 0 is highest, LOD 2 is lowest.
	 */
	void	V_GLOBAL	vglSetFontLOD(gl_outline_font_t *fnt,u32_t lod);
	void	V_GLOBAL	vglSetFontAdjustment(gl_outline_font_t *fnt,
		gl_font_adjustment_t adjustment);
	void	V_GLOBAL	vglSetFontCustomAdjustment(gl_outline_font_t *fnt,
		vec3_t custom_adjustment);
	void	V_GLOBAL	vglSetFontScale(gl_outline_font_t *fnt,
		vec_t scale);
	void	V_GLOBAL	vglSetFontFOV(gl_outline_font_t *fnt,vec_t fov);
	/*	Selects font in current context. Newly created font not selected
	 *	automaticaly; caller must explicitly select them.
	 *	NOTE. Fonts created in context A can be selected in context B,
	 *	even if contexts using different pixel formats (for example,
	 *	font created in HW - accelerated context can be selected into
	 *	context with software rendering), also first time such selection
	 *	takes VERY long time (about second on my Klamath-233), because
	 *	font is duplicated in target context. But subsequent selections
	 *	will be very quickly.
	 */
	vboolean	V_GLOBAL	vglSelectFont(gl_outline_font_t *fnt);
	/*	Returns current (selected) font */
	gl_outline_font_p	V_GLOBAL	vglGetCurrentFont(void);
	/*	Prints string. */
	void	V_GLOBAL	vglPuts(char *str);
	/*	No comments. */
	void	V_GLOBAL_VA	vglPrintf(const char *fmt,...);
	/*	Returns x-extent and height from the baseline of given text */
	void	V_GLOBAL	vglGetTextExtents(char *str,vec_t *width,vec_t *height);
	/*	Draws single line of text; adjusts visible font size, LOD, etc.
	 *	Parameters:
	 *		origin 	- relative to this point text will be aligned;
	 *		cam		- camera to look from (see v_camctl.h);
	 *		text	- text to draw.
	 */
	void	V_GLOBAL	vglDrawString(vec3_t origin,cam_t *cam,char *text);
    /*
     *	Puts OpenGL in feedback mode and returns pointer to
     *	malloc()'ed feedback buffer.
     *	Input:
     *		num_elements	- number of GLfloats in feedback buffer.
     */
    void_p		V_GLOBAL	vglBeginFeedback(int num_elements);
    /*
     *	Takes OpenGL out of feedback mode. If previous mode 
     *	was feedback mode, stores number of recorded GLfloats in
     *	'stored_elements' and returns true.
     */
    vboolean	V_GLOBAL	vglEndFeedback(u32_t *stored_elements);
	/*
     *	Decodes OpenGL feedback buffer.
     *	Input:
     *		back_api	- list of drawing callbacks;
     *		dev			- device handle to pass to back_api's;
     *		buffer		- filled OpenGL feedback buffer;
     *		stored_elemens - number of GLfloat's in feedback buffer.
     *	Output:
     *		None.
     */
    void	V_GLOBAL		vglDecodeFeedback(gl_tl_callbacks_t *back_api,
		void *dev,void_p buffer,int stored_elements);
    /*
     *	Creates metafile context, sets it's modelview and projection matrices
     *	idential to ones in current context, calls DrawProc(param) and closes
     *	metafile.
     *	Feedback buffer size automatically increased if needed.
     */
    vboolean	V_GLOBAL    vglDrawToMetafile(char *file_name,
    	u32_t mmWidth,u32_t mmHeight,
    	void (V_CALL *DrawProc)(void *param),void *param);
	/*
	 *	Utility to produce screenshots.
	 *	File format determines automatically by extension (bmp | tga).
	 */
	vboolean	V_GLOBAL	vglSaveFramebuffer(char *file_name,
		u32_t mmWidth,u32_t mmHeight);
	/*	Returns handle for glBindTexture() / glDeleteTextures(). */
	GLuint	V_GLOBAL		vglLoadTexture(char *file_name,
		u8_p (V_CALL *LoadProc)(char *name,u32_t *width,u32_t *height),
		GLenum filter,GLenum clamp,GLenum env);
	/*	Loads six sky textures for vglDrawSky() */
	vboolean	V_GLOBAL	vglLoadSky(quake_sky_t dest,char *path,
		char *sky_name,char *file_ext);
	/*	Draws QuakeII sky box */
	void		V_GLOBAL	vglDrawSky(quake_sky_t sky_textures,
		vec_t sky_intensity,vec_t sky_opacity);
#ifdef 	__cplusplus
}
#endif	/* __cplusplus */

#endif	/* V_GL_H */
/*</pre></strong></body></html>*/
/*EOF*/
