/*<html><body bgcolor=#B4C4B0><font color=#000000><strong><pre>*/
#ifndef V_WMF_H
#define V_WMF_H

/*
 *		Can be ported to other OSes; but currently implemented only on
 *	native environment.
 */
typedef	struct wmf_s
{
	void	*dc;						/* Win32: DC of metafile */
	void	*hfont;						/* Win32: handle of current font */
	u32_t	width, height;				/* Dimensions in millimeters */
	u32_t	in1000_width, in1000_height;	/* ", in inches*1000 */
	char	name[_MAX_PATH];			/* name or '//./clipboard' */
}	wmf_t, *wmf_p;

#ifdef 	__cplusplus
extern	"C" {
#endif	/* __cplusplus */
	/*
	 *	Must be called BEFORE any wmfCreateXXX.
	 */
	void	V_GLOBAL	wmfSetExtents(u32_t ext_x,u32_t ext_y);
	/*	Returns metafile extents */
	u32_t	V_GLOBAL	wmfGetMaxX(void);
	u32_t	V_GLOBAL	wmfGetMaxY(void);
	/*
	 *	wmfCreate() - creates metafile DC.
	 *	Input:
	 *		mmWidth, mmHeight - width and height in millimeters.
	 *	Output:
	 *		Allocated and initialized wmf_t struct or NULL on error.
	 *	Notes.
	 *		1. If at least one of mmWidth and mmHeight parameters is zero,
	 *	extents of metafile set equals to size of A4 sheet.
	 *		2. Extents of metafile can be always 1000 by 1000 units.
	 *		3. Win32 only: if name is "//./clipboard", subsequent call
	 *	to wmfClose() places data into system clipboard.
	 */
	wmf_p	V_GLOBAL	wmfCreate(char *name,u32_t mmWidth,u32_t mmHeight);
	/*
	 *	wmfClose() - saves metafile.
	 *	Input:
	 *		Address of wmf_t object.
	 *	Output:
	 *		true if success, false otherwise; wmf_t object cleaned and freed.
	 */
	vboolean	V_GLOBAL	wmfClose(wmf_t *mf);
	/*
	 *	GDI's GetTextExtentPoint32 not always works with metafiles.
	 */
	vboolean V_GLOBAL	wmfTextExtent(wmf_t *mf,char *str,u32_t *cx,u32_t *cy);
	/*
	 *	Changes font for DC of metafile.
	 *	Input:
	 *		mf - wmf_t object;
	 *		char_height - height of character (in metafile's logical
	 *			units; 100 mean 1/10 of metafile's height etc.);
	 *		deg_ori - orientation in degrees (counter-clockwise);
	 *		weight - weight from 0 to 900;
	 *		italic - if 1, italic font used;
	 *		face - font face name.
	 */
	void	V_GLOBAL	wmfChangeFont(wmf_t *mf,int char_height,int deg_ori,int weight,int italic,char *face);
    /*
     *	wmfCreateOffscreenWindow() - creates 'Metafile' OpenGL context.
     *	Input:
     *		name - name of WMF file to save;
     *		mmWidth, mmHeight - dimensions of metafile in millimeters;
     *		feedback_buffer_hint - appr. size of OpenGL feedback buffer
     *			in BYTES.
     *	Output:
     *		Feedback metafile handle or NULL.
     *	Note. Alters default extents.
     */
    void_p		V_GLOBAL	wmfCreateOffscreenWindow(char *name,
    	u32_t mmWidth,u32_t mmHeight,u32_t feedback_buffer_hint);
    /*
     *	Saves captured OpenGL data and frees all objects.
     */	
    vboolean	V_GLOBAL	wmfDestroyWindow(void *fdmf);
	void_p		V_GLOBAL	wmfGetDevice(wmf_t *mf);
#if	0
#define	wmfGetDevice(mf)	((mf)->dc)
#endif	/* 0 */
#ifdef 	__cplusplus
}
#endif	/* __cplusplus */

#endif	/* V_WMF_H */
/*</pre></strong></body></html>*/
/*EOF*/
