/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	<X11/Xlib.h>
#include	<GL/glx.h>
#define		SGI_GL_H
#include	"../../include/v_lib.h"
#include	"../v_mt.h"
#include	"../v_os.h"

/*
 *	Internal GL context: unix form.
 */
typedef struct		gl_context_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;			/* GL_MAGIC */
#endif	/* !CHECKED_BUILD */
	/*				Platform - independent part */
	u32_t			x, y;			/* Position of window */
	u32_t			width,height;	/* Window extents */
	vboolean		fullscreen;		/* == true for FS window */
	u32_t			bpp;			/* bit depth (FS only) */
	u32_t			refresh;		/* refresh rate (FS only) */
	vboolean		bitmap_based;	/* == 1 for offscreen windows */
	vboolean		accelerated;	/* == 1 for HW renderer */
	vboolean		was_selected;	/* == 1 for 2nd and other selections */
	char			*vendor;		/* information about OpenGL driver */
	char			*renderer;
	char			*version;
	u32_t			num_extensions;
	char			**extensions;
	/*	GLX-specific part */
	XSetWindowAttributes attr;
	Display			*dpy;
	Window			parent;			/* used for child windows */
	char			title[128];		/* window title */
	Window			win;			/* window handle */
	vboolean		focused;		/* activity flag (focused) */
	GLXContext		ctx;
	fifo_t			mqueue;			/* message queue */
	fifo_t			aqueue;			/* deferred action queue, CURRENTLY NOT USED, see glut sources */
}	gl_context_t, *gl_context_p;

#ifdef	CHECKED_BUILD
#define	gl_magic	0x56474C57
#endif	/* CHECKED_BUILD */

/*		Utility functions */
V_LOCAL	gl_context_p
#ifdef	CHECKED_BUILD
vglSafeGetCurrentContext(char *caller)
#else	/* !CHECKED_BUILD */
vglSafeGetCurrentContextQuiet(void)
#define	vglSafeGetCurrentContext(caller)	vglSafeGetCurrentContextQuiet()
#endif	/* !CHECKED_BUILD */
{
	gl_context_t *ctx;
	ctx=(gl_context_t*)GetThreadData()->gl_context;
#ifdef	CHECKED_BUILD
	if(!ctx)
		conlPrintf(LOG_ERROR,"%s(): OpenGL not initialized.\n",
			caller);
	else
		if(ctx->magic != gl_magic)
		{
			conlPrintf(LOG_ERROR,
				"%s(): bad current context, magic wrong.\n",
				caller);
			ctx=NULL;
		}
#endif	/* !CHECKED_BUILD */
	return ctx;
}

#define	vglSafeCheckContext(caller,win) \
		((gl_context_p)libvIsHandle(win,caller,"GL context",gl_magic))

V_LOCAL	gl_context_p
vglAllocContext(u32_t x,u32_t y,u32_t width,u32_t height,
	vboolean fullscreen,void *parent,char *title,
	u32_t bpp,u32_t refresh,vboolean bitmap_based)
{
	gl_context_t *ctx;
	ALLOC_OBJECT(ctx,gl_context_t,0,gl_magic);
	ctx->x=x;
	ctx->y=y;
	ctx->width=width;
	ctx->height=height;
	ctx->fullscreen=fullscreen;
	ctx->parent=(Window)parent;
	fifoInit(&ctx->mqueue,sizeof(input_record_t));
	fifoInit(&ctx->aqueue,sizeof(input_record_t));
	if(title)
		strCopy(ctx->title,title,128);
	ctx->bpp=bpp;
	ctx->refresh=refresh;
	ctx->bitmap_based=bitmap_based;
	return ctx;
}

/*
 *	Destroys GLX context.
 */
void	V_LOCAL
vglFinalizeContext(gl_context_t *ctx)
{
	u32_t i;
	if(!ctx)
		return;
#ifdef	CHECKED_BUILD
	if(ctx->vendor && ctx->renderer && ctx->version)
		conlPrintf(LOG_INFO,"vglFinalizeContext(): %s OpenGL on "
			"%s ver. %s\n",ctx->vendor,ctx->renderer,ctx->version);
#endif	/* CHECKED_BUILD */
	/* Make no context current */
	if(ctx->ctx)
	{
		glXMakeCurrent(NULL,0,NULL);
		glXDestroyContext(ctx->dpy,ctx->ctx);
		ctx->ctx=NULL;
	}
	fifoClear(&ctx->mqueue);
	fifoClear(&ctx->aqueue);
	/* free strdup()'ed memory */
	if(ctx->vendor)
		free(ctx->vendor);
	if(ctx->renderer)
		free(ctx->renderer);
	if(ctx->version)
		free(ctx->version);
	if(ctx->extensions)
	{
		for(i=0;i<ctx->num_extensions;i++)
			free(ctx->extensions[i]);
		free(ctx->extensions);
	}
}

/*
 *	Creates simple window and initializes OpenGL
 */
vboolean	V_LOCAL
vglCreateWindowedContext(gl_context_t *ctx)
{
	int	attrib[] =
	{
		GLX_RGBA,
		GLX_RED_SIZE,1,
		GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1,
		GLX_DEPTH_SIZE, 1,
		GLX_DOUBLEBUFFER,
		None
	};
	int	scrnum;
	XSizeHints *size;
 	unsigned long mask;
	Window root;
	XVisualInfo *visinfo;
	ctx->dpy=XOpenDisplay(NULL);
	if(!ctx->dpy)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,"vglCreateWindowedContext(): unable to "
			"open default display, trying 'localhost:0'...\n");
#endif	/* CHECKED_BUILD */
		ctx->dpy=XOpenDisplay("localhost:0");
		if(!ctx->dpy)
		{
#ifdef	CHECKED_BUILD
			conlPuts(LOG_ERROR,
				"vglCreateWindowedContext(): XOpenDisplay() failed.\n");
#endif	/* CHECKED_BUILD */
			return false;
		}
	}
#ifndef ENABLE_AUTOREPEAT
	XAutoRepeatOff(ctx->dpy);
#endif /* ENABLE_AUTOREPEAT */
	scrnum=DefaultScreen(ctx->dpy);
	root=RootWindow(ctx->dpy,scrnum);
	visinfo = glXChooseVisual(ctx->dpy,scrnum,attrib);
	if(!visinfo)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglCreateWindowedContext(): "
			"couldn't get an RGB, Double-buffered visual.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	ctx->attr.background_pixel = 0;
	ctx->attr.border_pixel = 0;
	ctx->attr.colormap = XCreateColormap(ctx->dpy,root,visinfo->visual,AllocNone);
	ctx->attr.event_mask = StructureNotifyMask | ExposureMask 
		| KeyPressMask | KeyReleaseMask | KeymapStateMask
		| ButtonPressMask | ButtonReleaseMask
		| EnterWindowMask | LeaveWindowMask 
		| PointerMotionMask;
	mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
	ctx->win = XCreateWindow(
		ctx->dpy,root,
		0,0,ctx->width,ctx->height,
		0,visinfo->depth,
		InputOutput,
		visinfo->visual,
		mask,
		&ctx->attr);
	if(!ctx->win)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglCreateWindowedContext(): "
			"couldn't create window.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	/* Set window title. Is this correct ? */
	XStoreName(ctx->dpy,ctx->win,ctx->title);
    /*
    XSetWindowAttributes ...
    XSetWMName(ctx->dpy,ctx->win,&attrib);
    */
	/* Set size hints */
	if(ctx->fullscreen)
	{
		size=XAllocSizeHints();
		size->flags=USPosition|USSize|PPosition|PSize|PMinSize|PMaxSize;
		size->x=0;
		size->y=0;
		size->width=size->min_width=size->max_width=size->base_width=\
			ctx->width;
		size->height=size->min_height=size->max_height=size->base_height=\
			ctx->height;
		XSetWMNormalHints(ctx->dpy,ctx->win,size);
		XFree(size);
	}
	/* Set no backgroung pixmap: we will paint window in each frame */
	XSetWindowBackgroundPixmap(ctx->dpy,ctx->win,None);
	/* Create context */
	ctx->ctx = glXCreateContext(ctx->dpy,visinfo,NULL,True);
	if(!ctx->ctx)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglCreateWindowedContext(): "
			"couldn't create OpenGL context.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	XMapWindow(ctx->dpy,ctx->win);
	return vglMakeCurrent((glw_t)ctx);
}

/*
 *	Translates XLib keycode info LibV key code.
 *	This is HACK, but this working ... :(
 */
keycode_t	V_LOCAL
keycodeXCode(unsigned int keycode)
{
	keycode_t key;
	switch(keycode)
	{
		case 9:			key=key_escape; break;
		case 67:		key=key_f1;	break;
		case 68:		key=key_f2;	break;
		case 69:		key=key_f3;	break;
		case 70:		key=key_f4;	break;
		case 71:		key=key_f5;	break;
		case 72:		key=key_f6;	break;
		case 73:		key=key_f7;	break;
		case 74:		key=key_f8;	break;
		case 75:		key=key_f9;	break;
		case 76:		key=key_f10;break;
		
		case 95:		key=key_f11;	break;
		case 96:		key=key_f12;	break;
		
		case 111:		key=key_printscreen;	break;
		case 78:		key=key_scrolllock;		break;
		case 110:		key=key_pause;			break;
		
		case 49:		key=key_back_quote; break;
		case 10:		key=key_1; break;
		case 11:		key=key_2; break;
		case 12:		key=key_3; break;
		case 13:		key=key_4; break;
		case 14:		key=key_5; break;
		case 15:		key=key_6; break;
		case 16:		key=key_7; break;
		case 17:		key=key_8; break;
		case 18:		key=key_9; break;
		case 19:		key=key_0; break;
		case 20:		key=key_minus; break;
		case 21:		key=key_equal; break;
		
		case 51:		key=key_back_slash; break;
		case 22:		key=key_backspace; break;
		
		case 23:		key=key_tab; break;
		case 24:		key=key_Q; break;
		case 25:		key=key_W; break;
		case 26:		key=key_E; break;
		case 27:		key=key_R; break;
		case 28:		key=key_T; break;
		case 29:		key=key_Y; break;
		case 30:		key=key_U; break;
		case 31:		key=key_I; break;
		case 32:		key=key_O; break;
		case 33:		key=key_P; break;
		case 34:		key=key_left_bracket; break;
		case 35:		key=key_right_bracket; break;
		case 36:		key=key_enter; break;

		case 66:		key=key_capslock; break;
		case 38:		key=key_A; break;
		case 39:		key=key_S; break;
		case 40:		key=key_D; break;
		case 41:		key=key_F; break;
		case 42:		key=key_G; break;
		case 43:		key=key_H; break;
		case 44:		key=key_J; break;
		case 45:		key=key_K; break;
		case 46:		key=key_L; break;
		case 47:		key=key_colon; break;
		case 48:		key=key_apostrophe; break;
		
		case 50:		key=key_left_shift; break;
		case 52:		key=key_Z; break;
		case 53:		key=key_X; break;
		case 54:		key=key_C; break;
		case 55:		key=key_V; break;
		case 56:		key=key_B; break;
		case 57:		key=key_N; break;
		case 58:		key=key_M; break;
		case 59:		key=key_comma; break;
		case 60:		key=key_dot; break;
		case 61:		key=key_slash; break;
		case 62:		key=key_right_shift; break;
		
		case 37:		key=key_left_control; break;
		case 64:		key=key_left_alt; break;
		case 65:		key=key_space; break;
		case 113:		key=key_right_alt; break;
		case 116:		key=key_right_win; break;
		case 117:		key=key_context; break;
		case 109:		key=key_right_control; break;
		
		case 106:		key=key_insert; break;
		case 97:		key=key_home; break;
		case 99:		key=key_pageup; break;
		case 107:		key=key_delete; break;
		case 103:		key=key_end; break;
		case 105:		key=key_pagedown; break;
		
		case 100:		key=key_left; break;
		case 98:		key=key_up; break;
		case 104:		key=key_down; break;
		case 102:		key=key_right; break;
		
		case 77:		key=key_numlock; break;
		case 112:		key=key_grey_slash; break;
		case 63:		key=key_grey_asterisk; break;
		case 82:		key=key_grey_minus; break;
		case 79:		key=key_numpad_7; break;
		case 80:		key=key_numpad_8; break;
		case 81:		key=key_numpad_9; break;
		case 86:		key=key_grey_plus; break;
		case 83:		key=key_numpad_4; break;
		case 84:		key=key_numpad_5; break;
		case 85:		key=key_numpad_6; break;
		case 87:		key=key_numpad_1; break;
		case 88:		key=key_numpad_2; break;
		case 89:		key=key_numpad_3; break;
		case 108:		key=key_grey_enter; break;
		case 90:		key=key_numpad_0; break;
		case 91:		key=key_numpad_dot; break;
		default:		key=key_unknown; break;
	}
	return key;
}


/*
 *	Returns state of modifier keys
 */
vboolean	V_LOCAL	x_caps_on(u32_t ext)
{
	u32_t shift, caps;
	shift = ext & 1;
	caps = (ext & 2) ? 1 : 0;
	return (caps != shift) ? true : false;
}

#if 0

vboolean	V_LOCAL	x_num_lock_on(u32_t ext)
{
	return (ext & 16) ? true : false;
}

vboolean	V_LOCAL	x_scroll_lock_on(u32_t ext)
{
	return (ext & 128) ? true : false;
}

vboolean	V_LOCAL	x_locale_ru(u32_t ext)
{
	return (ext & 8192) ? true : false;
}

#endif

keycode_t	V_LOCAL	x_get_button(unsigned int button)
{
	switch(button)
	{
		case Button1: return key_left_mouse;
		case Button2: return key_middle_mouse;
		case Button3: return key_right_mouse;
		default: break;
	}
	return key_unknown;
}

vboolean V_LOCAL
x_process_message(gl_context_t *ctx,XEvent *event)
{
	input_record_t ev;
	vboolean mouse_event=false;
	switch(event->type)
	{
		case KeyPress:
			ev.event=ev_key_down;
			ev.first_code=keycodeXCode(event->xkey.keycode);
			ev.second_code=keycodeGetChar(ev.first_code,
				!x_caps_on(event->xkey.state),NULL);
			break;
		case KeyRelease:
			ev.event=ev_key_up;
			ev.first_code=keycodeXCode(event->xkey.keycode);
			ev.second_code=0;
			break;
		case MotionNotify:
			ev.event=ev_mouse_move;
			ev.first_code=event->xmotion.x;
			ev.second_code=event->xmotion.y;
			break;
		case ButtonPress:
			ev.event=ev_key_down;
			ev.first_code=x_get_button(event->xbutton.button);
			ev.second_code=0;
			mouse_event=true;
			break;
		case ButtonRelease:
			ev.event=ev_key_up;
			ev.first_code=x_get_button(event->xbutton.button);
			ev.second_code=0;
			mouse_event=true;
			break;
		default:
			return false;
	}
	fifoPush(&ctx->mqueue,&ev);
	/*
	 *	Post mouse_move event after mouse press/release events.
	 */
	if(mouse_event)
	{
		ev.event=ev_mouse_move;
		ev.first_code=event->xmotion.x;
		ev.second_code=event->xmotion.y;
		fifoPush(&ctx->mqueue,&ev);
	}
	return true;
}

#if 0

/*
 *	Changes screen resolution, display depth and refresh rate.
 */
vboolean	V_LOCAL
vidSetMode(int width,int height,int bpp,int refresh_rate)
{
	/* TODO: add support for XF86-VidModeExtension */
	return false;
}

#endif

/*
 *	Sets default display settings.
 */
void	V_LOCAL
vidDefaultMode(void)
{
	return;
}

glw_t	V_GLOBAL
vglCreateWindow(char *title,u32_t x,u32_t y,u32_t width,u32_t height)
{
	gl_context_t *ctx;
	ctx=vglAllocContext(x,y,width,height,false,NULL,title,0,0,false);
	if(vglCreateWindowedContext(ctx))
		return (glw_t)ctx;
	else
	{
		vglDestroyWindow(ctx);
		return NULL;
	}
}

glw_t	V_GLOBAL
vglCreateFullscreenWindow(char *title,u32_t width,u32_t height,u32_t bpp,u32_t refresh)
{
	return vglCreateWindow(title,0,0,width,height);
}

glw_t	V_GLOBAL
vglCreateChildWindow(char *title,u32_t x,u32_t y,u32_t width,u32_t height,
	void *parent)
{
	return NULL;
}

glw_t	V_GLOBAL
vglCreateMappedWindow(u32_t width,u32_t height)
{
	return NULL;
}

void	V_GLOBAL
vglDestroyWindow(glw_t win)
{
	gl_context_t *ctx=(gl_context_t*)win;
	vglFinalizeContext(ctx);
	if(ctx->fullscreen)
		vidDefaultMode();
	if(ctx->attr.colormap)
		XFreeColormap(ctx->dpy,ctx->attr.colormap);
	if(ctx->win)
		XDestroyWindow(ctx->dpy,ctx->win);
	memFree(ctx);
}

vboolean	V_GLOBAL
vglMakeCurrent(glw_t win)
{
	gl_context_t *ctx;
	thread_data_t *td;
	if(!win)
	{
		glXMakeCurrent(NULL,0,NULL);
		return true;
	}
	ctx=vglSafeCheckContext("vglMakeCurrent",win);
	if(!ctx)
		return false;
	glXMakeCurrent(ctx->dpy,ctx->win,ctx->ctx);
	td=GetThreadData();
	td->gl_context=(void*)ctx;
	if(!ctx->was_selected)
	{
		int i;
		char *exts;
		ctx->was_selected=true;
		ctx->vendor=(char*)glGetString(GL_VENDOR);
		if(ctx->vendor)
			ctx->vendor=strdup((const char*)ctx->vendor);
		ctx->renderer=(char*)glGetString(GL_RENDERER);
		if(ctx->renderer)
			ctx->renderer=strdup((const char*)ctx->renderer);
		ctx->version=(char*)glGetString(GL_VERSION);
		if(ctx->version)
			ctx->version=strdup((const char*)ctx->version);
		ctx->extensions=NULL;
		ctx->num_extensions=0;
		exts=(char*)glGetString(GL_EXTENSIONS);
		if(exts)
		{
			parser_t *ep=tokStart(exts,NULL);
			while(tokNext(ep))
				ctx->num_extensions++;
			if(ctx->num_extensions)
			{
				ctx->extensions=(char**)malloc(ctx->num_extensions*sizeof(char*));
				tokRestart(ep,NULL);
				for(i=0;tokNext(ep);i++)
					ctx->extensions[i]=strdup(tokLast(ep));
			}
			free(ep);
		}
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,"vglMakeCurrent(): %s OpenGL on %s ver. %s\n",
			ctx->vendor,ctx->renderer,ctx->version);
#if		0
		conlPuts(LOG_INFO,"OpenGL extensions: \n");
		for(i=0;i<(int)ctx->num_extensions;i++)
			conlPrintf(LOG_INFO,"    %s\n",ctx->extensions[i]);
#endif	/* !0 */
#endif	/* CHECKED_BUILD */
	}
    return true;
}

glw_t	V_GLOBAL
vglGetCurrentWindow(void)
{
	return (glw_t)(GetThreadData()->gl_context);
}

u32_t V_LOCAL
vglPendingCTX(gl_context_t *ctx)
{
	XEvent event;
	while(XPending(ctx->dpy))
	{
		XNextEvent(ctx->dpy,&event);
		switch(event.type)
		{
			case FocusIn:
				ctx->focused=true;
				puts("focus "); /* TODO */
				break;
			case FocusOut:
				puts("focus lost"); /* TODO */
				ctx->focused=false;
				break;
			case Expose:
				break;
			case ConfigureNotify:
				glViewport(0,0,event.xconfigure.width,event.xconfigure.height);
				break;
			default:
				x_process_message(ctx,&event);
				break;
		}
	}
	return fifoCount(&ctx->mqueue);
}

int	V_GLOBAL
vglPending()
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglPending");
	if(ctx)
		return (int)vglPendingCTX(ctx);
	else
		return 0;
}

vboolean	V_GLOBAL
vglNextEvent(input_record_t *event_return)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglNextEvent");
	if(!ctx)
		return false;
	if(!event_return)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"vglNextEvent(NULL): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(vglPendingCTX(ctx))
		fifoPop(&ctx->mqueue,event_return);
	return fifoCount(&ctx->mqueue) > 0;
}

vboolean	V_GLOBAL
vglSetFocus()
{
	return true;
}

vboolean	V_GLOBAL
vglResizeWindow(u32_t width,u32_t height)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglResizeWindow");
	if(!ctx)
		return false;
	XResizeWindow(ctx->dpy,ctx->win,width,height);
	return false;
}

vboolean	V_GLOBAL
vglMoveWindow(u32_t left,u32_t top)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglMoveWindow");
	if(!ctx)
		return false;
	XMoveWindow(ctx->dpy,ctx->win,left,top);
	return false;
}

void		V_GLOBAL
vglSwapBuffers(void)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglSwapBuffers");
	if(!ctx)
		return;
	glFlush();
	glFinish();
	glXSwapBuffers(ctx->dpy,ctx->win);
	/* If window is not active, reduce CPU & 3Daccel overhead */
	if(!ctx->focused)
		timeDelay(10);	/* Maximum framerate limited to 100 FPS */
}

u8_p		V_GLOBAL
vglReadPixels(u32_t *width,u32_t *height)
{
	u32_t half;
	void *buffer;
	u8_t *upper, *lower;
	u8_t *pixels;
	u32_t stride_adjust;
	thread_data_t *td;
	gl_context_t *ctx=vglSafeGetCurrentContext("vglReadPixels");
	if(!ctx)
		return NULL;
	td=GetThreadData();
	pixels=(u8_t*)malloc(ctx->width*ctx->height*4);
	stride_adjust=ctx->width*3;
	stride_adjust=INT_ALIGN(stride_adjust);
	stride_adjust-=ctx->width*3;
	glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
	glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glReadPixels(0,0,ctx->width,ctx->height,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
	buffer=malloc(ctx->width*4);
	half=ctx->height / 2;
	upper=pixels;
	lower=upper+(ctx->height-1)*ctx->width*4;
	while(half--)
	{
		memCopy32(buffer,upper,ctx->width);
		memCopy32(upper,lower,ctx->width);
		memCopy32(lower,buffer,ctx->width);
		upper+=ctx->width*4;
		lower-=ctx->width*4;
	}
	free(buffer);
	if(width)
		*width=ctx->width;
	if(height)
		*height=ctx->height;
	return pixels;
}
