/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#ifndef   V_MT_H
#define   V_MT_H

#include <../source/vfs/vfsutil.h>

/*	======================================================================
 *	Here are private declarations which are used only by LibV functions.
 *	======================================================================
 */

#ifdef	RIGHT_SLASHED
#define	PATH_DELIMITER				'/'
#else	/* !RIGHT_SLASHED */
#define	PATH_DELIMITER				'\\'
#endif	/* !RIGHT_SLASHED */

#ifdef		__WIN32__
#define		NT_CONSOLE
#endif		/* __WIN32__ */

/*
 *	LibV internal data, shared between all threads.
 */
#define			LIBV_GLOBAL_MAGIC		0x4C564744		/* LVGD */
typedef struct	global_data_s
{
	u32_t		magic;				/* LIBV_GLOBAL_MAGIC */
	u32_t		thread_count;		/* number of user-created threads */
	/* v_log.c */
	FILE		*log_file;
	u32_t		log_upper_level;
#ifdef			NT_CONSOLE
	void		*inhandle, *outhandle;
	vboolean	own_console;
#endif			/* NT_CONSOLE */
	/* all callbacks are thread-safe, so caller can accumulate
	 * pointers to them, unlock global data and than call callbacks
	 * secuentially. TODO: debug console logs: disallow to set cb twice ? */
	vp_stack_t	log_callbacks;
#ifdef	__WIN32__
	/* v_gl.c: counter for unical class names */
	u32_t		context_count;
	/* TODO: workaround NT ICD OpenGL bug - implement some method to 
	 * disallow context creation from non-main thread after destruction of
	 * the first such context (GPF raised from NtWaitForMultipleObjectEx) */
	/*
	 *	And ... may be we need to implement another R3D abstraction layer ???
	 *	Example: to support - OGL, D3D (execute/DP), Glide, MeTal, 
	 *	S3DTK (Hmm... S3 Inc. NO LONGER EXIST !), SGL (tell me, were I can
	 *	buy ancient PowerVR ?), Intel 3DR (ya, and implement dynamic link 
	 *	to software GSI opengl.dll, MesaGL)... this is nuts.
	 */
#endif	/* __WIN32__ */
	/* v_reg.c */
	ntree_p		reg_root;		/* registry root node */
	ntree_p		current_key;	/* registry current key */
}	global_data_t, *global_data_p;

#ifndef	RIGHT_SLASHED

/*
 *	TODO: move to v_fsth.h
 */

typedef struct file_apis_s
{
	void (V_CALL *fsEnumFiles)(
		void (V_CALL *dir_callback)(char *full_unix_style_dir_name),
		void (V_CALL *file_callback)(char *full_unix_style_file_name),
		vboolean recursive);
	void (V_CALL *fsSplitPath)(char *src,char *dir,char *name,char *ext);
	void (V_CALL *fsMakePath)(char *dest_path,char *dir,
		char *name,char *ext);
	void (V_CALL *fsReplaceExt)(char *dst,char *name,char *ext);
	file_t	(V_CALL *fileOpenRead)(char *name);
	file_t	(V_CALL	*fileOpenWrite)(char *name);
	file_t	(V_CALL *fileOpenAppend)(char *name);
}	file_apis_t, *file_apis_p;

void	V_CALL	uvfsReplaceExt(char *dst,char *name,char *ext);
void	V_CALL	nvfsReplaceExt(char *dst,char *name,char *ext);
void	V_CALL	nvfsSplitPath(char *src,char *dir,char *name,char *ext);
void	V_CALL	nvfsMakePath(char *dest_path,char *dir,char *name,char *ext);
file_t	V_CALL	ufileOpenRead(char *name);
file_t	V_CALL	nfileOpenRead(char *name);
file_t	V_CALL	ufileOpenWrite(char *name);
file_t	V_CALL	nfileOpenWrite(char *name);
file_t	V_CALL	ufileOpenAppend(char *name);
file_t	V_CALL	nfileOpenAppend(char *name);

#endif	/* RIGHT_SLASHED */

/*
 *	LibV internal TSD.
 */
#define			LIBV_TSD_MAGIC		0x4C565444			/* LVTD */
typedef struct	thread_data_s
{
	u32_t		magic;
	/* v_image.c */
	u8_t		constant_alpha;
	/* v_wmf.c */
	u32_t		meta_x_extent,meta_y_extent;
	/* v_fs.c */
#ifndef		RIGHT_SLASHED
	file_apis_t *file_apis;	/* file IO routines */
#endif		/* RIGHT_SLASHED */
	/* v_gl.c */
	void		*gl_context;
#ifdef	__WIN32__
	/* BUGBUG: these fields are obolete and should be removed */
	vboolean	lower_case;
	vboolean	caps_lock;
	vboolean	num_lock;
	/* v_incon.c */
	u32_t		con_prev_mouse_flags;
#elif	defined(__UNIX__)
	/* used to determine status of the thread - see v_sched.c */
	void		*this_thread;
#endif	/* __UNIX__ */
}	thread_data_t, *thread_data_p;

/*
 *	Internally used types and macros.
 */

#define GetThreadData() \
		((thread_data_t*)(libvGetSpecific()))

/*
 *		Default constructor
 */
#define	ALLOC_OBJECT(dest,type,extra_memory,magic)\
		dest=(type*)memAlloc(sizeof(type)+extra_memory);\
		memZero((void*)dest,sizeof(type)+extra_memory);\
		MARK_OBJECT(dest,magic);

#ifdef	CHECKED_BUILD
#define	MARK_OBJECT(object,magic)	*((u32_t*)(object))=(u32_t)(magic)
#else	/* !CHECKED_BUILD */
#define	MARK_OBJECT(object,magic)
#endif	/* !CHECKED_BUILD */

/*	Vertex drawing function */
typedef	void (*vertex_draw_proc_t)(vec_t *v,vec_t *c,vec_t *n,vec_t *t);
/*	Triangle drawing function */
typedef	void (*tri_draw_proc_t)(tlist_face_t *tri);
/*	Draw flags */
#define	NORMAL_DRAW_BIT			1
#define	TEXTURE_DRAW_BIT		2
#define	COLOR_DRAW_BIT			4

/*
 *	Internally used functions.
 */
#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
	/*
	**	Prints error message in case when conlPrintf() can't be used.
	**	TODO: something is wrong there ... but WHAT ???
	*/
	void		V_CALL			libvError(char *msg);
	/*
	**	Changing contents of private TSD, used for internal TSD
	**	managing. LibV clients SHOULDN'T call these functions !
	**
	**	Implemented in v_sched.c
	*/
	void		V_CALL		libvSetSpecific(void_p specific);
	void_p		V_CALL		libvGetSpecific(void);
	/*
	**	Returns contents of guarded section. After use, section
	**	MUST be unlocked by libvUnlockGlobal().
	**	Implemented in v_sched.c
	*/
	global_data_p	V_CALL	libvLockGlobal(void);
	/*
	**	Unlocks global data, makes it possible to use by another threads.
	**	Implemented in v_sched.c
	*/
	void			V_CALL	libvUnlockGlobal(void);
	/*
	**	General - purpose function for testing valididy of object's
	**	handles. Returns NULL (and complains an error) if:
	**		a)	agrument is NULL;
	**		b)	(when CHECKED_BUILD) first DWORD of memory pointed
	**			by 'handle' != magic.
	**	Otherwise returns 'handle' back.
	**	Implemented in v_mt.c.
	*/
#ifdef	CHECKED_BUILD
	void_p	V_CALL	libvIsHandle(void *handle,
		char *caller,	/* Name of function handle passed to */
		char *object,	/* Sence of handle: file, ipc, etc... */
		u32_t magic);	/* Magic: must be in FIRST DWORD of object */
#else	/* !CHECKED_BUILD */
#define	libvIsHandle(handle,caller,object,magic)	(handle)
#endif	/* CHECKED_BUILD */
	/*
	**	GL draw procs: returns functions for vertex/triangle drawing
	**	via OpenGL, depending on current OpenGL state.
	*/
	u32_t	/* returns "draw mask", which reflects current GL state */
			V_CALL	vglGetDrawBits(void);
	vertex_draw_proc_t	/* returns proc for vertex drawing */
			V_CALL	vglGetVertexDrawProc(int draw_bits);
	tri_draw_proc_t	/* same as above, for triangles */
			V_CALL	vglGetTriangleDrawProc(int draw_bits);
#ifdef	__cplusplus
}
#endif	/* __cplusplus */

/*
 *	Platform - dependent functions, used internally.
 */
#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
#ifdef	__WIN32__
	/*				WE ALWAYS WILL LOVE WINDOWS 95 !!! 				*/
	/*	Returns false if we are under GOOD OS, otherwise returns true.	*/
	vboolean	V_CALL	WeAreUnderMustDie(void);
#endif	/* __WIN32__ */
#ifdef	__cplusplus
}
#endif	/* __cplusplus */

#endif	/* V_MT_H */

/*</pre></strong></body></html>*/
/*EOF*/
