/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	<v_lib.h>
#include	<../source/v_mt.h>

/*
 *	TODO: implement unicode & utf-8 text files conversion based on
 *	selected codepage. Cyrillic letters in ../devel/unitxt/u[l|b]ei.txt & utf.txt
 */

#ifdef	CHECKED_BUILD
#define	PARSER_MAGIC	0x53525052
V_LOCAL	char			*parser_object="parser";
#endif	/* CHECKED_BUILD */

typedef struct	real_parser_s
{
#ifdef	CHECKED_BUILD
	u32_t		magic;
#endif	/* CHECKED_BUILD */
	i32_t		tos;
	i32_t		unget;
	char_p		token_buffer;
	char_p		last_token;
	char_p		token_ptr;
	char		delim[MAX_DELIMITERS];
	char		token[MAX_TOKENSIZE];
	char_p		posstack[MAX_TOKENPOSSTACK];
}	real_parser_t;

#ifdef	THIS
#undef	THIS
#endif	/* THIS */
#define	THIS(handle)	((real_parser_t*)handle)

char_p		V_LOCAL
tokSetLast(real_parser_t *tok,char *ret)
{
	tok->last_token=ret;
	return ret;
}

parser_t	V_GLOBAL
tokStart(char *buffer,char *delim)
{
	real_parser_t *t;
	if(!buffer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"tokStart('[NULL]','%s'): incorrect parameter.\n",dbs(delim));
#endif
		return NULL;
	}
	ALLOC_OBJECT(t,real_parser_t,0,PARSER_MAGIC);
	t->last_token=t->token;
	t->tos=t->unget=0;
	t->token_ptr=t->token_buffer=buffer;
	if(delim)
		strCopy(t->delim,delim,MAX_DELIMITERS);
	return (parser_t)t;
}

parser_t	V_GLOBAL
tokStartFile(char *filename,char *delim)
{
	real_parser_t *t;
	file_t infile;
	u32_t insize;
	if(!filename || !delim)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"tokStartFile('%s','%s'): NULL parameter(s).\n",
			dbs(filename),dbs(delim));
#endif
		return NULL;
	}
	infile=fileOpenRead(filename);
	if(!infile)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"tokStartFile('%s','%s'): can't open file.\n",
			dbs(filename),dbs(delim));
#endif
		return NULL;
	}
	insize=fileSize(infile);
	if(!insize)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"tokStartFile('%s','%s'): empty file.\n",
			dbs(filename),dbs(delim));
#endif
		fileClose(infile);
		return NULL;
	}
	ALLOC_OBJECT(t,real_parser_t,insize+1,PARSER_MAGIC);
	t->tos=t->unget=0;
	t->token_ptr=t->token_buffer=(char*)(t+1);
	/*
	 *	TODO1: determine type of file: ASCII, LE/BE UNICODE, or UTF-8.
	 *	TODO2: convert to user pctl_map_ispace
	 */
	if(fileRead(infile,t->token_buffer,insize)!=insize)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"tokStartFile('%s','%s'): can't read %d bytes.\n",
			dbs(filename),dbs(delim),insize);
#endif
		free(t);
		fileClose(infile);
		return NULL;
	}
	if(delim)
		strCopy(t->delim,delim,MAX_DELIMITERS);
	t->last_token=t->token;
	return (parser_t)t;
}

void	V_GLOBAL
tokRestart(parser_t tok,char *new_buffer)
{
	if(!libvIsHandle(tok,"tokRestart",parser_object,PARSER_MAGIC))
		return;
	if(new_buffer)
		THIS(tok)->token_buffer=new_buffer;
	THIS(tok)->token_ptr=THIS(tok)->token_buffer;
	THIS(tok)->last_token=THIS(tok)->token;
	THIS(tok)->tos=0;
	return;
}

vboolean	V_GLOBAL
tokPush(parser_t tok)
{
	if(!libvIsHandle(tok,"tokPush",parser_object,PARSER_MAGIC))
		return false;
	if(THIS(tok)->tos==MAX_TOKENPOSSTACK)
		return false;
	THIS(tok)->posstack[THIS(tok)->tos++]=THIS(tok)->token_ptr;
	return true;
}

vboolean V_GLOBAL
tokPop(parser_t tok)
{
	if(!libvIsHandle(tok,"tokPop",parser_object,PARSER_MAGIC))
		return false;
	if(!THIS(tok)->tos)
		return false;
	THIS(tok)->token_ptr=THIS(tok)->posstack[--THIS(tok)->tos];
	return true;
}

void	V_GLOBAL
tokDelim(parser_t tok,char *delim)
{
	if(!libvIsHandle(tok,"tokDelim",parser_object,PARSER_MAGIC))
		return;
	/* Let strCopy() checks for second arg, Ok ? */
	strCopy(THIS(tok)->delim,delim,MAX_DELIMITERS);
	return;
}

char_p	V_GLOBAL
tokNext(parser_t tok)
{
	int i=0;
	if(!libvIsHandle(tok,"tokNext",parser_object,PARSER_MAGIC))
		return NULL;
	if(!*THIS(tok)->token_ptr)
		return tokSetLast(THIS(tok),NULL);
	if(THIS(tok)->unget)
	{
		THIS(tok)->unget=0;
		return tokSetLast(THIS(tok),THIS(tok)->token);
	}
	/* Skip whitespace */
	while(*THIS(tok)->token_ptr<=32 && *THIS(tok)->token_ptr)
		THIS(tok)->token_ptr++;
	if(!*THIS(tok)->token_ptr)	/* no more characters */
		tokSetLast(THIS(tok),NULL);
	if(*THIS(tok)->token_ptr=='"')
	{
		/* Extract quoted token */
		THIS(tok)->token_ptr++;	/* skip quote */
		while(*THIS(tok)->token_ptr != '"' && *THIS(tok)->token_ptr
			&& i<MAX_TOKENSIZE)
			{
				THIS(tok)->token[i]=*THIS(tok)->token_ptr;
				THIS(tok)->token_ptr++;
				i++;
			}
		if(*THIS(tok)->token_ptr=='"')
			THIS(tok)->token_ptr++;
		THIS(tok)->token[i]=0;
	}
	else
	{
		if(strchr(THIS(tok)->delim,*THIS(tok)->token_ptr))
		{
			/* return delimiter */
			THIS(tok)->token[0]=*THIS(tok)->token_ptr;
			THIS(tok)->token[1]=0;
			THIS(tok)->token_ptr++;
		}
		else
		{
			/* extract next token enclosed in delimiters */
			while((i<MAX_TOKENSIZE) &&
				!strchr(THIS(tok)->delim,*THIS(tok)->token_ptr)
				&& *THIS(tok)->token_ptr>32)
			{
				THIS(tok)->token[i]=*THIS(tok)->token_ptr;
				THIS(tok)->token_ptr++;
				i++;
			}
			THIS(tok)->token[i]=0;
		}
	}
	return tokSetLast(THIS(tok),THIS(tok)->token);
}

/*
 *	Same as tokNext(), but whitespace chars threats as usual chars.
 */
char_p	V_GLOBAL
tokStrict(parser_t tok)
{
	int i=0;
	if(!libvIsHandle(tok,"tokStrict",parser_object,PARSER_MAGIC))
		return NULL;
	if(!*THIS(tok)->token_ptr)
		return tokSetLast(THIS(tok),NULL);
	if(THIS(tok)->unget)
	{
		THIS(tok)->unget=0;
		return tokSetLast(THIS(tok),THIS(tok)->token);
	}
	if(!*THIS(tok)->token_ptr)	/* no more characters */
		tokSetLast(THIS(tok),NULL);
	if(strchr(THIS(tok)->delim,*THIS(tok)->token_ptr))
	{
		/* return delimiter */
		THIS(tok)->token[0]=*THIS(tok)->token_ptr;
		THIS(tok)->token[1]=0;
		THIS(tok)->token_ptr++;
	}
	else
	{
		/* extract next token enclosed in delimiters */
		while((i<MAX_TOKENSIZE) &&
			!strchr(THIS(tok)->delim,*THIS(tok)->token_ptr))
		{
			THIS(tok)->token[i]=*THIS(tok)->token_ptr;
			THIS(tok)->token_ptr++;
			i++;
		}
		THIS(tok)->token[i]=0;
	}
	return tokSetLast(THIS(tok),THIS(tok)->token);
}

char_p	V_GLOBAL
tokString(parser_t tok)
{
	int i;
	if(!libvIsHandle(tok,"tokString",parser_object,PARSER_MAGIC))
		return NULL;
	if(!*THIS(tok)->token_ptr)
		return tokSetLast(THIS(tok),NULL);
	while((*THIS(tok)->token_ptr == 0x0d || *THIS(tok)->token_ptr==0x0a)
		&& *THIS(tok)->token_ptr)
		THIS(tok)->token_ptr++;
	if(!*THIS(tok)->token_ptr)
		return tokSetLast(THIS(tok),NULL);
	i=0;
	while(*THIS(tok)->token_ptr!=0x0d && *THIS(tok)->token_ptr!=0x0a && 
		i<MAX_TOKENSIZE && *THIS(tok)->token_ptr)
	{
		THIS(tok)->token[i]=*THIS(tok)->token_ptr;
		THIS(tok)->token_ptr++;
		i++;
	}
	THIS(tok)->token[i]=0;
	return tokSetLast(THIS(tok),THIS(tok)->token);
}

char_p	V_GLOBAL
tokLast(parser_t tok) 
{
	if(!libvIsHandle(tok,"lokLast",parser_object,PARSER_MAGIC))
		return NULL;
	return THIS(tok)->last_token;
}

void	V_GLOBAL
tokFree(parser_t tok)
{
	if(!libvIsHandle(tok,"lokLast",parser_object,PARSER_MAGIC))
		return;
	memFree(tok);
}

#undef	THIS

/*</pre></strong></body></html>*/
/*EOF*/
