/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	<v_lib.h>
#include	<../source/v_mt.h>

/*	=======================================================================
**	Helper API.
*/

btree_p	V_LOCAL
alloc_btree_node(btree_t *parent,int child_index,u32_t extra)
{
	btree_t *n;
	if(parent && parent->children[child_index])
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"alloc_btree_node(): %s child already present.\n",
			child_index ? "right" : "left");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	n=(btree_t*)memAlloc(sizeof(btree_t)+extra);
	n->parent=parent;
	n->children[0]=n->children[1]=NULL;
	if(parent)
		parent->children[child_index]=n;
	return n;
}

void	V_LOCAL
btDeleteAllNodesFast_r(btree_t *root)
{
	if(!root)
		return;
	btDeleteAllNodesFast_r(root->children[0]);
	btDeleteAllNodesFast_r(root->children[1]);
	memFree(root);
}

void	V_LOCAL
btDeleteAllNodes_r(btree_t *root,void (V_CALL *destructor)(void*))
{
	if(!root)
		return;
	btDeleteAllNodes_r(root->children[0],destructor);
	btDeleteAllNodes_r(root->children[1],destructor);
	destructor(root->data);
	memFree(root);
}

ntree_p	V_LOCAL
alloc_tree_node(u32_t extra,ntree_t *parent)
{
	ntree_t *n=(ntree_t*)memAlloc(sizeof(ntree_t)+extra);
	n->parent=parent;
	vpsInit(&n->children);
	if(parent)
		vpsPush(&parent->children,n);
	return n;
}

void	V_LOCAL
treeDeleteAllNodesFast_r(ntree_t *root)
{
	u32_t ind,count;
	ntree_p *ch;
	if(!root)
		return;
	ch=(ntree_p*)vpsGetStorage(&root->children);
	count=vpsCount(&root->children);
	for(ind=0;ind<count;ind++)
		treeDeleteAllNodesFast_r(ch[ind]);
	vpsClear(&root->children);
	memFree(root);
}

void	V_LOCAL
treeDeleteAllNodes_r(ntree_t *root,void (V_CALL *destructor)(void*))
{
	u32_t ind,count;
	ntree_p *ch;
	if(!root)
		return;
	ch=(ntree_p*)vpsGetStorage(&root->children);
	count=vpsCount(&root->children);
	for(ind=0;ind<count;ind++)
		treeDeleteAllNodes_r(ch[ind],destructor);
	vpsClear(&root->children);
	destructor(root->data);
	memFree(root);
}

/*	=======================================================================
**	Global API.
*/

btree_p	V_GLOBAL
btAllocNode(void *data,btree_t *parent,int child_index)
{
	btree_t *n=alloc_btree_node(parent,child_index,0);
	if(n)
		n->data=data;
	return n;
}

btree_p	V_GLOBAL
btStore(void *data,u32_t len,btree_t *parent,int child_index)
{
	btree_t *n;
	if(!data || !len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"btStore(): nothing to store.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	n=alloc_btree_node(parent,child_index,len);
	if(n)
	{
		n->data=(void*)(n+1);
		memCopy(n->data,data,len);
	}
	return n;
}

btree_p	V_GLOBAL
btRootNode(btree_t *node)
{
	if(!node)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"btRootNode(): couldn't ascend from NULL.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	while(node->parent)
		node=node->parent;
	return node;
}

/*
**	Removes all nodes from leftmost to rightmost
*/
void	V_GLOBAL
btDeleteAllNodes(btree_t *root,void (V_CALL *destructor)(void*))
{
	if(!root)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"btDeleteAllNodes(): nothing to delete.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(destructor)
	{
		btDeleteAllNodes_r(root->children[0],destructor);
		btDeleteAllNodes_r(root->children[1],destructor);
		destructor(root->data);
	}
	else
	{
		btDeleteAllNodesFast_r(root->children[0]);
		btDeleteAllNodesFast_r(root->children[1]);
	}
	memFree(root);
}

ntree_p	V_GLOBAL
treeAllocNode(void *data,ntree_t *parent)
{
	ntree_t *n;
	n=alloc_tree_node(0,parent);
	n->data=data;
	return n;
}

ntree_p	V_GLOBAL
treeStore(void *data,u32_t len,ntree_t *parent)
{
	ntree_t *n;
	if(!data || !len)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"treeStore(): nothing to store.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	n=alloc_tree_node(len,parent);
	n->data=(void*)(n+1);
	memCopy(n->data,data,len);
	return n;
}

ntree_p	V_GLOBAL
treeRootNode(ntree_t *node)
{
	if(!node)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"treeRootNode(): couldn't ascend from NULL.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	while(node->parent)
		node=node->parent;
	return node;
}

void	V_GLOBAL
treeDeleteAllNodes(ntree_t *root,void (V_CALL *destructor)(void*))
{
	if(!root)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"treeDeleteAllNodes(): nothing to delete.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(destructor)
		treeDeleteAllNodes_r(root,destructor);
	else
		treeDeleteAllNodesFast_r(root);
}

/*</pre></strong></body></html>*/
/*EOF*/
