/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	<v_lib.h>
#include	<../source/v_mt.h>
#include	<../source/v_os.h>

#ifdef	__WIN32__
#ifndef	__CYGWIN32__
#include <mmsystem.h>
#else	/* !__CYGWIN32__ */
#error	CYGWIN distribution BROKEN: no <mmsystem.h> !!!
#endif	/* !__CYGWIN32__ */
#elif	defined(__UNIX__)
#include <sys/ioctl.h>
#ifdef	__linux
#include <linux/soundcard.h>
#elif	defined(__bsdi__) /* !__linux */
#include <sys/soundcard.h>
#else	/* !__bsdi__ */
#include <machine/soundcard.h>
#endif	/* !__bsdi__ */
#endif	/* __UNIX__ */

/*
 *	LINUX-TODO: determine DSP caps and MAP_IN_BUF, if possible.
 */

/*
 *	TODO: move these into protected section.
 */
vboolean	initialized=false;
vboolean	recording=false;

int         flipflop;
rwl_fifo_t	accum;
u32_t		buf_len;

#ifdef	__WIN32__

char			*buf1, *buf2;
HWAVEIN			hDev=NULL;
WAVEFORMATEX	wfe;
WAVEHDR			whd[2];
vboolean		input_pending=true;

/*
 * Callback for waveInOpen()
 */
#include <../source/argsused.h>
V_LOCAL void CALLBACK
waveInProc(HWAVEIN hwi,UINT uMsg,DWORD dwInstance,
	DWORD dwParam1,DWORD dwParam2)
{
	int done;
	switch(uMsg)
	{
		case WIM_DATA:
			done=flipflop;
			flipflop = (flipflop+1) & 1;
			if(input_pending)
			{
				waveInAddBuffer(hDev,&whd[flipflop],sizeof(WAVEHDR));
				waveInStart(hDev);
			}
			waveInUnprepareHeader(hDev,&whd[done],sizeof(WAVEHDR));
			if(whd[done].dwBytesRecorded)
				rwlPushBuffer(accum,whd[done].lpData,
					whd[done].dwBytesRecorded);
			if(input_pending)
				waveInPrepareHeader(hDev,&whd[done],sizeof(WAVEHDR));
			return;
		case WIM_OPEN:
			return;
		case WIM_CLOSE:
			rwlDestroy(accum);
			memFree(buf1);
			memFree(buf2);
			buf1=buf2=NULL;
			accum=NULL;
			hDev=NULL;
			return;
	}
}

/*
 *	Prepares for capturing wave data.
 */
vboolean	V_GLOBAL
waveCaptureInit(u32_t freq,u32_t nbits,u32_t nchannels)
{
	UINT dev,nd;
	MMRESULT errcode;
#ifdef	CHECKED_BUILD
	WAVEINCAPS wc;
#endif	/* CHECKED_BUILD */
	if(initialized)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"waveCaptureInit(%d,%d,%d): wave capturing "
			"already initialized.\n",freq,nbits,nchannels);
#endif	/* CHECKED_BUILD */
		return false;
	}
	memZero(&wfe,sizeof(WAVEFORMATEX));
	wfe.wFormatTag=WAVE_FORMAT_PCM;
	wfe.nSamplesPerSec=freq;
	wfe.wBitsPerSample=(USHORT)nbits;
	wfe.nChannels=(USHORT)nchannels;
	wfe.nBlockAlign=(USHORT)((wfe.nChannels*wfe.wBitsPerSample) >> 3);
	wfe.nAvgBytesPerSec=wfe.nSamplesPerSec*wfe.nBlockAlign;
	wfe.cbSize=0;
	nd=waveInGetNumDevs();
	/* Select appropriate sound device */
	for(dev=0;dev<nd;dev++)
	{
		errcode=waveInOpen(&hDev,dev,&wfe,(DWORD)waveInProc,0,
			CALLBACK_FUNCTION);
		if(errcode!=MMSYSERR_NOERROR)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_TRACE,"waveCaptureInit(%d,%d,%d): waveInOpen()"
				" failed (code %d).\n",freq,nbits,nchannels,errcode);
#endif		/* CHECKED_BUILD */
			continue;
		}
#ifdef	CHECKED_BUILD
		memZero(&wc,sizeof(WAVEINCAPS));
		waveInGetDevCaps(dev,&wc,sizeof(WAVEINCAPS));
		conlPrintf(LOG_INFO,"waveCaptureInit(): using device '%s'.\n",
			wc.szPname);
#endif	/* CHECKED_BUILD */
		break;
	}
	if(dev==nd)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"waveCaptureInit(%d,%d,%d): "
			"no appropriate hardware found.\n",freq,nbits,nchannels);
#endif	/* CHECKED_BUILD */
		return false;
	}
	/* The ONLY way to work with non-standart sampling frequencies */
	buf_len=nbits * nchannels * freq;
	/* Avoid using too small buffers */
	while(buf_len < 0x4000)
		buf_len <<= 1;
	/* Allocate space for audio data */
	buf1=(char*)memAlloc(buf_len);
	buf2=(char*)memAlloc(buf_len);
	memZero(buf1,buf_len);
	memZero(buf2,buf_len);
	memZero(whd,sizeof(whd));
	whd[0].lpData=buf1;
	whd[0].dwBufferLength=buf_len;
	whd[1].lpData=buf2;
	whd[1].dwBufferLength=buf_len;
	flipflop=0;
	if((waveInPrepareHeader(hDev,&whd[0],sizeof(WAVEHDR))
		==MMSYSERR_NOERROR) && 
		(waveInPrepareHeader(hDev,&whd[1],sizeof(WAVEHDR))
		==MMSYSERR_NOERROR))
	{
		if(waveInAddBuffer(hDev,&whd[flipflop],sizeof(WAVEHDR))
			==MMSYSERR_NOERROR)
		{
			accum=rwlCreate();
			if(accum)
			{
				/* We are ready to capture audio data */
				initialized=true;
				return true;
			}
		}
	}
	memFree(buf1);
	memFree(buf2);
	buf1=buf2=NULL;
	waveInClose(hDev);
	return false;
}

vboolean	V_GLOBAL
waveCaptureResume(void)
{
	if(!initialized)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"waveCaptureResume(): capture was not initialized.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	input_pending=true;
	if(waveInStart(hDev)==MMSYSERR_NOERROR)
	{
		recording=true;
		return true;
	}
	input_pending=false;
	return false;
}

void	V_GLOBAL
waveCapturePause(void)
{
	if(recording)
	{
		waveInStop(hDev);
		recording=false;
	}
#ifdef	CHECKED_BUILD
	else
	{
		conlPuts(LOG_ERROR,
			"waveCapturePause(): capture was not started.\n");
	}
#endif	/* CHECKED_BUILD */
}

void	V_GLOBAL
waveCaptureShutdown(void)
{
	if(initialized)
	{
		input_pending=false;
		waveInReset(hDev);
		waveInClose(hDev);
		recording=false;
		initialized=false;
	}
#ifdef	CHECKED_BUILD
	else
	{
		conlPuts(LOG_ERROR,
			"waveCaptureShutdown(): capture was not initialazed.\n");
	}
#endif	/* CHECKED_BUILD */
}

#elif	defined(__linux)	/* !__WIN32__ */

int			dev=-1;
thread_t	cap_thread=NULL;
cond_t		pause_cond=NULL;
char		*buffer;

V_LOCAL	void V_CALL
waveCaptureThread(void *param)
{
	ssize_t rl=0;
	while(initialized)
	{
		condWait(pause_cond);
		condReset(pause_cond);
		while(recording)
		{
			rl=read(dev,buffer,buf_len);
			if(rl)
				rwlPushBuffer(accum,buffer,rl);
			else
				timeDelay(10);
		}
	}
	threadExit(0);
}

V_LOCAL	void
waveClean(void)
{
	recording=false;
	initialized=false;
	if(pause_cond)
	{
		condSet(pause_cond);
		if(cap_thread)
		{
			threadJoin(cap_thread,NULL);
			cap_thread=NULL;
		}
		condDestroy(pause_cond);
		pause_cond=NULL;
	}
	if(accum)
	{
		rwlDestroy(accum);
		accum=NULL;
	}
	if(dev!=-1)
	{
		close(dev);
		dev=-1;
	}
	if(buffer)
	{
		memFree(buffer);
		buffer=NULL;
	}
}

vboolean	V_GLOBAL
waveCaptureInit(u32_t freq,u32_t nbits,u32_t nchannels)
{
	int inburst,maxlen;
	int dsp_rate, dsp_channels;
#ifdef	USE_SAMPLESIZE
	int dsp_samplesize;
#else	/* !USE_SAMPLESIZE */
	int fmt;
#endif	/* !USE_SAMPLESIZE */
	if(initialized)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"waveCaptureInit(): capture already initialized.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	dev=open("/dev/dsp",O_RDONLY | O_SYNC,S_IRUSR | S_IRGRP);
	if(dev==-1)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"waveCaptureInit(): no input hardware.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	maxlen=(int)(freq*nbits*nchannels);
	inburst=maxlen;
	if(ioctl(dev,SNDCTL_DSP_GETBLKSIZE,&inburst)<0)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"waveCaptureInit(): can't determine DMA buffer length.\n");
#endif	/* CHECKED_BUILD */
		goto init_failed;
	}
	buf_len=(maxlen < inburst) ? (u32_t)maxlen : (u32_t)inburst;
	buffer=(char*)memAlloc(buf_len);
	if(!buffer)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveCaptureInit(): can't allocate %u bytes.\n",buf_len);
#endif	/* CHECKED_BUILD */
		goto init_failed;
	}
	dsp_rate=(int)freq;
    if(ioctl(dev,SNDCTL_DSP_SPEED,&dsp_rate)<0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveCaptureInit(%u,%u,%u): can't set rate %u Hz.\n",
			freq,nbits,nchannels,freq);
#endif	/* CHECKED_BUILD */
		goto init_failed;
	}
	dsp_channels=(int)(nchannels-1);
	if(ioctl(dev,SNDCTL_DSP_STEREO,&dsp_channels)<0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveCaptureInit(%u,%u,%u): can't setup for %s input.\n",
			freq,nbits,nchannels,nchannels==1 ? "mono" : "stereo");
#endif	/* CHECKED_BUILD */
		goto init_failed;
	}
#ifdef	USE_SAMPLESIZE
	dsp_samplesize=(int)nbits;
	if(ioctl(devn,SNDCTL_DSP_SAMPLESIZE,&dsp_samplesize)<0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveCaptureInit(%u,%u,%u): can't set %u bits per sample.\n",
			freq,nbits,nchannels,nbits);
#endif	/* CHECKED_BUILD */
		goto init_failed;
	}
#else	/* !USE_SAMPLESIZE */
	switch(nbits)
	{
		case 16:
			fmt = AFMT_S16_LE;
      		break;
		case 8:
			fmt = AFMT_U8;
			break;
		default:
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_WARNING,"waveCaptureInit(%u,%u,%u): "
				"strange sample size '%u'.\n",
				freq,nbits,nchannels,nbits);
#endif		/* CHECKED_BUILD */
			fmt = (int)nbits;
			break;
	}
	if(ioctl(dev,SNDCTL_DSP_SETFMT,&fmt)<0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"waveCaptureInit(%u,%u,%u): can't set %u bits per sample.\n",
			freq,nbits,nchannels,nbits);
#endif	/* CHECKED_BUILD */
		goto init_failed;
	}
#endif	/* !USE_SAMPLESIZE */
	/* Set up capture thread */
	initialized=true;
	recording=false;
	pause_cond=condCreate(false);
	condReset(pause_cond);
	accum=rwlCreate();
	rwlBlockMode(accum,true);
	cap_thread=threadCreate(waveCaptureThread,NULL);
	return true;
init_failed:
	waveClean();
	return false;
}

vboolean	V_GLOBAL
waveCaptureResume(void)
{
	if(!initialized)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"waveCaptureResume(): capture was not initialized.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	if(recording)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,
			"waveCaptureResume(): capture already started.\n");
#endif	/* CHECKED_BUILD */
		return true;
	}
	recording=true;
	condSet(pause_cond);
	return true;
}

void	V_GLOBAL
waveCapturePause(void)
{
	if(!initialized || !recording)
		return;
	recording=false;
	return;
}

void	V_GLOBAL
waveCaptureShutdown()
{
	if(!initialized)
		return;
	waveClean();
	return;
}

#endif	/* defined(__linux) */

u32_t	V_GLOBAL
waveCaptureLength(void)
{
	if(recording)
		return rwlCount(accum);
	return 0;
}

vboolean	V_GLOBAL
waveCaptureBlockMode(vboolean blocking)
{
	if(initialized && !recording)
	{
		rwlBlockMode(accum,blocking);
		return true;
	}
#ifdef	CHECKED_BUILD
	if(!initialized)
		conlPuts(LOG_ERROR,
			"waveCaptureBlockMode(): capture was not initialized.\n");
	else
		conlPuts(LOG_ERROR,
			"waveCaptureBlockMode(): capture already started.\n");
#endif	/* CHECKED_BUILD */
	return false;
}

u32_t		V_GLOBAL
waveCaptureGet(void *dest_buf,u32_t requested_bytes)
{
	if(recording)
		return rwlPopBuffer(accum,dest_buf,requested_bytes);
	return 0;
}

/*</pre></strong></body></html>*/
/*EOF*/
