/*
 * fsbase.c: implementation of base filesystem object.
 */

#include <v_lib.h>
#include <../source/v_mt.h>
#include <../source/vfs/iobase.h>
#include <../source/vfs/fsbase.h>
#include <../source/vfs/vfsutil.h>

#ifdef	CHECKED_BUILD
#define	VFS_MAGIC				0x524F4653	/* ROFS */
static char	*vfs_object			= "virtual FS";
#endif	/* CHECKED_BUILD */

V_LOCAL	u32_t	V_CALL
vfs_generic_addref(struct vfs_dev_s *self)
{
	if(!libvIsHandle(self,"VFS->AddRef",vfs_object,VFS_MAGIC))
		return 0;
	return ++self->refcount;
}

V_LOCAL u32_t	V_CALL
vfs_generic_release(struct vfs_dev_s *self)
{
	if(!libvIsHandle(self,"VFS->Release",vfs_object,VFS_MAGIC))
		return 0;
	if(self->refcount==1)
	{
    	self->refcount=0;
		self->destroy(self);
#ifdef	CHECKED_BUILD
		conlPuts(LOG_TRACE,
        	"VFS->Release(): ref_count=0, FS deleted.\n");
#endif	/* CHECKED_BUILD */
		return 0;
	}
	return --self->refcount;
}

V_LOCAL void	V_CALL
vfs_generic_destroy(struct vfs_dev_s *self)
{
	if(!libvIsHandle(self,"VFS->destroy",vfs_object,VFS_MAGIC))
		return;
	if(self->refcount)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"VFS->destroy(): FS is busy, %d ref(s).\n",self->refcount);
#endif	/* CHECKED_BUILD */
		return;
	}
	memFree(self);
}

/*
 *	Default stub procedures.
 */
#include <../source/argsused.h>
V_LOCAL u32_t	V_CALL
vfs_get_inode_count_stub(struct vfs_dev_s *self)
{
	if(!libvIsHandle(self,"stubfs->get_inode_count",vfs_object,VFS_MAGIC))
		return 0;
	return 0;
}

#include <../source/argsused.h>
V_LOCAL vboolean	V_CALL
vfs_get_file_name_stub(struct vfs_dev_s *self,u32_t inode,char *name_ret)
{
	if(!libvIsHandle(self,"stubfs->get_file_name",vfs_object,VFS_MAGIC))
		return false;
	return false;
}

#include <../source/argsused.h>
V_LOCAL file_t	V_CALL
vfs_create_file_stub(struct vfs_dev_s *self,u32_t inode)
{
	if(!libvIsHandle(self,"stubfs->create_file",vfs_object,VFS_MAGIC))
		return false;
	return NULL;
}

vfs_dev_p	V_CALL
vfsAlloc(u32_t size_of_object)
{
	vfs_dev_t *fs;
	if(size_of_object<sizeof(vfs_dev_t))
		size_of_object=sizeof(vfs_dev_t);
	ALLOC_OBJECT(fs,vfs_dev_t,size_of_object-sizeof(vfs_dev_t),VFS_MAGIC);
	fs->refcount=1;
	fs->addref=vfs_generic_addref;
	fs->release=vfs_generic_release;
	fs->destroy=vfs_generic_destroy;
	fs->get_inode_count=vfs_get_inode_count_stub;
	fs->get_file_name=vfs_get_file_name_stub;
	fs->create_file=vfs_create_file_stub;
	return fs;
}
