#include <v_lib.h>

#include <../source/vfs/iobase.h>
#include <../source/vfs/fbase.h>
#include <../source/vfs/mbase.h>

/*
 *	Implementation of memory-mapped io read-only port
 */

#define	MEMPORT(obj)		((mapped_port_t*)(obj))

V_LOCAL u32_t V_CALL mmp_read(void *object,void *buf,u32_t len)
{
	u32_t avail=MEMPORT(object)->size-MEMPORT(object)->ptr;
	if(avail<len)
		len=avail;
	memCopy(buf,MEMPORT(object)->base+MEMPORT(object)->ptr,len);
	MEMPORT(object)->ptr+=len;
	return len;
}

V_LOCAL u32_t V_CALL mmp_seek(void *object,i32_t offset,int whence)
{
	switch(whence)
	{
		case SEEK_SET:
			/* this cast also handles incorrect negative offsets */
			if((u32_t)offset > MEMPORT(object)->size)
				return INVALID_IO_OFFSET;
			MEMPORT(object)->ptr=(u32_t)offset;
			break;
		case SEEK_CUR:
			if(offset<0)
			{
				if((u32_t)(-offset) > MEMPORT(object)->ptr)
					return INVALID_IO_OFFSET;
			}
			else
			{
				if((u32_t)(offset) > 
					(MEMPORT(object)->size-MEMPORT(object)->ptr) )
					return INVALID_IO_OFFSET;
			}
			MEMPORT(object)->ptr=MEMPORT(object)->ptr + offset;
			break;
		case SEEK_END:
			if((u32_t)(-offset) > MEMPORT(object)->size)
				return INVALID_IO_OFFSET;
			MEMPORT(object)->ptr=MEMPORT(object)->size + offset;
			break;
		default:
			return INVALID_IO_OFFSET;
	}
	return MEMPORT(object)->ptr;
}

V_LOCAL u32_t V_CALL mmp_tell(void *object)
{
	return MEMPORT(object)->ptr;
}

V_LOCAL	int V_CALL mmp_ioctl(void *object,int code,void *parameters)
{
	switch(code)
    {
    	case pctl_get_type:
        	if(!parameters)
            	return -1;
           	*((u32_t*)parameters)=PCTL_BLOCK_DEVICE | PCTL_SUPPORT_MMAP;
            return 0;
		case pctl_map_ispace:
        	if(!parameters)
            	return -1;
           	*((void**)parameters)=MEMPORT(object)->base;
			return 0;
	}
	return -1;
}

V_LOCAL vboolean	V_CALL mmp_close(void *object)
{
	memFree(object);
	return true;
}

#undef	MEMPORT

/*
 *	Factory helper function.
 */
file_t V_CALL
ioAllocMappedPort(u32_t real_port_size,void *mmap_base,u32_t mmap_size)
{
	mapped_port_t *port;
	if(real_port_size<sizeof(mapped_port_t))
	{
		/* Descendant can't be less in size than ancestor */
		return NULL;
	}
	port=(mapped_port_t*)ioAllocPort(real_port_size);
	port->base=(u8_t*)mmap_base;
	port->size=mmap_size;
	port->ptr=0;
	port->vtbl.read=mmp_read;
	port->vtbl.seek=mmp_seek;
	port->vtbl.tell=mmp_tell;
    port->vtbl.close=mmp_close;
	port->vtbl.ioctl=mmp_ioctl;
	return (file_t)port;
}
