#include <v_lib.h>
#include <../source/v_mt.h>

#include <../source/vfs/vfsmap.h>
#include <../source/vfs/vfsutil.h>
#include <../source/v_os.h>

#ifdef		__UNIX__

#include 	<sys/io.h>
#include 	<sys/fcntl.h>
#include 	<sys/stat.h>
#include 	<sys/types.h>
#include 	<sys/mman.h>

#elif		defined(__CYGWIN32__)

#include 	<io.h>
#include 	<sys/fcntl.h>
#include 	<sys/stat.h>
#include 	<sys/types.h>
#include 	<sys/mman.h>

#endif		/* defined(__CYGWIN32__) */

#ifdef		CHECKED_BUILD
#define		MMAP_MAGIC		0x4D4D4150 		/* MMAP */
static char	*mmap_object="memory map";
#endif		/* CHECKED_BUILD */

typedef struct file_map_s
{
#ifdef		CHECKED_BUILD
	u32_t	mmap_magic;
#endif		/* CHECKED_BUILD */
#ifdef		__UNIX__
	int		fd;
#elif		defined(__WIN32__)
	HANDLE	fd;
	HANDLE	md;
#endif
	size_t	len;
	void	*base;
}	file_map_t, *file_map_p;

#define	FILEMAP(handle)		((file_map_t*)(handle))

#ifdef	__UNIX__

mmap_t	V_GLOBAL
mmapOpenRead(char *file_name)
{
	int fd;
	struct stat st;
	void *base;
	file_map_p res;
	fd=open(file_name,0600);
	if(fd==-1)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mmapOpen('%s'): open() failed (code %d).\n",
            dbs(file_name),errno);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	fstat(fd,&st);
	base=mmap(NULL,st.st_size,PROT_READ,MAP_PRIVATE,fd,0);
	if(!base || (int)base==-1)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mmapOpen('%s'): mmap() failed (code %d).\n",file_name,errno);
#endif	/* CHECKED_BUILD */
		close(fd);
		return NULL;
	}
	ALLOC_OBJECT(res,file_map_t,0,MMAP_MAGIC);
	res->fd=fd;
	res->len=st.st_size;
	res->base=base;
	return (mmap_t)res;
}

void	V_GLOBAL
mmapClose(mmap_t handle)
{
	if(!libvIsHandle(handle,"mmapClose",mmap_object,MMAP_MAGIC))
		return;
	munmap(FILEMAP(handle)->base,FILEMAP(handle)->len);
	close(FILEMAP(handle)->fd);
	memFree(handle);
}

#elif	defined(__WIN32__)

mmap_t	V_GLOBAL
mmapOpenRead(char *file_name)
{
	DWORD type,size;
	HANDLE fd, md;
	void *base;
	file_map_p res;
	char os_name[_MAX_PATH];
    if(!file_name)
    {
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"mmapOpenRead(NULL): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
		return NULL;
    }
    vfsUnix2Native(os_name,file_name);
	fd=CreateFile(os_name,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,0,NULL);
	if(fd==INVALID_HANDLE_VALUE)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mmapOpen(%s): CreateFile(%s) failed (code %u).\n",
			file_name,os_name,GetLastError());
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	type=GetFileType(fd);
	if(type!=FILE_TYPE_DISK)
	{
#if 	0
		switch(type)
		{
			case FILE_TYPE_CHAR:
				/* Error: attempt to map character device */
				break;
			case FILE_TYPE_PIPE:
				/* Error: attempt to map pipe */
				break;
			default:
				/* Error: unknown file type */
				break;
		}
#endif	/* 0 */
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"mmapOpen('%s'): can't CreateFileMapping():"
			" isn't regular file.\n",file_name);
#endif	/* CHECKED_BUILD */
		CloseHandle(fd);
		return NULL;
	}
	size=GetFileSize(fd,NULL);
	md=CreateFileMapping(fd,NULL,PAGE_READONLY,0,size,NULL);
	if(md==INVALID_HANDLE_VALUE)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mmapOpen(%s): CreateFileMapping() failed (code %u).\n",
			file_name,GetLastError());
#endif	/* CHECKED_BUILD */
		CloseHandle(fd);
		return NULL;
	}
	base=MapViewOfFile(md,FILE_MAP_READ,0,0,size);
	if(!base)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"mmapOpen(%s): MapViewOfFile() failed (code %u).\n",
			file_name,GetLastError());
#endif	/* CHECKED_BUILD */
		CloseHandle(md);
		CloseHandle(fd);
		return NULL;
	}
	ALLOC_OBJECT(res,file_map_t,0,MMAP_MAGIC);
	res->fd=fd;
	res->md=md;
	res->len=size;
	res->base=base;
	return (mmap_t)res;
}

void	V_GLOBAL
mmapClose(mmap_t handle)
{
	if(!libvIsHandle(handle,"mmapClose",mmap_object,MMAP_MAGIC))
		return;
	if(FILEMAP(handle)->md && FILEMAP(handle)->md!=INVALID_HANDLE_VALUE)
	{
		if(FILEMAP(handle)->base)
			UnmapViewOfFile(FILEMAP(handle)->base);
		CloseHandle(FILEMAP(handle)->md);
	}
	if(FILEMAP(handle)->fd && FILEMAP(handle)->fd!=INVALID_HANDLE_VALUE)
		CloseHandle(FILEMAP(handle)->fd);
	memFree(handle);
}

#endif	/* __WIN32__ */

void_p	V_GLOBAL
mmapBase(mmap_t handle)
{
	if(!libvIsHandle(handle,"mmapClose",mmap_object,MMAP_MAGIC))
		return NULL;
	return FILEMAP(handle)->base;
}

u32_t	V_GLOBAL
mmapSize(mmap_t handle)
{
	if(!libvIsHandle(handle,"mmapClose",mmap_object,MMAP_MAGIC))
		return 0;
	return FILEMAP(handle)->len;
}

#undef	FILEMAP
