/*<html><body bgcolor=#000080><font color=#C0C0B0><strong><pre>*/
#include	"../../include/v_lib.h"
#include	"../../source/v_mt.h"
#include	"../v_os.h"

/*
 *	wglXXX functions declared in wingdi.h; all other
 *	protos for OpenGL functions are in ../include/sgi/gl.h;
 *	so we DON'T need to include extra OpenGL header (we assume
 *	that calling conventions properly set in sgi/gl.h).
 */

/*
 *	Internal GL context.
 */
#ifdef	CHECKED_BUILD
static char			*vgl_ctx_object="VGL context";
static u32_t		gl_magic	= 0x56474C57;
#endif	/* !CHECKED_BUILD */

typedef struct		gl_context_s
{
#ifdef	CHECKED_BUILD
	u32_t			magic;			/* GL_MAGIC */
#endif	/* !CHECKED_BUILD */
	/*				Platform - independent part */
	u32_t			x, y;			/* Position of window */
	u32_t			width,height;	/* Window extents */
	vboolean		fullscreen;		/* == true for FS window */
	u32_t			bpp;			/* bit depth (FS only) */
	u32_t			refresh;		/* refresh rate (FS only) */
	vboolean		bitmap_based;	/* == 1 for offscreen windows */
	vboolean		accelerated;	/* == 1 for HW renderer */
	vboolean		was_selected;	/* == 1 for 2nd and other selections */
	char			*vendor;		/* information about OpenGL driver */
	char			*renderer;
	char			*version;
	u32_t			num_extensions;
	char			**extensions;
	/*	Win32-specific part */
	char			class_name[64];
	WNDCLASS		wc;
	HWND			parent;			/* used for child windows */
	HBITMAP			bmp;			/* used for offscreen windows */
	BITMAPINFOHEADER	bmh;
	void			*bits;
	HPALETTE		pal;			/* used in 8-bpp mode */
	char			title[128];		/* window title */
	HWND			wnd;			/* window handle */
	BOOL			focused;		/* activity flag (focused) */
	HDC				dc;				/* window of bitmap DC */
	HGLRC			rc;				/* OpenGL RC */
	gl_outline_font_t *font;		/* selected font */
	vp_stack_t		fonts;			/* fonts belongs to this context */
	fifo_t			mqueue;			/* message queue */
	/* rest is HACKS ... */
	UINT			last_msg;			/* used to store last msg */
	i32_t			last_mx, last_my;	/* used to store last mousemove */
}	gl_context_t, *gl_context_p;

/*		Utility functions */
V_LOCAL	gl_context_p
#ifdef	CHECKED_BUILD
vglSafeGetCurrentContext(char *caller)
#else	/* !CHECKED_BUILD */
vglSafeGetCurrentContextQuiet(void)
#define	vglSafeGetCurrentContext(caller)	vglSafeGetCurrentContextQuiet()
#endif	/* !CHECKED_BUILD */
{
	gl_context_t *ctx;
	ctx=(gl_context_t*)GetThreadData()->gl_context;
#ifdef	CHECKED_BUILD
	if(!ctx)
		conlPrintf(LOG_ERROR,"%s(): OpenGL not initialized.\n",
			caller);
	else
		if(ctx->magic != gl_magic)
		{
			conlPrintf(LOG_ERROR,
				"%s(): bad current context, magic wrong.\n",
				caller);
			ctx=NULL;
		}
#endif	/* !CHECKED_BUILD */
	return ctx;
}

#define	vglSafeCheckContext(caller,win) \
		((gl_context_p)libvIsHandle(win,caller,"GL context",gl_magic))

V_LOCAL	gl_context_p
vglAllocContext(u32_t x,u32_t y,u32_t width,u32_t height,
	vboolean fullscreen,void *parent,char *title,
	u32_t bpp,u32_t refresh,vboolean bitmap_based)
{
	gl_context_t *ctx=(gl_context_t*)memAlloc(sizeof(gl_context_t));
	memZero(ctx,sizeof(gl_context_t));
#ifdef	CHECKED_BUILD
	ctx->magic=gl_magic;
#endif	/* CHECKED_BUILD */
	ctx->x=x;
	ctx->y=y;
	ctx->width=width;
	ctx->height=height;
	ctx->fullscreen=fullscreen;
	ctx->parent=parent;
	if(title)
		strCopy(ctx->title,title,128);
	ctx->bpp=bpp;
	ctx->refresh=refresh;
	ctx->bitmap_based=bitmap_based;
	return ctx;
}

/*	======================================================================
 *	Common OpenGL stuff.
 */

V_LOCAL	u8_t	threeto8[8] = {0, 0111>>1, 0222>>1, 0333>>1, 0444>>1,
  0555>>1, 0666>>1, 0377 };
V_LOCAL	u8_t	twoto8[4] = {  0, 0x55, 0xaa, 0xff };
V_LOCAL	u8_t	oneto8[2] = {  0, 255 };
V_LOCAL	int 	defaultOverride[13] = { 0, 3, 24, 27, 64, 67, 88, 173,
	181, 236, 247, 164, 91 };

V_LOCAL	PALETTEENTRY defaultPalEntry[20] = {
	{ 0,   0,   0,    0 },
	{ 0x80,0,   0,    0 },
	{ 0,   0x80,0,    0 },
	{ 0x80,0x80,0,    0 },
	{ 0,   0,   0x80, 0 },
	{ 0x80,0,   0x80, 0 },
	{ 0,   0x80,0x80, 0 },
	{ 0xC0,0xC0,0xC0, 0 },

	{ 192, 220, 192,  0 },
	{ 166, 202, 240,  0 },
	{ 255, 251, 240,  0 },
	{ 160, 160, 164,  0 },

	{ 0x80,0x80,0x80, 0 },
	{ 0xFF,0,   0,    0 },
	{ 0,   0xFF,0,    0 },
	{ 0xFF,0xFF,0,    0 },
	{ 0,   0,   0xFF, 0 },
	{ 0xFF,0,   0xFF, 0 },
	{ 0,   0xFF,0xFF, 0 },
	{ 0xFF,0xFF,0xFF, 0 }
};

V_LOCAL	u8_t
ComponentFromIndex(int i,u32_t nbits,u32_t shift)
{
	u8_t val;
	val=(u8_t)(i>>shift);
	switch(nbits)
	{
		case 1:
			val &= 0x1;
			return oneto8[val];
		case 2:
			val &= 0x3;
			return twoto8[val];
		case 3:
			val &= 0x7;
			return threeto8[val];
		default:
			return 0;
	}
}

/*
 *	Uses 'dc' and 'bitmap_based' fields. Modifies 'rc' and 'pal' fields.
 */
vboolean	V_LOCAL
vglInitializeContext(gl_context_t *ctx)
{
	int rep, pixfmt;
	PIXELFORMATDESCRIPTOR pfd;
	ctx->pal=NULL;
	ctx->rc=NULL;
	memZero(&pfd,sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize=sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion=1;
	if(ctx->bitmap_based)
	{
		pfd.cDepthBits=16;
		pfd.cColorBits=24;
		pfd.dwFlags=PFD_DRAW_TO_BITMAP|PFD_SUPPORT_OPENGL;
	}
	else
	{
		pfd.cDepthBits=(unsigned char)ctx->bpp;
		pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
	}
	pfd.iPixelType=PFD_TYPE_RGBA;
	pfd.iLayerType=PFD_MAIN_PLANE;
	pixfmt=ChoosePixelFormat(ctx->dc,&pfd);
	if(!pixfmt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateContext(): ChoosePixelFormat() failed (code %d).\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		return false;
	}
	for(rep=4;rep;rep--)
	{
		if(SetPixelFormat(ctx->dc,pixfmt,&pfd))
			break;
		Sleep(1000);
	}
	if(!rep)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglCreateContext(): SetPixelFormat() failed"
			" (code %d, OpenGL %d).\n",GetLastError(),glGetError());
#endif	/* CHECKED_BUILD */
		return false;
	}
	DescribePixelFormat(ctx->dc,pixfmt,sizeof(pfd),&pfd);
	if(pfd.dwFlags & PFD_NEED_PALETTE)
	{
		unsigned int i, n;
		LOGPALETTE *lp;
		n = 1 << pfd.cColorBits;
		lp=(PLOGPALETTE)memAlloc(sizeof(LOGPALETTE)+n*sizeof(PALETTEENTRY));
		lp->palVersion=0x300;
		lp->palNumEntries=(unsigned short)n;
		for(i=0;i<n;i++)
		{
			lp->palPalEntry[i].peRed =
				ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
			lp->palPalEntry[i].peGreen =
				ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
			lp->palPalEntry[i].peBlue =
				ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
			lp->palPalEntry[i].peFlags = 0;
		}
		if((pfd.cColorBits==8)&&(pfd.cRedBits==3)&&(pfd.cRedShift==0)&&
			(pfd.cGreenBits==3)&&(pfd.cGreenShift==3)&&
			(pfd.cBlueBits == 2)&&(pfd.cBlueShift==6))
		{
			for(i=1;i<=12;i++)
				lp->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
		}
		ctx->pal = CreatePalette(lp);
		memFree(lp);
		SelectPalette(ctx->dc,ctx->pal,FALSE);
		RealizePalette(ctx->dc);
	}
	ctx->rc=wglCreateContext(ctx->dc);
	if(!ctx->rc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateContext(): wglCreateContext failed (code %d).\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		if(ctx->pal)
		{
			SelectPalette(ctx->dc,GetStockObject(DEFAULT_PALETTE),FALSE);
			DeleteObject(ctx->pal);
			ctx->pal=NULL;
		}
		return false;
	}
	/* Other non-critical operations */
#ifndef	PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED     0x00001000
#endif	/* PFD_GENERIC_ACCELERATED */
	/* BUGBUG !!! __asm int 3 */
	ctx->accelerated=(pfd.dwFlags & PFD_GENERIC_FORMAT) == 
		(pfd.dwFlags & PFD_GENERIC_ACCELERATED);
	/* Initialize font cache */
	vpsInit(&ctx->fonts);
	fifoInit(&ctx->mqueue,sizeof(input_record_t));
	ctx->font=NULL;
	return true;
}

void	V_LOCAL	vglReleaseFont(gl_outline_font_t *fnt);

/*
 *	Uses and modifies 'rc' and 'pal' fields.
 */
void	V_LOCAL
vglFinalizeContext(gl_context_t *ctx)
{
	u32_t i,num;
	gl_outline_font_t *f;
	thread_data_t *td;
#ifdef	CHECKED_BUILD
	conlPrintf(LOG_INFO,"vglFinalizeContext(): %s OpenGL on %s ver. %s\n",
		ctx->vendor,ctx->renderer,ctx->version);
#endif	/* CHECKED_BUILD */
	/* Delete all fonts */
	num=vpsCount(&ctx->fonts);
	for(i=0;i<num;i++)
	{
		f=(gl_outline_font_t*)vpsGet(&ctx->fonts,i);
		if(f)
		{
#ifdef		CHECKED_BUILD
			/* There is nothing wrong, but we must notify caller */
			if(f->ref)
				/* font was created implicitly and there was no 
				 * way for caller to delete it */
				conlPrintf(LOG_TRACE,"vglDeleteContext(): duplicated "
					"font at slot %d have been released.\n",i);
			else
				/* caller must be more accurate ! */
				conlPrintf(LOG_WARNING,
					"vglDeleteContext(): non-empty font slot %d.\n",i);
#endif	/* CHECKED_BUILD */
			vglReleaseFont(f);
		}
	}
	/* Make no context current */
	td=GetThreadData();
	if(td->gl_context == ctx)
		vglMakeCurrent(NULL);
	if(ctx->rc)
	{
		wglDeleteContext(ctx->rc);
		ctx->rc=NULL;
	}
	/* Delete palette */
	if(ctx->pal)
	{
		SelectPalette(ctx->dc,GetStockObject(DEFAULT_PALETTE),FALSE);
		DeleteObject(ctx->pal);
		ctx->pal=NULL;
	}
	/* free strdup()'ed memory */
	if(ctx->vendor)
		free(ctx->vendor);
	if(ctx->renderer)
		free(ctx->renderer);
	if(ctx->version)
		free(ctx->version);
	if(ctx->extensions)
	{
		for(i=0;i<ctx->num_extensions;i++)
			free(ctx->extensions[i]);
		free(ctx->extensions);
	}
}

/*
 *	Translates wParam and lParam into LibV key code.
 */
keycode_t	V_LOCAL
keycodeWin32Code(u32_t wp,u32_t lp)
{
	keycode_t key;
	unsigned char scan=(unsigned char)((lp>>16) & 0xFF);
	unsigned char ext=(unsigned char)((lp & (1<<24)) != 0);
	switch(wp)
	{
		case VK_ESCAPE:	key=key_escape; break;
		case '1':		key=key_1; break;
		case '2':		key=key_2; break;
		case '3':		key=key_3; break;
		case '4':		key=key_4; break;
		case '5':		key=key_5; break;
		case '6':		key=key_6; break;
		case '7':		key=key_7; break;
		case '8':		key=key_8; break;
		case '9':		key=key_9; break;
		case '0':		key=key_0; break;
		case 189:		key=key_minus; break;
		case 187:		key=key_equal; break;
		case VK_BACK:	key=key_backspace; break;
		case VK_TAB:	key=key_tab; break;
		case 'Q':		key=key_Q; break;
		case 'W':		key=key_W; break;
		case 'E':		key=key_E; break;
		case 'R':		key=key_R; break;
		case 'T':		key=key_T; break;
		case 'Y':		key=key_Y; break;
		case 'U':		key=key_U; break;
		case 'I':		key=key_I; break;
		case 'O':		key=key_O; break;
		case 'P':		key=key_P; break;
		case 219:		key=key_left_bracket; break;
		case 221:		key=key_right_bracket; break;
		case VK_RETURN:	key=ext ? key_grey_enter : key_enter; break;
		case 17:		key=
			ext ? key_right_control : key_left_control; break;
		case 18:		key=ext ? key_right_alt : key_left_alt; break;
		case 'A':		key=key_A; break;
		case 'S':		key=key_S; break;
		case 'D':		key=key_D; break;
		case 'F':		key=key_F; break;
		case 'G':		key=key_G; break;
		case 'H':		key=key_H; break;
		case 'J':		key=key_J; break;
		case 'K':		key=key_K; break;
		case 'L':		key=key_L; break;
		case 186:		key=key_colon; break;
		case 222:		key=key_apostrophe; break;
		case 192:		key=key_back_quote; break;
		case 16:		key=
			(scan==42) ? key_left_shift : key_right_shift; break;
		case 220:		key=key_back_slash; break;
		case 'Z':		key=key_Z; break;
		case 'X':		key=key_X; break;
		case 'C':		key=key_C; break;
		case 'V':		key=key_V; break;
		case 'B':		key=key_B; break;
		case 'N':		key=key_N; break;
		case 'M':		key=key_M; break;
		case 188:		key=key_comma; break;
		case 190:		key=key_dot; break;
		case 191:		key=key_slash; break;
		case VK_F1:		key=key_f1; break;
		case VK_F2:		key=key_f2; break;
		case VK_F3:		key=key_f3; break;
		case VK_F4:		key=key_f4; break;
		case VK_F5:		key=key_f5; break;
		case VK_F6:		key=key_f6; break;
		case VK_F7:		key=key_f7; break;
		case VK_F8:		key=key_f8; break;
		case VK_F9:		key=key_f9; break;
		case VK_F10:	key=key_f10; break;
		case VK_F11:	key=key_f11; break;
		case VK_F12:	key=key_f12; break;
		case VK_NUMLOCK:	key=key_numlock; break;
		case VK_SCROLL:	key=key_scrolllock; break;
		case 32:		key=key_space; break;
		case 20:		key=key_capslock; break;
		case 91:		key=key_left_win; break;
		case 92:		key=key_right_win; break;
		case 93:		key=key_context; break;
		case 111:		key=key_grey_slash; break;
		case 106:		key=key_grey_asterisk; break;
		case 109:		key=key_grey_minus; break;
		case 107:		key=key_grey_plus; break;
		default:
			switch(scan)
			{
				case 71:	key=(ext) ? key_home : key_numpad_7; break;
				case 72:	key=(ext) ? key_up : key_numpad_8; break;
				case 73:	key=(ext) ? key_pageup : key_numpad_9; break;
				case 75:	key=(ext) ? key_left : key_numpad_4; break;
				case 76:	key=key_numpad_5; break;
				case 77:	key=(ext) ? key_right : key_numpad_6; break;
				case 79:	key=(ext) ? key_end : key_numpad_1; break;
				case 80:	key=(ext) ? key_down : key_numpad_2; break;
				case 81:	key=(ext) ? key_pagedown : key_numpad_3; break;
				case 82:	key=(ext) ? key_insert : key_numpad_0; break;
				case 83:	key=(ext) ? key_delete : key_numpad_dot; break;
				case 69:	key=key_pause; break;
				case 55:	key=key_printscreen; break;
				default:	key=key_unknown; break;
			}
		break;
	}
	return key;
}

vboolean V_LOCAL
wm_process_message(gl_context_t *ctx,UINT msg,WPARAM wparam,
	LPARAM lparam,thread_data_t *td)
{
	input_record_t ev;
	vboolean mouse_event=false;
	switch(msg)
	{
		case WM_SYSKEYDOWN:
		case WM_KEYDOWN:
#ifndef		ENABLE_AUTOREPEAT
			if(lparam & 0x40000000)
				return true;
#endif		/* ENABLE_AUTOREPEAT */
			ev.event=ev_key_down;
			ev.first_code=keycodeWin32Code(wparam,lparam);
			switch(ev.first_code)
			{
				case key_left_shift:
				case key_right_shift:
					td->lower_case=false;
					break;
				case key_capslock:
					td->caps_lock=GetKeyState(20) & 1;
					break;
				case key_numlock:
					td->num_lock=GetKeyState(VK_NUMLOCK) & 1;
					break;
				case key_space:
					if(GetAsyncKeyState(VK_SPACE) & 0x8000)
						SetFocus(NULL);
			}
			ev.second_code=keycodeGetChar((keycode_t)ev.first_code,
				td->caps_lock ? !td->lower_case : td->lower_case,NULL);
			break;
		case WM_KEYUP:
		case WM_SYSKEYUP:
			ev.event=ev_key_up;
			ev.first_code=keycodeWin32Code(wparam,lparam);
			ev.second_code=0;
			switch(ev.first_code)
			{
				/* TODO: workaround Win32 bug with shifts 
				   (see Descent 3 readme for details) */
				case key_left_shift:
					td->lower_case=true;
					/* if(!(GetAsyncKeyState(VK_SHIFT) & 0x8000)) */
					/* Generate key_up + key_right_shift event (NOT WORKING !) */
					/* ev.first_code=key_right_shift ... */
					fifoPush(&ctx->mqueue,&ev);
					return true;
				case key_right_shift:
					td->lower_case=true;
					/* if(!(GetAsyncKeyState(VK_SHIFT) & 0x8000)) */
					/* Generate key_up + key_left_shift (NOT WORKING !) */
					/* ev.first_code=key_left_shift ... */
					fifoPush(&ctx->mqueue,&ev);
					return true;
			}
			break;
		case WM_MOUSEMOVE:
			ev.event=ev_mouse_move;
			ev.first_code=LOWORD(lparam);	/* TODO: SIGN_EXTEND(...) */
			ev.second_code=HIWORD(lparam);
			if(ev.first_code & 0x8000)
				ev.first_code|=0xFFFF0000;
			if(ev.second_code & 0x8000)
				ev.second_code|=0xFFFF0000;
#ifndef		DISABLE_MOUSE_FILTERING
			if(ctx->last_mx==ev.first_code &&
				ctx->last_my==ev.second_code)
				return true;
			ctx->last_mx=ev.first_code;
			ctx->last_my=ev.second_code;
#endif		/* !DISABLE_MOUSE_FILTERING */
			break;
		case WM_LBUTTONDOWN:
			ev.event=ev_key_down;
			ev.first_code=key_left_mouse;
			ev.second_code=0;
			mouse_event=true;
			break;
		case WM_MBUTTONDOWN:
			ev.event=ev_key_down;
			ev.first_code=key_middle_mouse;
			ev.second_code=0;
			mouse_event=true;
			break;
		case WM_RBUTTONDOWN:
			ev.event=ev_key_down;
			ev.first_code=key_right_mouse;
			ev.second_code=0;
			mouse_event=true;
			break;
		case WM_LBUTTONUP:
			ev.event=ev_key_up;
			ev.first_code=key_left_mouse;
			ev.second_code=0;
			mouse_event=true;
			break;
		case WM_MBUTTONUP:
			ev.event=ev_key_up;
			ev.first_code=key_middle_mouse;
			ev.second_code=0;
			mouse_event=true;
			break;
		case WM_RBUTTONUP:
			ev.event=ev_key_up;
			ev.first_code=key_right_mouse;
			ev.second_code=0;
			mouse_event=true;
			break;
		default:
			return false;
	}
	fifoPush(&ctx->mqueue,&ev);
	if(ctx->parent)
		PostMessage(ctx->parent,msg,wparam,lparam);
	/*
	 *	Post mouse_move event after mouse press/release events.
	 */
	if(mouse_event)
	{
		if(ev.event==ev_key_down)
		{
			SetCapture(ctx->wnd);
			SetFocus(ctx->wnd);	/* Is this still needed ? */
		}
		else
			ReleaseCapture();
		ev.event=ev_mouse_move;
		ev.first_code=LOWORD(lparam);
		ev.second_code=HIWORD(lparam);
		if(ev.first_code & 0x8000)
			ev.first_code|=0xFFFF0000;
		if(ev.second_code & 0x8000)
			ev.second_code|=0xFFFF0000;
		fifoPush(&ctx->mqueue,&ev);
	}
	return true;
}

/*	======================================================================
 *	Callback for all OpenGL windows
 */
LRESULT CALLBACK
gl_window_proc(HWND wnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	gl_context_t *ctx;
	input_record_t ev;
	PAINTSTRUCT ps;
	ctx=(gl_context_t*)GetWindowLong(wnd,GWL_USERDATA);
	/* Supress warning message */
	if(!ctx || !vglSafeCheckContext("gl_window_proc",(glw_t)ctx))
		return DefWindowProc(wnd,msg,wparam,lparam);
	switch(msg)
	{
		case WM_ACTIVATE:
			ctx->focused=LOWORD(wparam) != WA_INACTIVE;
			/* Extra stuff for fullscreen windows */
			if(ctx->fullscreen)
			{
				/* TODO: implement video mode switching */
				if(ctx->focused)
					SetWindowPos(ctx->wnd,HWND_TOP,0,0,0,0,
						SWP_NOSIZE|SWP_NOMOVE);
				else
					SetWindowPos(ctx->wnd,HWND_NOTOPMOST,0,0,0,0,
						SWP_NOSIZE|SWP_NOMOVE);
			}
			break;
		case WM_CLOSE:
			break;
		case WM_PAINT:
			BeginPaint(ctx->wnd,&ps);
			EndPaint(wnd,&ps);
			if(ctx->last_msg!=WM_PAINT)
			{
				ev.event=ev_os_dependent;
				ev.first_code=WM_PAINT;
				ev.second_code=0;
				fifoPush(&ctx->mqueue,&ev);
				if(ctx->parent)
					PostMessage(ctx->parent,msg,wparam,lparam);
			}
			break;
		case WM_MOVE:
		{
			ctx->x=(u32_t)LOWORD(lparam);
			ctx->y=(u32_t)HIWORD(lparam);
			ev.event=ev_os_dependent;
			ev.first_code=WM_MOVE;
			ev.second_code=0;
			fifoPush(&ctx->mqueue,&ev);
			if(ctx->parent)
				PostMessage(ctx->parent,msg,wparam,lparam);
			break;
		}
		case WM_SIZE:
		{
			ctx->width=(u32_t)LOWORD(lparam);
			ctx->height=(u32_t)HIWORD(lparam);
			ev.event=ev_os_dependent;
			ev.first_code=WM_SIZE;
			ev.second_code=0;
			fifoPush(&ctx->mqueue,&ev);
			if(ctx->parent)
				PostMessage(ctx->parent,msg,wparam,lparam);
			break;
		}
		case WM_PALETTECHANGED:
		{
			if(wnd!=(HWND)wparam)
			{
				SelectPalette(ctx->dc,ctx->pal,FALSE);
				RealizePalette(ctx->dc);
			}
			break;
		}
		case WM_QUERYNEWPALETTE:
		{
			SelectPalette(ctx->dc,ctx->pal,FALSE);
			RealizePalette(ctx->dc);
			return TRUE;
		}
		default:
		{
			if(wm_process_message(ctx,msg,wparam,lparam,GetThreadData()))
				break;
			return DefWindowProc(wnd,msg,wparam,lparam);
		}
	}
	ctx->last_msg=msg;
	return 0;
}

/*
 *	Registers OpenGL window class.
 */
vboolean	V_LOCAL
vglRegisterWindowClass(gl_context_t *ctx)
{
	u32_t context_count;
	global_data_t *gd=libvLockGlobal();
	context_count=gd->context_count+1;
	gd->context_count=context_count;
	libvUnlockGlobal();
	wsprintf(ctx->class_name,"GLWC%X",context_count);
	memZero(&ctx->wc,sizeof(WNDCLASS));
	ctx->wc.lpszClassName=ctx->class_name;
	ctx->wc.hInstance=GetModuleHandle(NULL);
	ctx->wc.style = CS_CLASSDC	| CS_NOCLOSE;
	ctx->wc.lpfnWndProc=gl_window_proc;
	ctx->wc.hIcon = LoadIcon(NULL,IDI_WINLOGO);
	ctx->wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	ctx->wc.hbrBackground = GetStockObject(HOLLOW_BRUSH);
	ctx->wc.lpszMenuName=NULL;
	if(RegisterClass(&ctx->wc))
		return true;
#ifdef	CHECKED_BUILD
	conlPrintf(LOG_ERROR,
		"vglRegisterWindowClass(): RegisterClass() failed (code %d).\n",
		GetLastError());
#endif	/* CHECKED_BUILD */
	return false;
}

/*
 *	Changes screen resolution, display depth and, under NT,
 *	display refresh rate.
 */
vboolean	V_LOCAL
vidSetMode(int width,int height,int bpp,int refresh_rate)
{
	DEVMODE dm;
	LONG retval;
	memZero(&dm,sizeof(dm));
	dm.dmSize=sizeof(dm);
	dm.dmPelsWidth=width;
	dm.dmPelsHeight=height;
	dm.dmBitsPerPel=bpp;
	dm.dmFields=DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	if(refresh_rate>0)
	{
		dm.dmDisplayFrequency=refresh_rate;
		dm.dmFields|=DM_DISPLAYFREQUENCY;
	}
	retval=ChangeDisplaySettings(&dm,0);
#ifndef	CHECKED_BUILD
	return retval==DISP_CHANGE_SUCCESSFUL;
#else	/* CHECKED_BUILD */
	switch(retval)
	{
		case DISP_CHANGE_SUCCESSFUL:
			conlPuts(LOG_INFO,"success.\n");
			return true;
		case DISP_CHANGE_RESTART:
			conlPuts(LOG_INFO,"failed, restart required.\n");
			break;
		case DISP_CHANGE_BADMODE:
			conlPuts(LOG_INFO,"failed, bad mode.\n");
			break;
		case DISP_CHANGE_BADFLAGS:
			conlPuts(LOG_INFO,"failed, bad flags.\n");
			break;
		case DISP_CHANGE_FAILED:
			conlPuts(LOG_INFO,"failed, driver error.\n");
			break;
		default:
			conlPrintf(LOG_INFO,"failed, error code %d.\n",retval);
			break;
	}
	return false;
#endif	/* CHECKED_BUILD */
}

/*
 *	Sets default display settings.
 */
void	V_LOCAL
vidDefaultMode()
{
	/*
	BUGBUG
	SwitchToGraphicsMode !!!
	I can't make it work from full-screen text mode. Sorry.
	*/
	ChangeDisplaySettings(NULL,0);
}

/*
 *	Creates OpenGL windowed context.
 *	Returns context on success, NULL on error.
 */
gl_context_p V_LOCAL
vglCreateWindowedContext(gl_context_t *ctx)
{
	u32_t sx, sy;
	thread_data_t *td;
	if(!vglRegisterWindowClass(ctx))
	{
		vglDestroyWindow((glw_t)ctx);
		return NULL;
	}
	if(ctx->fullscreen)
	{
		HDC refdc;
		u32_t sbpp;
		u32_t refresh;
		refdc=GetDC(NULL);
		sbpp=GetDeviceCaps(refdc,BITSPIXEL);
		refresh=GetDeviceCaps(refdc,VREFRESH);
		ReleaseDC(NULL,refdc);
		sx=GetSystemMetrics(SM_CXSCREEN);
		sy=GetSystemMetrics(SM_CYSCREEN);
		ctx->wnd=CreateWindow(
			ctx->class_name,
			ctx->title,
			WS_POPUP,
			0,0,
			sx,sy,
			NULL,
			NULL,
			GetModuleHandle(NULL),
			NULL);
		ShowWindow(ctx->wnd,SW_SHOWNORMAL);
		if(!ctx->wnd)
		{
			/* Can't create window */
#ifdef		CHECKED_BUILD
				conlPrintf(LOG_ERROR,"vglCreateWindowedContext(): "
					"CreateWindow() failed (code %d).\n",GetLastError());
#endif		/* CHECKED_BUILD */
			memFree(ctx);
			return NULL;
		}
		ctx->dc=GetDC(ctx->wnd);
		/*	under Chicago we can't change bpp; under Nashvile and
		 *	Memphis we can, but can't set refresh rate ...
		 */
		if(ctx->bpp==0 || WeAreUnderMustDie())
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vglCreateWindowedContext(): %susing desktop display "
				"depth %d.\n",
				WeAreUnderMustDie() ? "Windows 95/98 detected, " : "",
				sbpp);
#endif		/* CHECKED_BUILD */
			ctx->bpp=sbpp;
		}
		if(WeAreUnderMustDie())
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vglCreateWindowedContext(): can't set change refresh "
				"rate (%d Hz) under Windows 9x.\n",ctx->refresh);
#endif		/* CHECKED_BUILD */
			ctx->refresh=0;
		}
		SetWindowLong(ctx->wnd,GWL_USERDATA,(LONG)ctx);
		if(ctx->width<=0 || ctx->height<=0)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vglCreateWindowedContext(): using desktop dimensions "
				"%d by %d.\n",sx,sy);
#endif		/* CHECKED_BUILD */
			ctx->width=sx;
			ctx->height=sy;
		}
		if(sbpp!=ctx->bpp || sx!=ctx->width || sy!=ctx->height)
		{
			/* Change screen resolution */
#ifdef		CHECKED_BUILD
			if((ctx->width>sx || ctx->height>sy) && ctx->refresh>=refresh)
				conlPrintf(LOG_WARNING,
					"vglCreateWindowedContext(): refresh rate %dHz may "
					"be to high (current is %dHz).",ctx->refresh,refresh);
			conlPrintf(LOG_INFO,
				"vglCreateWindowedContext(): "
				"calling CDS(%d,%d,%d,%d Hz) ... ",
				ctx->width,ctx->height,ctx->bpp,ctx->refresh);
#endif		/* CHECKED_BUILD */
			if(!vidSetMode(ctx->width,ctx->height,ctx->bpp,ctx->refresh))
			{
				/* Stay quiet: vidSetMode() already report error */
				DestroyWindow(ctx->wnd);
				memFree(ctx);
				return NULL;
			}
			/* Resize window */
			SetWindowPos(ctx->wnd,HWND_NOTOPMOST,0,0,ctx->width,ctx->height,
				SWP_NOZORDER);
		}
	}
	else	/* top-level/child (not fullscreen) */
	{
		if(ctx->parent)
		{
			/* Create child window */
			ctx->wnd=CreateWindow(
				ctx->class_name,
				ctx->title,
				WS_CHILD /* | WS_CLIPCHILDREN | WS_CLIPSIBLINGS */,
				ctx->x,ctx->y,
				ctx->width,ctx->height,
				ctx->parent,
				NULL,
				GetModuleHandle(NULL),
				NULL);
		}
		else /* top-level window */
		{
			/* Create top-level window without sizing border */
			/* @&%# !!! damned user32.dll !!! */
			/*
			if(ctx->wnd && ctx->width<sx && ctx->height<sy)
			{
				RECT rc;
				u32_t rw, rh;
				GetClientRect(ctx->wnd,&rc);
				rw=2*ctx->width-rc.right;
				rh=2*ctx->height-rc.bottom;
				
					works only if started from cmd.exe !!! DAMN
				
				SetWindowPos(ctx->wnd,HWND_NOTOPMOST,0,0,rw,rh,SWP_NOMOVE);
			}
			*/
			/* Workaround: */
			ctx->wnd=CreateWindow(
				ctx->class_name,
				ctx->title,
				WS_POPUP | WS_CAPTION /* | WS_CLIPCHILDREN | WS_CLIPSIBLINGS */,
				ctx->x,ctx->y,
				ctx->width + 2 * GetSystemMetrics(SM_CXFIXEDFRAME),
				ctx->height + 2 * GetSystemMetrics(SM_CYFIXEDFRAME) +
					GetSystemMetrics(SM_CYCAPTION),
				NULL,
				NULL,
				GetModuleHandle(NULL),
				NULL);
		}
		if(!ctx->wnd)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_INFO,
				"vglCreateWindowedContext(): "
				"CreateWindow() returns error '%d'.\n",
				GetLastError());
#endif	/* CHECKED_BUILD */
			memFree(ctx);
			return NULL;
		}
		ctx->dc=GetDC(ctx->wnd);
		SetWindowLong(ctx->wnd,GWL_USERDATA,(LONG)ctx);
		ShowWindow(ctx->wnd,SW_SHOWNORMAL);
	}
	SetActiveWindow(ctx->wnd);
	/* clear window */
	PatBlt(ctx->dc,0,0,ctx->width,ctx->height,BLACKNESS);
	if(!vglInitializeContext(ctx) || !vglMakeCurrent((glw_t)ctx))
	{
		vglDestroyWindow((glw_t)ctx);
		return NULL;
	}
	td=GetThreadData();
	td->lower_case=true;
	td->caps_lock=GetKeyState(20) & 1;
	td->num_lock=GetKeyState(VK_NUMLOCK) & 1;
	return ctx;
}

glw_t	V_GLOBAL
vglCreateWindow(char *title,u32_t x,u32_t y,u32_t width,u32_t height)
{
	gl_context_t *ctx;
	ctx=vglAllocContext(x,y,width,height,false,NULL,title,0,0,false);
	return (glw_t)vglCreateWindowedContext(ctx);
}

glw_t	V_GLOBAL
vglCreateFullscreenWindow(char *title,u32_t width,u32_t height,u32_t bpp,u32_t refresh)
{
	gl_context_t *ctx;
	ctx=vglAllocContext(0,0,width,height,true,NULL,title,
		bpp,refresh,false);
	return (glw_t)vglCreateWindowedContext(ctx);
}

glw_t	V_GLOBAL
vglCreateChildWindow(char *title,u32_t x,u32_t y,u32_t width,u32_t height,void *parent)
{
	gl_context_t *ctx;
	ctx=vglAllocContext(x,y,width,height,false,parent,title,0,0,false);
	return (glw_t)vglCreateWindowedContext(ctx);
}

glw_t	V_GLOBAL
vglCreateMappedWindow(u32_t width,u32_t height)
{
	gl_context_t *ctx;
	ctx=vglAllocContext(0,0,width,height,false,NULL,NULL,0,0,true);
	/*
	 *	For bitmap - based contexts, we must only create
	 *	output bitmap.
	 */
	ctx->bmh.biSize=sizeof(BITMAPINFOHEADER);
	ctx->bmh.biWidth=ctx->width;
	ctx->bmh.biHeight=ctx->height;
	ctx->bmh.biPlanes=1;
	ctx->bmh.biBitCount=24;
	ctx->bmh.biCompression=BI_RGB;
	ctx->dc=CreateCompatibleDC(NULL);
	ctx->bmp=CreateDIBSection(ctx->dc,(LPBITMAPINFO)&ctx->bmh,
		DIB_RGB_COLORS,&ctx->bits,NULL,0);
	SelectObject(ctx->dc,ctx->bmp);
	if(!vglInitializeContext(ctx) || !vglMakeCurrent((glw_t)ctx))
	{
		vglDestroyWindow((glw_t)ctx);
		return NULL;
	}
	return (glw_t)ctx;
}

void	V_GLOBAL
vglDestroyWindow(glw_t win)
{
	gl_context_t *ctx;
	ctx=(gl_context_t*)libvIsHandle(win,"vglDestroyWindow",vgl_ctx_object,
		gl_magic);
	if(!ctx)
		return;
	vglFinalizeContext(ctx);
	if(ctx->fullscreen)
		vidDefaultMode();
	if(ctx->dc)
		ReleaseDC(ctx->wnd,ctx->dc);
	if(ctx->wnd)
		DestroyWindow(ctx->wnd);
	if(!ctx->bmp)
		UnregisterClass(ctx->wc.lpszClassName,ctx->wc.hInstance);
	else
		DeleteObject(ctx->bmp);
	memFree(ctx);
}

vboolean	V_GLOBAL
vglMakeCurrent(glw_t win)
{
	gl_context_t *ctx;
	thread_data_t *td;
	if(!win)
	{
		SetFocus(NULL);
		wglMakeCurrent(NULL,NULL);
		td=GetThreadData();
		td->gl_context=NULL;
		return true;
	}
	ctx=vglSafeCheckContext("vglMakeCurrent",win);
	if(!ctx)
		return false;
	/*	DAMN. SOMETHING IS WRONG HERE. I CANT FIX THE BUG, SORRY ...
	 *		Under NT 4.0 wglMakeCurrent() crashes the program if
	 *	called twice from thread other then current (exception in
	 *	'NtWaitForMultipleObjects()'). Under Win98 and NT 5.0 - all Ok.
	 */
	/*
	 *	TODO: catch exception.
	 */
	if(wglMakeCurrent(ctx->dc,ctx->rc)==0)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglMakeCurrent(): wglMakeCurrent() failed (code %d).\n",
			GetLastError());
#endif	/* CHECKED_BUILD */
		return false;
	}
	td=GetThreadData();
	td->gl_context=(void*)ctx;
	if(!ctx->was_selected)
	{
		int i;
		char *exts;
		ctx->was_selected=true;
		ctx->vendor=(char*)glGetString(GL_VENDOR);
		if(ctx->vendor)
			ctx->vendor=strdup((const char*)ctx->vendor);
		ctx->renderer=(char*)glGetString(GL_RENDERER);
		if(ctx->renderer)
			ctx->renderer=strdup((const char*)ctx->renderer);
		ctx->version=(char*)glGetString(GL_VERSION);
		if(ctx->version)
			ctx->version=strdup((const char*)ctx->version);
		ctx->extensions=NULL;
		ctx->num_extensions=0;
		exts=(char*)glGetString(GL_EXTENSIONS);
		if(exts)
		{
			parser_t ep=tokStart(exts,NULL);
			while(tokNext(ep))
				ctx->num_extensions++;
			if(ctx->num_extensions)
			{
				ctx->extensions=(char**)memAlloc(ctx->num_extensions*sizeof(char*));
				tokRestart(ep,NULL);
				for(i=0;tokNext(ep);i++)
					ctx->extensions[i]=strdup(tokLast(ep));
			}
			tokFree(ep);
		}
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_INFO,"vglMakeCurrent(): %s OpenGL on %s ver. %s\n",
			ctx->vendor,ctx->renderer,ctx->version);
#if		0
		conlPuts(LOG_INFO,"OpenGL extensions: \n");
		for(i=0;i<(int)ctx->num_extensions;i++)
			conlPrintf(LOG_INFO,"    %s\n",ctx->extensions[i]);
#endif	/* !0 */
#endif	/* CHECKED_BUILD */
	}
    return true;
}

glw_t	V_GLOBAL
vglGetCurrentWindow(void)
{
	return (glw_t)(GetThreadData()->gl_context);
}

u32_t V_LOCAL
vglPendingCTX(gl_context_t *ctx)
{
	MSG msg;
	while(PeekMessage(&msg,ctx->wnd,0,0,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return fifoCount(&ctx->mqueue);
}

/*
**	Extensions
*/
vboolean	V_GLOBAL
vglHasExtension(char *name)
{
	int i,n;
	gl_context_t *ctx=vglSafeGetCurrentContext("vglHasExtension");
	if(!ctx)
		return false;
	n = ctx->num_extensions;
	for(i=0;i<n;i++)
		if(!strcmp(name,ctx->extensions[i]))
			return true;
	return false;
}

#ifdef	_MSC_VER
#pragma warning(disable:4054)	/* Cast from function to data pointer */
#endif	/* _MSC_VER */

vboolean	V_GLOBAL
vglQueryExtension(char *extension,void **entry)
{
#ifdef	__WIN32__
	*entry=(void*)wglGetProcAddress(extension);
#elif	defined(__UNIX__)
	May be:
	*entry=(void*)glGetProcAddressARB(extension); ???
#endif
	return *entry ? true : false;
}

#ifdef	_MSC_VER
#pragma warning(default:4054)
#endif	/* _MSC_VER */

vboolean	V_GLOBAL
vglPending()
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglPending");
	if(ctx)
		return vglPendingCTX(ctx) ? true : false;
	else
		return false;
}

vboolean	V_GLOBAL
vglNextEvent(input_record_t *event_return)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglNextEvent");
	if(!ctx)
		return false;
	if(!event_return)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"vglNextEvent(NULL): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
		return false;
	}
	while(!vglPendingCTX(ctx))
		WaitMessage();
    return fifoPop(&ctx->mqueue,event_return);
}

vboolean	V_GLOBAL
vglSetFocus()
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglSetFocus");
	if(!ctx)
		return false;
	SetFocus(ctx->wnd);
	return true;
}

vboolean	V_GLOBAL
vglResizeWindow(u32_t width,u32_t height)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglResizeWindow");
	if(!ctx)
		return false;
	return SetWindowPos(ctx->wnd,HWND_NOTOPMOST,
    	0,0,width,height,SWP_NOMOVE|SWP_NOZORDER) ? true : false;
}

vboolean	V_GLOBAL
vglShowWindow(void)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglShowWindow");
	if(!ctx)
		return false;
	if(ShowWindow(ctx->wnd,SW_SHOW))
		if(SetActiveWindow(ctx->wnd))
			return true;
	return false;
}

vboolean	V_GLOBAL
vglHideWindow(void)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglHideWindow");
	if(!ctx)
		return false;
	return ShowWindow(ctx->wnd,SW_HIDE) ? true : false;
}

vboolean	V_GLOBAL
vglMoveWindow(u32_t left,u32_t top)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglMoveWindow");
	if(!ctx)
		return false;
	return SetWindowPos(ctx->wnd,HWND_NOTOPMOST,
    	left,top,640,480,SWP_NOSIZE|SWP_NOZORDER) ? true : false;
}

void		V_GLOBAL
vglSwapBuffers(void)
{
	gl_context_t *ctx=vglSafeGetCurrentContext("vglSwapBuffers");
	if(!ctx)
		return;
	glFlush();
	glFinish();
	SwapBuffers(ctx->dc);
	/* If window is not active, reduce CPU & 3Daccel overhead */
	if(!ctx->focused && !ctx->bitmap_based)
		Sleep(5);	/* Maximum framerate limited to 200 FPS */
}

u8_p		V_GLOBAL
vglReadPixels(u32_t *width,u32_t *height)
{
	u8_t *pixels;
	u32_t x,y,offset,w3,stride_adjust;
	thread_data_t *td;
	gl_context_t *ctx=vglSafeGetCurrentContext("vglReadPixels");
	if(!ctx)
		return NULL;
	td=GetThreadData();
	pixels=(u8_t*)memAlloc(ctx->width*ctx->height*4);
    w3=ctx->width*3;
	stride_adjust=INT_ALIGN(w3)-w3;
    if(ctx->bitmap_based)
	{
		u8_t *ptr=(u8_t*)ctx->bits;
		for(y=ctx->height-1;;y--)
		{
			offset=y*4*ctx->width;
			for(x=0;x<ctx->width;x++)
			{
				*(pixels+offset+2)=ptr[0];
				*(pixels+offset+1)=ptr[1];
				*(pixels+offset)=ptr[2];
				*(pixels+offset+3)=td->constant_alpha;
				offset+=4;
				ptr+=3;
			}
			ptr+=stride_adjust;
			if(!y)
				break;
		}
	}
	else
	{
		u32_t half;
		void *buffer;
		u8_t *upper, *lower;
		glPixelStorei(GL_UNPACK_LSB_FIRST,GL_FALSE);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,0);
		glPixelStorei(GL_UNPACK_SKIP_ROWS,0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS,0);
		glPixelStorei(GL_UNPACK_ALIGNMENT,1);
		glReadPixels(0,0,ctx->width,ctx->height,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
		buffer=memAlloc(ctx->width*4);
		half=ctx->height / 2;
		upper=pixels;
		lower=upper+(ctx->height-1)*ctx->width*4;
		while(half--)
		{
			memCopy32(buffer,upper,ctx->width);
			memCopy32(upper,lower,ctx->width);
			memCopy32(lower,buffer,ctx->width);
			upper+=ctx->width*4;
			lower-=ctx->width*4;
		}
		memFree(buffer);
	}
	if(width)
		*width=ctx->width;
	if(height)
		*height=ctx->height;
	return pixels;
}

/*
**	======================================================================
**	FONT STUFF
**	======================================================================
*/

/*
 *	Validity check.
 */

#ifdef	CHECKED_BUILD 
#define	GLOF_MAGIC		0x474C4F46				  /* "GLOF" */
V_LOCAL	char *glof_object="OpenGL font";
#endif	/* !CHECKED_BUILD */

/*
 *	Returns {arg | current font handle if arg is NULL | NULL if error}.
 *	If defined(CHECKED_BUILD), reports errors to console.
 */
gl_outline_font_p	V_LOCAL
#ifdef	CHECKED_BUILD
vglGetFontForOp(gl_outline_font_t *fnt,char *caller_name)
#else	/* CHECKED_BUILD */
vglGetFontForOp(gl_outline_font_t *fnt)
#endif	/* CHECKED_BUILD */
{
	if(!fnt)
	{
		gl_context_t *ctx=vglSafeGetCurrentContext(caller_name);
		if(!ctx)
			return NULL;
		fnt=ctx->font;
		if(!fnt)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"%s(): no selected font in current context.\n",
				caller_name);
#endif		/* CHECKED_BUILD */
			return NULL;
		}
    }    
   	return (gl_outline_font_t*)libvIsHandle(fnt,caller_name,
       	glof_object,GLOF_MAGIC);
}

#ifdef	CHECKED_BUILD
#define	VGL_GET_FONT_FOR_OP(fnt,caller)		vglGetFontForOp(fnt,caller)
#else	/* CHECKED_BUILD */
#define	VGL_GET_FONT_FOR_OP(fnt,caller)		vglGetFontForOp(fnt)
#endif	/* CHECKED_BUILD */

/*
**		Creation / destruction.
*/

/*
 *	Allocates memory for font structure
 */
V_LOCAL gl_outline_font_p
vglAllocFont(LOGFONT **lf)
{
	gl_outline_font_t *fnt;
	ALLOC_OBJECT(fnt,gl_outline_font_t,
    	256*sizeof(GLYPHMETRICSFLOAT)+sizeof(LOGFONT),GLOF_MAGIC);
	fnt->logfont=(void*)(fnt+1);
	*lf=(LOGFONT*)fnt->logfont;
	fnt->gm=(void*)((*lf)+1);
	return fnt;
}

static	vec_t	font_lod_deviations[GL_FONT_LOD_LEVELS] = {0.001f,0.01f,0.1f};

/*
 *	Creates outlined or polygonal font.
 *	Input:
 *		face_name	- font face (may be NULL);
 *		polygonal	- if zero, creates outlined font;
 *		extrusion	- amount of extrusion;
 *		os_font		- os-dependent handle of font (may be NULL).
 *	Notes.
 *		1. Win32 only: os_font must be HFONT object that was previously
 *	created via CreateFont() / CreateFontIndirect().
 *		2. Win32 only: if os_font is non-NULL, font_face ignored.
 *		3. If both face_name and os_font are NULL, function creates
 *	font with face 'Arial'.
 *	Output:
 *		0 on error, or index+1 of font in current context's 'font' cache.
 */
gl_outline_font_p	V_GLOBAL
vglCreateFont(char *face_name,vboolean polygonal,vec_t extrusion,
	void *os_font)
{
	int i,j;
	HDC dc;				/* Current Win32 DC of RC */
	HFONT old,fnt;		/* Win32 font */
	int format;			/* Format of glyphs: either lines or polygons */
	gl_outline_font_t *f;
	LOGFONT *lf;
	u32_t slot,len;
	gl_context_t *ctx;
	ctx=vglSafeGetCurrentContext("vglCreateFont");
	if(!ctx)
		return NULL;
	/* Check if os_font is real font handle */
	if(os_font && GetObjectType(os_font)!=OBJ_FONT)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateFont('%s',%d,%X): os_font isn't font handle.\n",
			dbs(face_name),polygonal,os_font);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	/* Obtain current RC's DC */
	dc=wglGetCurrentDC();
	if(!dc)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateFont('%s',%d,%X): no current rendering context.\n",
			dbs(face_name),polygonal,os_font);
#endif	/* CHECKED_BUILD */
		return NULL;
	}
	/* Allocate memory for all structures */
	f=vglAllocFont(&lf);
	if(!f)
		return NULL;
	/* Generate OpenGL display lists */
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		f->bases[i]=glGenLists(256);
		if(!f->bases[i])
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglCreateFont('%s',%d,%X): glGenLists() failed, GL error"
				" %d.\n",dbs(face_name),polygonal,os_font,glGetError());
#endif		/* CHECKED_BUILD */
			for(j=i-1;j>=0;j--)
				glDeleteLists(f->bases[i],256);
			memFree(f);
			return NULL;
		}
	}
	/* Create Win32 font and select it into DC */
	if(os_font)
	{
		/* Obtain font info from OS */
		if(GetObjectType((HGDIOBJ)os_font)!=OBJ_FONT)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglCreateFont('%s',%d,%X): %X isn't a font handle.\n",
				dbs(face_name),polygonal,os_font,os_font);
#endif		/* CHECKED_BUILD */
			memFree(f);
			return NULL;
		}
		if(GetObject((HGDIOBJ)os_font,sizeof(LOGFONT),(LPVOID)lf)!=
			sizeof(LOGFONT))
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglCreateFont('%s',%d,%X): GetObject(%X) failed "
				"(code %d).\n",dbs(face_name),polygonal,os_font,os_font,
				GetLastError());
#endif		/* CHECKED_BUILD */
			memFree(f);
			return NULL;
		}
	}
	else
	{
		/* Setup font info explicitly */
		lf->lfHeight=10;
		lf->lfWidth=0;			/* can be any value ? */
		lf->lfEscapement=0;		/* we don't use GDI transforms */
		lf->lfOrientation=0;
		lf->lfWeight=FW_NORMAL; 
		lf->lfItalic=FALSE;
		lf->lfUnderline=FALSE;
		lf->lfStrikeOut=FALSE;
#ifdef	ALWAYS_ENGLISH
		lf->lfCharSet=DEFAULT_CHARSET;
#else	/* ALWAYS_ENGLISH */
		lf->lfCharSet=RUSSIAN_CHARSET;
#endif	/* ALWAYS_ENGLISH */
		lf->lfOutPrecision=OUT_TT_PRECIS;
		lf->lfClipPrecision=CLIP_DEFAULT_PRECIS;
		lf->lfQuality=DEFAULT_QUALITY;
		lf->lfPitchAndFamily=DEFAULT_PITCH;
#ifdef	_UNICODE
#error	Implement unicode version of of vglCreateFont().
#else	/* _UNICODE */
		if(face_name)
			strCopy(lf->lfFaceName,face_name,LF_FACESIZE);
		else
			strCopy(lf->lfFaceName,"Arial",LF_FACESIZE);
#endif	/* _UNICODE */
	}
	fnt=CreateFontIndirect(lf);
	if(!fnt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglCreateFont('%s',%d): CreateFontIndirect() failed "
			"(code %d).\n",dbs(face_name),polygonal,GetLastError());
#endif	/* CHECKED_BUILD */
		for(j=i-1;j>=0;j--)
			glDeleteLists(f->bases[i],256);
		memFree(f);
		return NULL;
	}
	f->hfont=fnt;
	old=SelectObject(dc,fnt);
	/* Convert quadratic B-spline glyphs to lines or polygons */
	format = polygonal ? WGL_FONT_POLYGONS : WGL_FONT_LINES;
	f->polygonal=polygonal;
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		wglUseFontOutlines(dc,0,256,f->bases[i],
			(GLfloat)font_lod_deviations[i],
			(GLfloat)extrusion,
			format,(LPGLYPHMETRICSFLOAT)f->gm);
	}
	/* Restore DC to it's previous state */
	SelectObject(dc,old);
	f->polygonal=polygonal;
	f->extrusion=extrusion;
	f->ctx_origin=(glw_t)ctx;
	f->ref=NULL;
	len=vpsCount(&ctx->fonts);
	for(slot=0;slot<len;slot++)
		if(!vpsGet(&ctx->fonts,slot))
			break;
	if(slot<len)
		vpsSet(&ctx->fonts,slot,f);
	else
		vpsPush(&ctx->fonts,f);
	f->cache_index=slot;
	/* Initial drawing values for font */
	f->sp.adjustment=fa_center_center;
	vec3Set(f->sp.custom_adjustment,-0.5f,-0.5f,0);
	f->sp.scale=1;
	f->sp.fov=ANGLE_10;
	f->sp.use_fov=false;
	f->magic=GLOF_MAGIC;
	f->lod=0;	/* force LOD update */
	vglSetFontLOD(f,1);
	return f;
}

/*
 *	Selects font created in another window; returns quickly if font 
 *	already already exists or duplicates font if not exists.
 *	Assumes 'origin' is valid font handle.
 */
gl_outline_font_p	V_LOCAL
vglDuplicateFont(gl_outline_font_t *origin)
{
	gl_context_t *ctx;
	u32_t slot,len;
	gl_outline_font_t *f;
	HFONT old,fnt;
	int i,j,format;
	LOGFONT *lf;
	ctx=vglSafeGetCurrentContext("vglDuplicateFont");
	if(!ctx)
		return NULL;
	len=vpsCount(&ctx->fonts);
	for(slot=0;slot<len;slot++)
	{
		f=(gl_outline_font_t*)vpsGet(&ctx->fonts,slot);
		if(f && (f==origin || f->ref==origin))
			/* Already duplicated */
			return f;
	}
	/* Duplicate font */
	f=vglAllocFont(&lf);
	if(!f)
		return NULL;
    for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		f->bases[i]=glGenLists(256);
		if(!f->bases[i])
		{
			int j;
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"vglDuplicateFont(%X): glGenLists() failed, GL "
				"error %d.\n",origin,glGetError());
#endif		/* CHECKED_BUILD */
			for(j=i-1;j>=0;j--)
				glDeleteLists(f->bases[i],256);
			memFree(f);
			return NULL;
		}
	}
	/* Create Win32 font and select it into DC */
	memCopy(lf,origin->logfont,sizeof(LOGFONT));
	fnt=CreateFontIndirect(lf);
	if(!fnt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vglDuplicateFont(): CreateFontIndirect() failed "
			"(code %d).\n",GetLastError());
#endif	/* CHECKED_BUILD */
		for(j=i-1;j>=0;j--)
			glDeleteLists(f->bases[i],256);
		memFree(f);
		return NULL;
	}
	f->hfont=fnt;
	old=SelectObject(ctx->dc,fnt);
	/* Convert quadratic B-spline glyphs to lines or polygons */
	format = origin->polygonal ? WGL_FONT_POLYGONS : WGL_FONT_LINES;
	f->polygonal=origin->polygonal;
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
	{
		wglUseFontOutlines(ctx->dc,0,256,f->bases[i],
			(GLfloat)font_lod_deviations[i],
			(GLfloat)origin->extrusion,
			format,(LPGLYPHMETRICSFLOAT)f->gm);
	}
	/* Restore DC to it's previous state */
	SelectObject(ctx->dc,old);
	f->magic=GLOF_MAGIC;
	f->extrusion=origin->extrusion;
	f->ctx_origin=(glw_t)ctx;
	f->ref=origin;
	f->line_width=origin->line_width;
	f->line_smooth=origin->line_smooth;
	f->lod=(1+origin->lod) % GL_FONT_LOD_LEVELS;	/* force LOD update */
	vglSetFontLOD(f,origin->lod);
	len=vpsCount(&ctx->fonts);
	for(slot=0;slot<len;slot++)
		if(!vpsGet(&ctx->fonts,slot))
			break;
	if(slot<len)
		vpsSet(&ctx->fonts,slot,f);
	else
		vpsPush(&ctx->fonts,f);
	f->cache_index=slot;
	return f;
}

/*
 *	Releases all objects and frees memory, but not updates state of 
 *	context.
 *	Assumes 'fnt' is valid font handle.
 */
void	V_LOCAL
vglReleaseFont(gl_outline_font_t *fnt)
{
	int i;
	if(!fnt)
		return;
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
		glDeleteLists(fnt->bases[i],256);
	if(fnt->hfont)
		DeleteObject(fnt->hfont);
	memFree(fnt);
}

/*
 *	Destroys given font. Font MUST belong to current context.
 */
void	V_GLOBAL
vglDeleteFont(gl_outline_font_t *fnt)
{
	u32_t i,len;
	gl_context_t *ctx;
	/* gl_outline_font_t *origin; */
    if(!libvIsHandle(fnt,"vglDeleteFont",glof_object,GLOF_MAGIC))
		return;
	/* origin=fnt->ref; */
	ctx=vglSafeGetCurrentContext("vglDeleteFont");
    if(!ctx)
    	return;
    if(ctx!=fnt->ctx_origin)
    {
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,"vglDeleteFont(): "
        	"font belongs to context other than current.\n");
#endif	/* CHECKED_BUILD */
    	return;
    }
	len=vpsCount(&ctx->fonts);
	for(i=0;i<len;i++)
		if(vpsGet(&ctx->fonts,i)==fnt)
		{
			vpsSet(&ctx->fonts,i,NULL);
			break;
		}
	if(i==len)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"vglDeleteFont(%X): font not fount in current context.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	vglReleaseFont(fnt);
	return;
}	

/*
 *	Adjust OpenGL state depending on font parameters.
 *	Assumes 'fnt' is valid.
 */
void	V_LOCAL
vglRealizeFont(gl_outline_font_t *fnt)
{
	glListBase(fnt->bases[fnt->lod]);
	if(!fnt->polygonal)
	{
		glLineWidth(fnt->line_width);
		if(fnt->line_smooth)
		{
			glEnable(GL_LINE_SMOOTH);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		}
		else
		{
			glDisable(GL_BLEND);
			glDisable(GL_LINE_SMOOTH);
		}
	}
}

/*
 *	Selects given font. Returns true on success.
 *	Fonts created in other glw CAN be selected into other (also it takes
 *	very long time at first call).
 */
vboolean	V_GLOBAL
vglSelectFont(gl_outline_font_t *fnt)
{
	gl_context_t *ctx;
    if(!libvIsHandle(fnt,"vglSelectFont",glof_object,GLOF_MAGIC))
		return false;
	ctx=vglSafeGetCurrentContext("vglSelectFont");
	if(ctx!=fnt->ctx_origin)
	{
		/* DON'T DO THAT: 'fnt=vglDuplicateFont(fnt)' ! */
    	vglDuplicateFont(fnt);
    }
    ctx->font=fnt;
	vglRealizeFont(fnt);
	return true;
}

/*
 *	Returns currently selected font or NULL.
 */
gl_outline_font_p	V_GLOBAL
vglGetCurrentFont(void)
{
	gl_context_t *ctx;
    ctx=vglSafeGetCurrentContext("vglGetCurrentFont");
	if(!ctx)
    	return NULL;
	return ctx->font;
}

/*
 *	Returns text x-extent and height from the base line.
 */
void	V_GLOBAL
vglGetTextExtents(char *str,vec_t *width,vec_t *height)
{
	LPGLYPHMETRICSFLOAT gmf;
	vec_t w,h,hc;
	unsigned char *ptr;
	gl_outline_font_t *fnt;
	if(!str)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglGetTextExtents([NULL],%X,%X): "
        	"incorrect parameter.\n",width,height);
#endif	/* CHECKED_BUILD */
		return;
	}
    fnt=vglGetCurrentFont();
	if(!fnt)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vglGetTextExtents(%s,%X,%X): "
        	"no selected font in current context.\n",dbs(str),
            width,height);
#endif	/* CHECKED_BUILD */
		return;
	}
	gmf=(LPGLYPHMETRICSFLOAT)fnt->gm;
	w=h=0;
	for(ptr=(unsigned char*)str;*ptr;ptr++)
	{
		hc=gmf[*ptr].gmfptGlyphOrigin.y;
		if(h<hc)
			h=hc;
		w+=gmf[*ptr].gmfCellIncX;
	}
	if(width)
		*width=w;
	if(height)
		*height=h;
}

/*
 *	Prints string.
 */
void	V_GLOBAL
vglPuts(char *str)
{
	if(str)
		glCallLists(strlen(str),GL_UNSIGNED_BYTE,str);
#ifdef	CHECKED_BUILD
	else
		conlPuts(LOG_ERROR,
			"vglPuts([NULL]): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
}

/*
 *	OpenGL "printf()".
 */
void	V_GLOBAL_VA
vglPrintf(const char *fmt,...)
{
	va_list marker;
	char 	*ptr;
	char 	fmt_buffer[MAX_FORMAT_BUFFER];
	if(fmt)
	{
		ptr=fmt_buffer;
		va_start(marker,fmt);
		vsprintf(ptr,fmt,marker);
		va_end(marker);
		strStripTrail(ptr);
		glCallLists(strlen(ptr),GL_UNSIGNED_BYTE,ptr);
	}
#ifdef	CHECKED_BUILD
	else
		conlPuts(LOG_ERROR,
			"vglPrintf([NULL]): incorrect parameter.\n");
#endif	/* CHECKED_BUILD */
}

/*
 *	Adjusts font LOD depending character size.
 */
void	V_GLOBAL
vglSetFontLOD(gl_outline_font_t *fnt,u32_t level_of_detail)
{
	fnt=VGL_GET_FONT_FOR_OP(fnt,"vglSetFontLOD");
	if(!fnt)
		return;
	if(level_of_detail>(u32_t)(GL_FONT_LOD_LEVELS-1))
		level_of_detail=(u32_t)(GL_FONT_LOD_LEVELS-1);
	if(fnt->lod!=level_of_detail)
	{
		fnt->lod=level_of_detail;
		switch(level_of_detail)
		{
			case 0:
				fnt->line_smooth=1;
				fnt->line_width=3.0;
				break;
			case 1:
				fnt->line_smooth=1;
				fnt->line_width=2.0;
				break;
			case 2:
				fnt->line_smooth=0;
				fnt->line_width=1.0;
				break;
		}
	}
	vglRealizeFont(fnt);
}

/*
 *	There are 9 useful pivot point positions for text:
 *			left+top;			center+top;				right+top;
 *			left+center;		center+center;			right+center;
 *			left+bottom;		center+bottom;			right+bottom.
 */
V_LOCAL	vec3_t	label_origins[9]=
{
	{0,-1,0},	{-0.5,-1,0},	{-1,-1,0},
	{0,-0.5,0},	{-0.5,-0.5,0},	{-1,-0.5,0},
	{0,0,0},	{-0.5,0,0},		{-1,0,0}
};


/*
	TODO: vglFontRasterQuality / vglFontMetafileQuality ...
*/
/*
V_LOCAL	vec_t	ref_flod_limits[GL_FONT_LOD_LEVELS]={0.25f,0.1f,0};
*/
V_LOCAL	vec_t	wmf_flod_limits[GL_FONT_LOD_LEVELS]={0.05f,0.02f,0};

V_LOCAL	vec_p	flod_limits=wmf_flod_limits;

/*
 *	Sets LOD based on scale and view size (font height is about 1.0).
 *	Used in vglDrawString() / vglDrawText().
 */
void	V_LOCAL
vglSetFontLODFromScale(gl_outline_font_t *fnt,vec_t scale)
{
	int i;
	scale=v_abs(scale);
	for(i=0;i<GL_FONT_LOD_LEVELS;i++)
		if(scale>flod_limits[i])
		{
			vglSetFontLOD(fnt,i);
			return;
		}
}

/*
 *	Loads appropriate matrix on modelview TOS.
 *	Assumes all args are valid.
 */
void	V_LOCAL
vglSetupStringTransform(
	gl_outline_font_t *fnt,
	vec3_t origin,
	cam_t *cam,
	char *text)
{
	vec3_t t,v;
	vec_t w,h,scale,lod_scale,t_len;
	glLoadIdentity();
	mat3x4VecMul(t,cam->transform,origin);
	t_len=vec3Abs(t);
	if(fnt->sp.use_fov || IsZero(t_len))
	{
		lod_scale=(vec_t)tan(fnt->sp.fov * .5);
		scale = lod_scale * t_len;
	}
	else
	{
		scale=fnt->sp.scale;
		lod_scale = scale / t_len;
	}
	vglGetTextExtents(text,&w,&h);
	if((unsigned)(fnt->sp.adjustment) > (unsigned)fa_right_bottom)
		vec3Negate(v,fnt->sp.custom_adjustment);
	else
		vec3Copy(v,label_origins[fnt->sp.adjustment]);
	v[0]*=w;
	v[1]*=h;
	v[2]=fnt->extrusion * .5f;
	vec3Scale(v,v,scale);
	vglSetFontLODFromScale(fnt,lod_scale);
	vec3Add(t,t,v);
	glTranslatef(t[0],t[1],t[2]);
	glScalef(scale,scale,scale);
}

/*
 *		vglSetFontAdjustment
 */
void	V_GLOBAL
vglSetFontAdjustment(gl_outline_font_t *fnt,gl_font_adjustment_t adjustment)
{
	if(!libvIsHandle(fnt,"vglSetFontAdjustment",glof_object,GLOF_MAGIC))
		return;
	fnt->sp.adjustment=adjustment;
}

/*
 *	custom_adjustment contain coordinates for text pivot point;
 *	{0,0} mean "left,bottom", {1,1} mean "right,top".
 */
void	V_GLOBAL
vglSetFontCustomAdjustment(gl_outline_font_t *fnt,vec3_t custom_adjustment)
{
	if(!libvIsHandle(fnt,"vglSetFontCustomAdjustment",glof_object,GLOF_MAGIC))
		return;
	if(!custom_adjustment)
	{
		return;
	}
	fnt->sp.adjustment=fa_custom;
	vec3Copy(fnt->sp.custom_adjustment,custom_adjustment);
}

/*
 *	vglSetFontScale()
 */
void	V_GLOBAL
vglSetFontScale(gl_outline_font_t *fnt,vec_t scale)
{
	if(!libvIsHandle(fnt,"vglSetFontScale",glof_object,GLOF_MAGIC))
		return;
	fnt->sp.use_fov=false;
	fnt->sp.scale=scale;
}

/*
 *	vglSetFontFOV
 */
void	V_GLOBAL
vglSetFontFOV(gl_outline_font_t *fnt,vec_t fov)
{
	if(!libvIsHandle(fnt,"vglSetFontFOV",glof_object,GLOF_MAGIC))
		return;
	fnt->sp.use_fov=true;
	fnt->sp.fov=fov;
}

/*
 *	vglDrawString()
 */
void	V_GLOBAL
vglDrawString(
	vec3_t origin,
	cam_t *cam,
	char *text)
{
	gl_outline_font_t *fnt=vglGetCurrentFont();
	if(!fnt)
	{
		/* this place can produce a lot of errors, so we set log level
		 * to LOG_TRACE only */
#ifdef	CHECKED_BUILD
		conlPuts(LOG_TRACE,"vglDrawString(): no current font.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	glPushMatrix();
	vglSetupStringTransform(fnt,origin,cam,text);
	vglPuts(text);
	glPopMatrix();
}

/*		=============================================================== */

/*</pre></strong></body></html>*/
/*EOF*/
